@@ -59,6 +59,7 @@ static void new_type(MVMThreadContext *tc, MVMCallsite *callsite, MVMRegister *a
     /* See if we were given a name; put it into the meta-object if so. */
     name = name_arg.exists ? name_arg.arg.s : instance->str_consts.anon;
     MVM_ASSIGN_REF(tc, &(HOW->header), ((MVMKnowHOWREPR *)HOW)->body.name, name);
+    type_object->st->debug_name = MVM_string_utf8_encode_C_string(tc, name);
 
     /* Set .WHO to an empty hash. */
     BOOTHash = tc->instance->boot_types.BOOTHash;
@@ -309,7 +310,7 @@ static void bootstrap_KnowHOW(MVMThreadContext *tc) {
 
     /* Stash the created KnowHOW. */
     tc->instance->KnowHOW = (MVMObject *)knowhow;
-    MVM_gc_root_add_permanent(tc, (MVMCollectable **)&tc->instance->KnowHOW);
+    MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&tc->instance->KnowHOW, "KnowHOW");
 }
 
 /* Takes a stub object that existed before we had bootstrapped things and
@@ -327,6 +328,7 @@ static void add_meta_object(MVMThreadContext *tc, MVMObject *type_obj, char *nam
         /* Set name. */
         name_str = MVM_string_ascii_decode_nt(tc, tc->instance->VMString, name);
         MVM_ASSIGN_REF(tc, &(meta_obj->header), ((MVMKnowHOWREPR *)meta_obj)->body.name, name_str);
+        type_obj->st->debug_name = strdup(name);
     });
 }
 
@@ -444,7 +446,9 @@ static void create_KnowHOWAttribute(MVMThreadContext *tc) {
 
         /* Stash the created type object. */
         tc->instance->KnowHOWAttribute = (MVMObject *)type_obj;
-        MVM_gc_root_add_permanent(tc, (MVMCollectable **)&tc->instance->KnowHOWAttribute);
+        MVM_gc_root_add_permanent_desc(tc,
+            (MVMCollectable **)&tc->instance->KnowHOWAttribute,
+            "KnowHOWAttribute");
     });
 }
 
@@ -453,7 +457,7 @@ static MVMObject * boot_typed_array(MVMThreadContext *tc, char *name, MVMObject
     MVMBoolificationSpec *bs;
     MVMObject  *repr_info;
     MVMInstance  *instance  = tc->instance;
-    const MVMREPROps *repr  = MVM_repr_get_by_id(tc, MVM_REPR_ID_MVMArray);
+    const MVMREPROps *repr  = MVM_repr_get_by_id(tc, MVM_REPR_ID_VMArray);
     MVMObject  *array = repr->type_object_for(tc, NULL);
     MVMROOT(tc, array, {
         /* Give it a meta-object. */
@@ -524,7 +528,7 @@ static void string_consts(MVMThreadContext *tc) {
 /* Set up some strings. */
 #define string_creator(variable, name) do { \
     instance->str_consts.variable = MVM_string_ascii_decode_nt(tc, tc->instance->VMString, (name)); \
-    MVM_gc_root_add_permanent(tc, (MVMCollectable **)&(instance->str_consts.variable)); \
+    MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&(instance->str_consts.variable), "VM string constant"); \
 } while (0)
 
     string_creator(empty, "");
@@ -557,18 +561,27 @@ static void string_consts(MVMThreadContext *tc) {
     string_creator(auto_viv_container, "auto_viv_container");
     string_creator(done, "done");
     string_creator(error, "error");
-    string_creator(stdout_chars, "stdout_chars");
     string_creator(stdout_bytes, "stdout_bytes");
-    string_creator(stderr_chars, "stderr_chars");
     string_creator(stderr_bytes, "stderr_bytes");
+    string_creator(merge_bytes, "merge_bytes");
     string_creator(buf_type, "buf_type");
     string_creator(write, "write");
+    string_creator(stdin_fd, "stdin_fd");
+    string_creator(stdout_fd, "stdout_fd");
+    string_creator(stderr_fd, "stderr_fd");
     string_creator(nativeref, "nativeref");
     string_creator(refkind, "refkind");
     string_creator(positional, "positional");
     string_creator(lexical, "lexical");
     string_creator(dimensions, "dimensions");
     string_creator(ready, "ready");
+    string_creator(multidim, "multidim");
+    string_creator(entry_point, "entry_point");
+    string_creator(kind, "kind");
+    string_creator(instrumented, "instrumented");
+    string_creator(heap, "heap");
+    string_creator(translate_newlines, "translate_newlines");
+    string_creator(platform_newline, MVM_TRANSLATE_NEWLINE_OUTPUT ? "\r\n" : "\n");
 }
 
 /* Drives the overall bootstrap process. */
@@ -603,7 +616,7 @@ void MVM_6model_bootstrap(MVMThreadContext *tc) {
     create_stub_boot_type(tc, MVM_REPR_ID_P6int, boot_types.BOOTInt, 1, MVM_BOOL_MODE_UNBOX_INT);
     create_stub_boot_type(tc, MVM_REPR_ID_P6num, boot_types.BOOTNum, 1, MVM_BOOL_MODE_UNBOX_NUM);
     create_stub_boot_type(tc, MVM_REPR_ID_P6str, boot_types.BOOTStr, 1, MVM_BOOL_MODE_UNBOX_STR_NOT_EMPTY);
-    create_stub_boot_type(tc, MVM_REPR_ID_MVMArray, boot_types.BOOTArray, 1, MVM_BOOL_MODE_HAS_ELEMS);
+    create_stub_boot_type(tc, MVM_REPR_ID_VMArray, boot_types.BOOTArray, 1, MVM_BOOL_MODE_HAS_ELEMS);
     create_stub_boot_type(tc, MVM_REPR_ID_MVMHash, boot_types.BOOTHash, 1, MVM_BOOL_MODE_HAS_ELEMS);
     create_stub_boot_type(tc, MVM_REPR_ID_MVMCFunction, boot_types.BOOTCCode, 0, MVM_BOOL_MODE_NOT_TYPE_OBJECT);
     create_stub_boot_type(tc, MVM_REPR_ID_MVMCode, boot_types.BOOTCode, 0, MVM_BOOL_MODE_NOT_TYPE_OBJECT);
@@ -630,7 +643,7 @@ void MVM_6model_bootstrap(MVMThreadContext *tc) {
     /* Give stub types meta-objects. */
 #define meta_objectifier(tc, slot, name) do { \
     add_meta_object((tc), (tc)->instance->slot, (name)); \
-    MVM_gc_root_add_permanent((tc), (MVMCollectable **)&(tc)->instance->slot); \
+    MVM_gc_root_add_permanent_desc((tc), (MVMCollectable **)&(tc)->instance->slot, name); \
 } while (0)
     meta_objectifier(tc, VMString, "VMString");
     meta_objectifier(tc, VMNull, "VMNull");
@@ -664,13 +677,19 @@ void MVM_6model_bootstrap(MVMThreadContext *tc) {
     /* Bootstrap typed arrays. */
     tc->instance->boot_types.BOOTIntArray = boot_typed_array(tc, "BOOTIntArray",
         tc->instance->boot_types.BOOTInt);
-    MVM_gc_root_add_permanent(tc, (MVMCollectable **)&tc->instance->boot_types.BOOTIntArray);
+    MVM_gc_root_add_permanent_desc(tc,
+        (MVMCollectable **)&tc->instance->boot_types.BOOTIntArray,
+        "BOOTIntArray");
     tc->instance->boot_types.BOOTNumArray = boot_typed_array(tc, "BOOTNumArray",
         tc->instance->boot_types.BOOTNum);
-    MVM_gc_root_add_permanent(tc, (MVMCollectable **)&tc->instance->boot_types.BOOTNumArray);
+    MVM_gc_root_add_permanent_desc(tc,
+        (MVMCollectable **)&tc->instance->boot_types.BOOTNumArray,
+        "BOOTNumArray");
     tc->instance->boot_types.BOOTStrArray = boot_typed_array(tc, "BOOTStrArray",
         tc->instance->boot_types.BOOTStr);
-    MVM_gc_root_add_permanent(tc, (MVMCollectable **)&tc->instance->boot_types.BOOTStrArray);
+    MVM_gc_root_add_permanent_desc(tc,
+        (MVMCollectable **)&tc->instance->boot_types.BOOTStrArray,
+        "BOOTStrArray");
 
     /* Set up HLL roles. */
     STABLE(tc->instance->boot_types.BOOTInt)->hll_role   = MVM_HLL_ROLE_INT;
