@@ -1,7 +1,7 @@
 #include "moar.h"
 
 /* This representation's function pointer table. */
-static const MVMREPROps this_repr;
+static const MVMREPROps CPPStruct_this_repr;
 
 /* Locates all of the attributes. Puts them onto a flattened, ordered
  * list of attributes (populating the passed flat_list). Also builds
@@ -309,7 +309,7 @@ static MVMint32 try_get_slot(MVMThreadContext *tc, MVMCPPStructREPRData *repr_da
 /* Creates a new type object of this representation, and associates it with
  * the given HOW. */
 static MVMObject * type_object_for(MVMThreadContext *tc, MVMObject *HOW) {
-    MVMSTable *st  = MVM_gc_allocate_stable(tc, &this_repr, HOW);
+    MVMSTable *st  = MVM_gc_allocate_stable(tc, &CPPStruct_this_repr, HOW);
 
     MVMROOT(tc, st, {
         MVMObject *obj = MVM_gc_allocate_type_object(tc, st);
@@ -673,7 +673,7 @@ static void gc_cleanup(MVMThreadContext *tc, MVMSTable *st, void *data) {
 
 /* Called by the VM in order to free memory associated with this object. */
 static void gc_free(MVMThreadContext *tc, MVMObject *obj) {
-	gc_cleanup(tc, STABLE(obj), OBJECT_BODY(obj));
+    gc_cleanup(tc, STABLE(obj), OBJECT_BODY(obj));
 }
 
 static const MVMStorageSpec storage_spec = {
@@ -786,10 +786,10 @@ static void deserialize_stable_size(MVMThreadContext *tc, MVMSTable *st, MVMSeri
 
 /* Initializes the representation. */
 const MVMREPROps * MVMCPPStruct_initialize(MVMThreadContext *tc) {
-    return &this_repr;
+    return &CPPStruct_this_repr;
 }
 
-static const MVMREPROps this_repr = {
+static const MVMREPROps CPPStruct_this_repr = {
     type_object_for,
     MVM_gc_allocate_object,
     initialize,
