@@ -1,12 +1,12 @@
 #include "moar.h"
 
 /* This representation's function pointer table. */
-static const MVMREPROps this_repr;
+static const MVMREPROps NFA_this_repr;
 
 /* Creates a new type object of this representation, and associates it with
  * the given HOW. */
 static MVMObject * type_object_for(MVMThreadContext *tc, MVMObject *HOW) {
-    MVMSTable *st = MVM_gc_allocate_stable(tc, &this_repr, HOW);
+    MVMSTable *st = MVM_gc_allocate_stable(tc, &NFA_this_repr, HOW);
 
     MVMROOT(tc, st, {
         MVMObject *obj = MVM_gc_allocate_type_object(tc, st);
@@ -243,10 +243,10 @@ static MVMuint64 unmanaged_size(MVMThreadContext *tc, MVMSTable *st, void *data)
 
 /* Initializes the representation. */
 const MVMREPROps * MVMNFA_initialize(MVMThreadContext *tc) {
-    return &this_repr;
+    return &NFA_this_repr;
 }
 
-static const MVMREPROps this_repr = {
+static const MVMREPROps NFA_this_repr = {
     type_object_for,
     MVM_gc_allocate_object,
     NULL, /* initialize */
@@ -318,8 +318,8 @@ MVMObject * MVM_nfa_from_statelist(MVMThreadContext *tc, MVMObject *states, MVMO
         num_states = MVM_repr_elems(tc, states) - 1;
         nfa->num_states = num_states;
         if (num_states > 0) {
-            nfa->num_state_edges = MVM_malloc(num_states * sizeof(MVMint64));
-            nfa->states = MVM_malloc(num_states * sizeof(MVMNFAStateInfo *));
+            nfa->num_state_edges = MVM_calloc(num_states, sizeof(MVMint64));
+            nfa->states = MVM_calloc(num_states, sizeof(MVMNFAStateInfo *));
         }
         for (i = 0; i < num_states; i++) {
             MVMObject *edge_info = MVM_repr_at_pos_o(tc, states, i + 1);
@@ -336,6 +336,8 @@ MVMObject * MVM_nfa_from_statelist(MVMThreadContext *tc, MVMObject *states, MVMO
                     MVM_repr_at_pos_o(tc, edge_info, j));
                 MVMint64 to  = MVM_coerce_simple_intify(tc,
                     MVM_repr_at_pos_o(tc, edge_info, j + 2));
+                if (to <= 0 && act != MVM_NFA_EDGE_FATE)
+                    MVM_exception_throw_adhoc(tc, "Invalid to edge %"PRId64" in NFA statelist", to);
 
                 nfa->states[i][cur_edge].act = act;
                 nfa->states[i][cur_edge].to = to;
@@ -454,7 +456,7 @@ static MVMint64 * nqp_nfa_run(MVMThreadContext *tc, MVMNFABody *nfa, MVMString *
         if (nfadeb) {
             if (offset < eos) {
                 MVMGrapheme32 cp = MVM_string_get_grapheme_at_nocheck(tc, target, offset);
-                fprintf(stderr,"%c with %ds target %lx offset %lld\n",cp,(int)numcur, (long)target, offset);
+                fprintf(stderr,"%c with %ds target %lx offset %"PRId64"\n",cp,(int)numcur, (long)target, offset);
             }
             else {
                 fprintf(stderr,"EOS with %ds\n",(int)numcur);
@@ -492,7 +494,6 @@ static MVMint64 * nqp_nfa_run(MVMThreadContext *tc, MVMNFABody *nfa, MVMString *
                         MVMint64 arg = edge_info[i].arg.i;
                         MVMint64 j;
                         MVMint64 found_fate = 0;
-                        arg &= 0xffffff;   /* XXX can go away after rebootstrap */
                         if (nfadeb)
                             fprintf(stderr, "fate(%016llx) ", (long long unsigned int)arg);
                         for (j = 0; j < total_fates; j++) {
@@ -528,10 +529,10 @@ static MVMint64 * nqp_nfa_run(MVMThreadContext *tc, MVMNFABody *nfa, MVMString *
                         continue;
                     }
                     else if (act == MVM_NFA_EDGE_EPSILON && to <= num_states && done[to] != gen) {
-			if (to)
-			    curst[numcur++] = to;
-			else if (nfadeb)  /* XXX should turn into a "can't happen" after rebootstrap */
-			    fprintf(stderr, "  oops, ignoring epsilon to 0\n");
+                        if (to)
+                            curst[numcur++] = to;
+                        else if (nfadeb)  /* XXX should turn into a "can't happen" after rebootstrap */
+                            fprintf(stderr, "  oops, ignoring epsilon to 0\n");
                         continue;
                     }
                 }
@@ -661,7 +662,7 @@ static MVMint64 * nqp_nfa_run(MVMThreadContext *tc, MVMNFABody *nfa, MVMString *
                             if (((act == MVM_NFA_EDGE_CODEPOINT_M)     && (ga == gb))
                              || ((act == MVM_NFA_EDGE_CODEPOINT_M_NEG) && (ga != gb)))
                                 nextst[numnext++] = to;
-
+                            MVM_unicode_normalizer_cleanup(tc, &norm);
                             continue;
                         }
                         case MVM_NFA_EDGE_CODEPOINT_IM:
@@ -677,6 +678,7 @@ static MVMint64 * nqp_nfa_run(MVMThreadContext *tc, MVMNFABody *nfa, MVMString *
                             MVM_unicode_normalizer_eof(tc, &norm);
                             if (!ready)
                                 uc_arg = MVM_unicode_normalizer_get_grapheme(tc, &norm);
+                            MVM_unicode_normalizer_cleanup(tc, &norm);
 
                             MVM_unicode_normalizer_init(tc, &norm, MVM_NORMALIZE_NFD);
                             ready = MVM_unicode_normalizer_process_codepoint_to_grapheme(tc, &norm, lc_arg, &lc_arg);
@@ -687,6 +689,7 @@ static MVMint64 * nqp_nfa_run(MVMThreadContext *tc, MVMNFABody *nfa, MVMString *
                             if (((act == MVM_NFA_EDGE_CODEPOINT_IM)     && (ord == lc_arg || ord == uc_arg))
                              || ((act == MVM_NFA_EDGE_CODEPOINT_IM_NEG) && (ord != lc_arg && ord != uc_arg)))
                                 nextst[numnext++] = to;
+                            MVM_unicode_normalizer_cleanup(tc, &norm);
                             continue;
                         }
                         case MVM_NFA_EDGE_CHARRANGE_M: {
