@@ -1,12 +1,12 @@
 #include "moar.h"
 
 /* This representation's function pointer table. */
-static const MVMREPROps this_repr;
+static const MVMREPROps NativeRef_this_repr;
 
 /* Creates a new type object of this representation, and associates it with
  * the given HOW. */
 static MVMObject * type_object_for(MVMThreadContext *tc, MVMObject *HOW) {
-    MVMSTable *st  = MVM_gc_allocate_stable(tc, &this_repr, HOW);
+    MVMSTable *st  = MVM_gc_allocate_stable(tc, &NativeRef_this_repr, HOW);
 
     MVMROOT(tc, st, {
         MVMObject *obj = MVM_gc_allocate_type_object(tc, st);
@@ -53,8 +53,8 @@ static void gc_mark(MVMThreadContext *tc, MVMSTable *st, void *data, MVMGCWorkli
     MVMNativeRefBody *ref = (MVMNativeRefBody *)data;
     MVMNativeRefREPRData *repr_data = (MVMNativeRefREPRData *)st->REPR_data;
     switch (repr_data->ref_kind) {
-        case MVM_NATIVEREF_REG_OR_LEX:
-            MVM_gc_worklist_add(tc, worklist, &ref->u.reg_or_lex.frame);
+        case MVM_NATIVEREF_LEX:
+            MVM_gc_worklist_add(tc, worklist, &ref->u.lex.frame);
             break;
         case MVM_NATIVEREF_ATTRIBUTE:
             MVM_gc_worklist_add(tc, worklist, &ref->u.attribute.obj);
@@ -103,7 +103,7 @@ static void compose(MVMThreadContext *tc, MVMSTable *st, MVMObject *info_hash) {
                 MVMuint16 kind;
                 MVMString *refkind_s = MVM_repr_get_str(tc, refkind);
                 if (MVM_string_equal(tc, refkind_s, str_consts->lexical)) {
-                    kind = MVM_NATIVEREF_REG_OR_LEX;
+                    kind = MVM_NATIVEREF_LEX;
                 }
                 else if (MVM_string_equal(tc, refkind_s, str_consts->attribute)) {
                     kind = MVM_NATIVEREF_ATTRIBUTE;
@@ -138,7 +138,7 @@ static void compose(MVMThreadContext *tc, MVMSTable *st, MVMObject *info_hash) {
 
 /* Initializes the representation. */
 const MVMREPROps * MVMNativeRef_initialize(MVMThreadContext *tc) {
-    return &this_repr;
+    return &NativeRef_this_repr;
 }
 
 static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpeshBB *bb, MVMSpeshIns *ins) {
@@ -148,7 +148,7 @@ static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpes
     if (!repr_data)
         return;
 
-    if (repr_data->ref_kind != MVM_NATIVEREF_REG_OR_LEX)
+    if (repr_data->ref_kind != MVM_NATIVEREF_LEX)
         return; /* TODO implement spesh for attribute and positional references */
 
     switch (opcode) {
@@ -161,7 +161,7 @@ static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpes
             ins->operands            = MVM_spesh_alloc(tc, g, 3 * sizeof(MVMSpeshOperand));
             ins->operands[0]         = target;
             ins->operands[1]         = value;
-            ins->operands[2].lit_i64 = offsetof(MVMNativeRef, body.u.reg_or_lex.var);
+            ins->operands[2].lit_i64 = offsetof(MVMNativeRef, body.u.lex.var);
             break;
         }
         case MVM_OP_decont_i: {
@@ -173,13 +173,13 @@ static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpes
             ins->operands            = MVM_spesh_alloc(tc, g, 3 * sizeof(MVMSpeshOperand));
             ins->operands[0]         = target;
             ins->operands[1]         = source;
-            ins->operands[2].lit_i64 = offsetof(MVMNativeRef, body.u.reg_or_lex.var);
+            ins->operands[2].lit_i64 = offsetof(MVMNativeRef, body.u.lex.var);
             break;
         }
     }
 }
 
-static const MVMREPROps this_repr = {
+static const MVMREPROps NativeRef_this_repr = {
     type_object_for,
     MVM_gc_allocate_object,
     NULL, /* initialize */
@@ -226,38 +226,17 @@ void MVM_nativeref_ensure(MVMThreadContext *tc, MVMObject *type, MVMuint16 wantp
     }
 }
 
-/* Creation of native references for registers. */
-static MVMObject * reg_or_lex_ref(MVMThreadContext *tc, MVMObject *type, MVMFrame *f, MVMRegister *r, MVMuint16 reg_type) {
+/* Creation of native references for lexicals. */
+static MVMObject * lex_ref(MVMThreadContext *tc, MVMObject *type, MVMFrame *f, MVMRegister *r, MVMuint16 reg_type) {
     MVMNativeRef *ref;
     MVMROOT(tc, f, {
         ref = (MVMNativeRef *)MVM_gc_allocate_object(tc, STABLE(type));
     });
-    MVM_ASSIGN_REF(tc, &(ref->common.header), ref->body.u.reg_or_lex.frame, f);
-    ref->body.u.reg_or_lex.var   = r;
-    ref->body.u.reg_or_lex.type  = reg_type;
+    MVM_ASSIGN_REF(tc, &(ref->common.header), ref->body.u.lex.frame, f);
+    ref->body.u.lex.var   = r;
+    ref->body.u.lex.type  = reg_type;
     return (MVMObject *)ref;
 }
-MVMObject * MVM_nativeref_reg_i(MVMThreadContext *tc, MVMRegister *r) {
-    MVMFrame *f = MVM_frame_force_to_heap(tc, tc->cur_frame);
-    MVMObject *ref_type = MVM_hll_current(tc)->int_lex_ref;
-    if (ref_type)
-        return reg_or_lex_ref(tc, ref_type, f, r, MVM_reg_int64);
-    MVM_exception_throw_adhoc(tc, "No int lexical reference type registered for current HLL");
-}
-MVMObject * MVM_nativeref_reg_n(MVMThreadContext *tc, MVMRegister *r) {
-    MVMFrame *f = MVM_frame_force_to_heap(tc, tc->cur_frame);
-    MVMObject *ref_type = MVM_hll_current(tc)->num_lex_ref;
-    if (ref_type)
-        return reg_or_lex_ref(tc, ref_type, f, r, MVM_reg_num64);
-    MVM_exception_throw_adhoc(tc, "No num lexical reference type registered for current HLL");
-}
-MVMObject * MVM_nativeref_reg_s(MVMThreadContext *tc, MVMRegister *r) {
-    MVMFrame *f = MVM_frame_force_to_heap(tc, tc->cur_frame);
-    MVMObject *ref_type = MVM_hll_current(tc)->str_lex_ref;
-    if (ref_type)
-        return reg_or_lex_ref(tc, ref_type, f, r, MVM_reg_str);
-    MVM_exception_throw_adhoc(tc, "No str lexical reference type registered for current HLL");
-}
 
 /* Creation of native references for lexicals. */
 static MVMFrame * get_lexical_outer(MVMThreadContext *tc, MVMuint16 outers) {
@@ -285,7 +264,7 @@ MVMObject * MVM_nativeref_lex_i(MVMThreadContext *tc, MVMuint16 outers, MVMuint1
                 type != MVM_reg_uint64 && type != MVM_reg_uint32 &&
                 type != MVM_reg_uint16 && type != MVM_reg_uint8)
             MVM_exception_throw_adhoc(tc, "getlexref_i: lexical is not an int");
-        return reg_or_lex_ref(tc, ref_type, f, &(f->env[idx]), type);
+        return lex_ref(tc, ref_type, f, &(f->env[idx]), type);
     }
     MVM_exception_throw_adhoc(tc, "No int lexical reference type registered for current HLL");
 }
@@ -301,7 +280,7 @@ MVMObject * MVM_nativeref_lex_n(MVMThreadContext *tc, MVMuint16 outers, MVMuint1
         MVMuint16 type = lexical_types[idx];
         if (type != MVM_reg_num64 && type != MVM_reg_num32)
             MVM_exception_throw_adhoc(tc, "getlexref_n: lexical is not a num");
-        return reg_or_lex_ref(tc, ref_type, f, &(f->env[idx]), type);
+        return lex_ref(tc, ref_type, f, &(f->env[idx]), type);
     }
     MVM_exception_throw_adhoc(tc, "No num lexical reference type registered for current HLL");
 }
@@ -316,7 +295,7 @@ MVMObject * MVM_nativeref_lex_s(MVMThreadContext *tc, MVMuint16 outers, MVMuint1
             : f->static_info->body.lexical_types;
         if (lexical_types[idx] != MVM_reg_str)
             MVM_exception_throw_adhoc(tc, "getlexref_s: lexical is not a str (%d, %d)", outers, idx);
-        return reg_or_lex_ref(tc, ref_type, f, &(f->env[idx]), MVM_reg_str);
+        return lex_ref(tc, ref_type, f, &(f->env[idx]), MVM_reg_str);
     }
     MVM_exception_throw_adhoc(tc, "No str lexical reference type registered for current HLL");
 }
@@ -329,7 +308,7 @@ static MVMObject * lexref_by_name(MVMThreadContext *tc, MVMObject *type, MVMStri
             MVM_HASH_GET(tc, lexical_names, name, entry)
             if (entry) {
                 if (cur_frame->static_info->body.lexical_types[entry->value] == kind) {
-                    return reg_or_lex_ref(tc, type, cur_frame, &cur_frame->env[entry->value], kind);
+                    return lex_ref(tc, type, cur_frame, &cur_frame->env[entry->value], kind);
                 }
                 else {
                     char *c_name = MVM_string_utf8_encode_C_string(tc, name);
@@ -479,31 +458,31 @@ MVMObject * MVM_nativeref_multidim_s(MVMThreadContext *tc, MVMObject *obj, MVMOb
  * they are more suited to calling from optimized code. The checking path is
  * in the native ref container implementation, in containers.c; after checks,
  * they delegate here. */
-MVMint64 MVM_nativeref_read_reg_or_lex_i(MVMThreadContext *tc, MVMObject *ref_obj) {
+MVMint64 MVM_nativeref_read_lex_i(MVMThreadContext *tc, MVMObject *ref_obj) {
     MVMNativeRef *ref = (MVMNativeRef *)ref_obj;
-    switch (ref->body.u.reg_or_lex.type) {
+    switch (ref->body.u.lex.type) {
         case MVM_reg_int8:
-            return ref->body.u.reg_or_lex.var->i8;
+            return ref->body.u.lex.var->i8;
         case MVM_reg_int16:
-            return ref->body.u.reg_or_lex.var->i16;
+            return ref->body.u.lex.var->i16;
         case MVM_reg_int32:
-            return ref->body.u.reg_or_lex.var->i32;
+            return ref->body.u.lex.var->i32;
         default:
-            return ref->body.u.reg_or_lex.var->i64;
+            return ref->body.u.lex.var->i64;
     }
 }
-MVMnum64 MVM_nativeref_read_reg_or_lex_n(MVMThreadContext *tc, MVMObject *ref_obj) {
+MVMnum64 MVM_nativeref_read_lex_n(MVMThreadContext *tc, MVMObject *ref_obj) {
     MVMNativeRef *ref = (MVMNativeRef *)ref_obj;
-    switch (ref->body.u.reg_or_lex.type) {
+    switch (ref->body.u.lex.type) {
         case MVM_reg_num32:
-            return ref->body.u.reg_or_lex.var->n32;
+            return ref->body.u.lex.var->n32;
         default:
-            return ref->body.u.reg_or_lex.var->n64;
+            return ref->body.u.lex.var->n64;
     }
 }
-MVMString * MVM_nativeref_read_reg_or_lex_s(MVMThreadContext *tc, MVMObject *ref_obj) {
+MVMString * MVM_nativeref_read_lex_s(MVMThreadContext *tc, MVMObject *ref_obj) {
     MVMNativeRef *ref = (MVMNativeRef *)ref_obj;
-    return ref->body.u.reg_or_lex.var->s;
+    return ref->body.u.lex.var->s;
 }
 MVMint64 MVM_nativeref_read_attribute_i(MVMThreadContext *tc, MVMObject *ref_obj) {
     MVMNativeRef *ref = (MVMNativeRef *)ref_obj;
@@ -546,38 +525,38 @@ MVMString * MVM_nativeref_read_multidim_s(MVMThreadContext *tc, MVMObject *ref_o
 }
 
 /* Reference write functions. Same (non-checking) rules as the reads above. */
-void MVM_nativeref_write_reg_or_lex_i(MVMThreadContext *tc, MVMObject *ref_obj, MVMint64 value) {
+void MVM_nativeref_write_lex_i(MVMThreadContext *tc, MVMObject *ref_obj, MVMint64 value) {
     MVMNativeRef *ref = (MVMNativeRef *)ref_obj;
-    switch (ref->body.u.reg_or_lex.type) {
+    switch (ref->body.u.lex.type) {
         case MVM_reg_int8:
-            ref->body.u.reg_or_lex.var->i8 = (MVMint8)value;
+            ref->body.u.lex.var->i8 = (MVMint8)value;
             break;
         case MVM_reg_int16:
-            ref->body.u.reg_or_lex.var->i16 = (MVMint16)value;
+            ref->body.u.lex.var->i16 = (MVMint16)value;
             break;
         case MVM_reg_int32:
-            ref->body.u.reg_or_lex.var->i32 = (MVMint32)value;
+            ref->body.u.lex.var->i32 = (MVMint32)value;
             break;
         default:
-            ref->body.u.reg_or_lex.var->i64 = value;
+            ref->body.u.lex.var->i64 = value;
             break;
     }
 }
-void MVM_nativeref_write_reg_or_lex_n(MVMThreadContext *tc, MVMObject *ref_obj, MVMnum64 value) {
+void MVM_nativeref_write_lex_n(MVMThreadContext *tc, MVMObject *ref_obj, MVMnum64 value) {
     MVMNativeRef *ref = (MVMNativeRef *)ref_obj;
-    switch (ref->body.u.reg_or_lex.type) {
+    switch (ref->body.u.lex.type) {
         case MVM_reg_num32:
-            ref->body.u.reg_or_lex.var->n32 = (MVMnum32)value;
+            ref->body.u.lex.var->n32 = (MVMnum32)value;
             break;
         default:
-            ref->body.u.reg_or_lex.var->n64 = value;
+            ref->body.u.lex.var->n64 = value;
             break;
     }
 }
-void MVM_nativeref_write_reg_or_lex_s(MVMThreadContext *tc, MVMObject *ref_obj, MVMString *value) {
+void MVM_nativeref_write_lex_s(MVMThreadContext *tc, MVMObject *ref_obj, MVMString *value) {
     MVMNativeRef *ref = (MVMNativeRef *)ref_obj;
-    MVM_ASSIGN_REF(tc, &(ref->body.u.reg_or_lex.frame->header),
-        ref->body.u.reg_or_lex.var->s, value);
+    MVM_ASSIGN_REF(tc, &(ref->body.u.lex.frame->header),
+        ref->body.u.lex.var->s, value);
 }
 void MVM_nativeref_write_attribute_i(MVMThreadContext *tc, MVMObject *ref_obj, MVMint64 value) {
     MVMNativeRef *ref = (MVMNativeRef *)ref_obj;
