@@ -58,10 +58,10 @@ static MVMuint8 in_handler_stack(MVMThreadContext *tc, MVMFrameHandler *fh, MVMF
 }
 
 /* Checks if a frame is still active. Naively, we could scan the call stack
- * for it, but we can use the same thing the GC uses to know if to scan the
- * work area. */
+ * for it, but since we always clean up ->work when a frame is removed from
+ * the call stack we can do it in O(1) that way. */
 static MVMuint8 in_caller_chain(MVMThreadContext *tc, MVMFrame *f_maybe) {
-    return f_maybe->tc ? 1 : 0;
+    return f_maybe->work ? 1 : 0;
 }
 
 
@@ -332,11 +332,11 @@ char * MVM_exception_backtrace_line(MVMThreadContext *tc, MVMFrame *cur_frame, M
         : NULL;
 
     char *filename_c = filename
-	? MVM_string_utf8_encode_C_string(tc, filename)
-	: "<ephemeral file>";
+        ? MVM_string_utf8_encode_C_string(tc, filename)
+        : "<ephemeral file>";
     char *name_c = name
-	? MVM_string_utf8_encode_C_string(tc, name)
-	: "<anonymous frame>";
+        ? MVM_string_utf8_encode_C_string(tc, name)
+        : "<anonymous frame>";
 
     snprintf(o, 1024, " %s %s:%u  (%s:%s)",
         not_top ? "from" : "  at",
@@ -346,9 +346,9 @@ char * MVM_exception_backtrace_line(MVMThreadContext *tc, MVMFrame *cur_frame, M
         name_c
     );
     if (filename)
-	MVM_free(filename_c);
+        MVM_free(filename_c);
     if (name)
-	MVM_free(name_c);
+        MVM_free(name_c);
 
     if (tmp1)
         MVM_free(tmp1);
@@ -850,7 +850,8 @@ void MVM_exception_throw_adhoc_free_va(MVMThreadContext *tc, char **waste, const
     MVMROOT(tc, ex, {
         char      *c_message = MVM_malloc(1024);
         int        bytes     = vsnprintf(c_message, 1024, messageFormat, args);
-        MVMString *message   = MVM_string_utf8_decode(tc, tc->instance->VMString, c_message, bytes);
+        int        to_encode = bytes > 1024 ? 1024 : bytes;
+        MVMString *message   = MVM_string_utf8_decode(tc, tc->instance->VMString, c_message, to_encode);
         MVM_free(c_message);
 
         /* Clean up after ourselves to avoid leaking C strings. */
