@@ -1,10 +1,13 @@
 #include "moar.h"
 
+/* This allows the dynlex cache to be disabled when bug hunting, if needed. */
+#define MVM_DYNLEX_CACHE_ENABLED 1
+
 /* Computes the initial work area for a frame or a specialization of a frame. */
 MVMRegister * MVM_frame_initial_work(MVMThreadContext *tc, MVMuint16 *local_types,
                                      MVMuint16 num_locals) {
     MVMuint16 i;
-    MVMRegister *work_initial = MVM_calloc(sizeof(MVMRegister), num_locals);
+    MVMRegister *work_initial = MVM_calloc(num_locals, sizeof(MVMRegister));
     for (i = 0; i < num_locals; i++)
         if (local_types[i] == MVM_reg_obj)
             work_initial[i].o = tc->instance->VMNull;
@@ -24,7 +27,7 @@ static void prepare_and_verify_static_frame(MVMThreadContext *tc, MVMStaticFrame
 
     /* Take compilation unit lock, to make sure we don't race to do the
      * frame preparation/verification work. */
-    MVM_reentrantmutex_lock(tc, (MVMReentrantMutex *)cu->body.update_mutex);
+    MVM_reentrantmutex_lock(tc, (MVMReentrantMutex *)cu->body.deserialize_frame_mutex);
     if (static_frame->body.instrumentation_level == 0) {
         /* Work size is number of locals/registers plus size of the maximum
         * call site argument list. */
@@ -61,7 +64,7 @@ static void prepare_and_verify_static_frame(MVMThreadContext *tc, MVMStaticFrame
     }
 
     /* Unlock, now we're finished. */
-    MVM_reentrantmutex_unlock(tc, (MVMReentrantMutex *)cu->body.update_mutex);
+    MVM_reentrantmutex_unlock(tc, (MVMReentrantMutex *)cu->body.deserialize_frame_mutex);
 }
 
 /* When we don't match the current instrumentation level, we hit this. It may
@@ -81,11 +84,15 @@ static void instrumentation_level_barrier(MVMThreadContext *tc, MVMStaticFrame *
         MVM_profile_instrument(tc, static_frame);
     else if (tc->instance->cross_thread_write_logging)
         MVM_cross_thread_write_instrument(tc, static_frame);
+    else if (tc->instance->coverage_logging)
+        MVM_line_coverage_instrument(tc, static_frame);
     else
         MVM_profile_ensure_uninstrumented(tc, static_frame);
 }
 
-/* Destroys a frame. */
+/* Called when the GC destroys a frame. Since the frame may have been alive as
+ * part of a continuation that was taken but never invoked, we should check
+ * things normally cleaned up on return don't need cleaning up also. */
 void MVM_frame_destroy(MVMThreadContext *tc, MVMFrame *frame) {
     if (frame->work) {
         MVM_args_proc_cleanup(tc, &frame->params);
@@ -114,15 +121,11 @@ static MVMFrame * create_context_only(MVMThreadContext *tc, MVMStaticFrame *stat
         frame = MVM_gc_allocate_frame(tc);
     });
     });
-    
-    /* Copy thread context into the frame. */
-    frame->tc = tc;
 
-    /* Set static frame. */
+    /* Set static frame, code ref, and handlers. */
     MVM_ASSIGN_REF(tc, &(frame->header), frame->static_info, static_frame);
-
-    /* Store the code ref. */
     MVM_ASSIGN_REF(tc, &(frame->header), frame->code_ref, code_ref);
+    frame->effective_handlers = static_frame->body.handlers;
 
     /* Allocate space for lexicals, copying the default lexical environment
      * into place and, if we're auto-closing, making sure anything we'd clone
@@ -132,22 +135,31 @@ static MVMFrame * create_context_only(MVMThreadContext *tc, MVMStaticFrame *stat
         frame->env = MVM_fixed_size_alloc(tc, tc->instance->fsa, static_frame->body.env_size);
         frame->allocd_env = static_frame->body.env_size;
         if (autoclose) {
-            MVMuint16 i;
-            for (i = 0; i < static_frame->body.num_lexicals; i++) {
+            MVMuint16 i, num_lexicals = static_frame->body.num_lexicals;
+
+            MVM_gc_root_temp_push(tc, (MVMCollectable **)&static_frame);
+
+            for (i = 0; i < num_lexicals; i++) {
                 if (!static_frame->body.static_env[i].o && static_frame->body.static_env_flags[i] == 1) {
                     MVMint32 scid, objid;
                     if (MVM_bytecode_find_static_lexical_scref(tc, static_frame->body.cu,
                             static_frame, i, &scid, &objid)) {
+                        MVMObject *resolved;
                         MVMSerializationContext *sc = MVM_sc_get_sc(tc, static_frame->body.cu, scid);
+
                         if (sc == NULL)
                             MVM_exception_throw_adhoc(tc,
                                 "SC not yet resolved; lookup failed");
+
+                        resolved = MVM_sc_get_object(tc, sc, objid);
+
                         MVM_ASSIGN_REF(tc, &(static_frame->common.header),
                             static_frame->body.static_env[i].o,
-                            MVM_sc_get_object(tc, sc, objid));
+                            resolved);
                     }
                 }
             }
+            MVM_gc_root_temp_pop(tc);
         }
         memcpy(frame->env, static_frame->body.static_env, static_frame->body.env_size);
     }
@@ -402,14 +414,14 @@ void MVM_frame_invoke(MVMThreadContext *tc, MVMStaticFrame *static_frame,
     if (spesh_cand >= 0 && spesh_cand < static_frame->body.num_spesh_candidates) {
         MVMSpeshCandidate *chosen_cand = &static_frame->body.spesh_candidates[spesh_cand];
         if (!chosen_cand->sg) {
-            MVMROOT(tc, static_frame, {
-            MVMROOT(tc, code_ref, {
-            MVMROOT(tc, outer, {
-                frame = allocate_frame(tc, static_frame, chosen_cand);
-            });
-            });
-            });
-            frame->effective_bytecode    = chosen_cand->bytecode;
+            frame = allocate_frame(tc, static_frame, chosen_cand);
+            if (chosen_cand->jitcode) {
+                frame->effective_bytecode = chosen_cand->jitcode->bytecode;
+                frame->jit_entry_label    = chosen_cand->jitcode->labels[0];
+            }
+            else {
+                frame->effective_bytecode = chosen_cand->bytecode;
+            }
             frame->effective_handlers    = chosen_cand->handlers;
             frame->effective_spesh_slots = chosen_cand->spesh_slots;
             frame->spesh_cand            = chosen_cand;
@@ -508,13 +520,7 @@ void MVM_frame_invoke(MVMThreadContext *tc, MVMStaticFrame *static_frame,
                 if (!chosen_cand->osr_logging && cur_idx < MVM_SPESH_LOG_RUNS) {
                     if (MVM_cas(&(chosen_cand->log_enter_idx), cur_idx, cur_idx + 1) == cur_idx) {
                         /* We get to log. */
-                        MVMROOT(tc, static_frame, {
-                        MVMROOT(tc, code_ref, {
-                        MVMROOT(tc, outer, {
-                            frame = allocate_frame(tc, static_frame, chosen_cand);
-                        });
-                        });
-                        });
+                        frame = allocate_frame(tc, static_frame, chosen_cand);
                         frame->effective_bytecode    = chosen_cand->bytecode;
                         frame->effective_handlers    = chosen_cand->handlers;
                         frame->effective_spesh_slots = chosen_cand->spesh_slots;
@@ -527,13 +533,7 @@ void MVM_frame_invoke(MVMThreadContext *tc, MVMStaticFrame *static_frame,
             }
             else {
                 /* In the post-specialize phase; can safely used the code. */
-                MVMROOT(tc, static_frame, {
-                MVMROOT(tc, code_ref, {
-                MVMROOT(tc, outer, {
-                    frame = allocate_frame(tc, static_frame, chosen_cand);
-                });
-                });
-                });
+                frame = allocate_frame(tc, static_frame, chosen_cand);
                 if (chosen_cand->jitcode) {
                     frame->effective_bytecode = chosen_cand->jitcode->bytecode;
                     frame->jit_entry_label    = chosen_cand->jitcode->labels[0];
@@ -550,20 +550,11 @@ void MVM_frame_invoke(MVMThreadContext *tc, MVMStaticFrame *static_frame,
         }
     }
     if (!found_spesh) {
-        MVMROOT(tc, static_frame, {
-        MVMROOT(tc, code_ref, {
-        MVMROOT(tc, outer, {
-            frame = allocate_frame(tc, static_frame, NULL);
-        });
-        });
-        });
+        frame = allocate_frame(tc, static_frame, NULL);
         frame->effective_bytecode = static_frame->body.bytecode;
         frame->effective_handlers = static_frame->body.handlers;
     }
 
-    /* Copy thread context into the frame. */
-    frame->tc = tc;
-
     /* Set static frame. */
     frame->static_info = static_frame;
 
@@ -620,8 +611,7 @@ void MVM_frame_invoke(MVMThreadContext *tc, MVMStaticFrame *static_frame,
                         }
                         else {
                             /* Allocate storage for state vars. */
-                            state = MVM_malloc(frame->static_info->body.env_size);
-                            memset(state, 0, frame->static_info->body.env_size);
+                            state = (MVMRegister *)MVM_calloc(1, frame->static_info->body.env_size);
                             ((MVMCode *)frame->code_ref)->body.state_vars = state;
                             state_act = 1;
 
@@ -754,7 +744,6 @@ MVMFrame * MVM_frame_create_for_deopt(MVMThreadContext *tc, MVMStaticFrame *stat
     });
     frame->effective_bytecode       = static_frame->body.bytecode;
     frame->effective_handlers       = static_frame->body.handlers;
-    frame->tc                       = tc;
     MVM_ASSIGN_REF(tc, &(frame->header), frame->static_info, static_frame);
     MVM_ASSIGN_REF(tc, &(frame->header), frame->code_ref, code_ref);
     MVM_ASSIGN_REF(tc, &(frame->header), frame->outer, code_ref->body.outer);
@@ -785,31 +774,15 @@ static MVMuint64 remove_one_frame(MVMThreadContext *tc, MVMuint8 unwind) {
         }
     }
 
-    /* Some cleanup we only need do if we're not a frame involved in a
-     * continuation (otherwise we need to allow for multi-shot
-     * re-invocation). */
-    if (!returner->in_continuation) {
-        /* Arguments buffer no longer in use (saves GC visiting it). */
-        returner->cur_args_callsite = NULL;
+    /* Clear up any continuation tags. */
+    if (returner->continuation_tags)
+        MVM_continuation_free_tags(tc, returner);
 
-        /* Clear up argument processing leftovers, if any. */
-        if (returner->work)
-            MVM_args_proc_cleanup_for_cache(tc, &returner->params);
-
-        /* Clear up any continuation tags. */
-        if (returner->continuation_tags)
-            MVM_continuation_free_tags(tc, returner);
-
-        /* Signal to the GC to ignore ->work */
-        returner->tc = NULL;
-
-        /* Unless we need to keep the caller chain in place, clear it up. */
-        if (caller) {
-            if (!returner->keep_caller)
-                returner->caller = NULL;
-            else if (unwind)
-                caller->keep_caller = 1;
-        }
+    /* Clean up frame working space. */
+    if (returner->work) {
+        MVM_args_proc_cleanup(tc, &returner->params);
+        MVM_fixed_size_free(tc, tc->instance->fsa, returner->allocd_work,
+            returner->work);
     }
 
     /* If it's a call stack frame, remove it from the stack. */
@@ -818,7 +791,17 @@ static MVMuint64 remove_one_frame(MVMThreadContext *tc, MVMuint8 unwind) {
         stack->alloc = (char *)returner;
         if ((char *)stack->alloc - sizeof(MVMCallStackRegion) == (char *)stack)
             MVM_callstack_region_prev(tc);
-        MVM_frame_destroy(tc, returner);
+        if (returner->env)
+            MVM_fixed_size_free(tc, tc->instance->fsa, returner->allocd_env, returner->env);
+    }
+
+    /* Otherwise, NULL  out ->work, to indicate the frame is no longer in
+     * dynamic scope. This is used by the GC to avoid marking stuff (this is
+     * needed for safety as otherwise we'd read freed memory), as well as by
+     * exceptions to ensure the target of an exception throw is indeed still
+     * in dynamic scope. */
+    else {
+        returner->work = NULL;
     }
 
     /* Switch back to the caller frame if there is one. */
@@ -879,21 +862,25 @@ MVMuint64 MVM_frame_try_return(MVMThreadContext *tc) {
         if (tc->cur_frame == tc->thread_entry_frame)
             MVM_exception_throw_adhoc(tc, "Thread entry point frame cannot have an exit handler");
 
-        switch (caller->return_type) {
-            case MVM_RETURN_OBJ:
-                result = caller->return_value->o;
-                break;
-            case MVM_RETURN_INT:
-                result = MVM_repr_box_int(tc, hll->int_box_type, caller->return_value->i64);
-                break;
-            case MVM_RETURN_NUM:
-                result = MVM_repr_box_num(tc, hll->num_box_type, caller->return_value->n64);
-                break;
-            case MVM_RETURN_STR:
-                result = MVM_repr_box_str(tc, hll->str_box_type, caller->return_value->s);
-                break;
-            default:
-                result = NULL;
+        if (caller->return_type == MVM_RETURN_OBJ) {
+            result = caller->return_value->o;
+        }
+        else {
+            MVMROOT(tc, cur_frame, {
+                switch (caller->return_type) {
+                    case MVM_RETURN_INT:
+                        result = MVM_repr_box_int(tc, hll->int_box_type, caller->return_value->i64);
+                        break;
+                    case MVM_RETURN_NUM:
+                        result = MVM_repr_box_num(tc, hll->num_box_type, caller->return_value->n64);
+                        break;
+                    case MVM_RETURN_STR:
+                        result = MVM_repr_box_str(tc, hll->str_box_type, caller->return_value->s);
+                        break;
+                    default:
+                        result = NULL;
+                }
+            });
         }
 
         handler = MVM_frame_find_invokee(tc, hll->exit_handler, NULL);
@@ -1010,16 +997,23 @@ void MVM_frame_unwind_to(MVMThreadContext *tc, MVMFrame *frame, MVMuint8 *abs_ad
 MVMObject * MVM_frame_get_code_object(MVMThreadContext *tc, MVMCode *code) {
     if (REPR(code)->ID != MVM_REPR_ID_MVMCode)
         MVM_exception_throw_adhoc(tc, "getcodeobj needs a code ref");
+
     if (!code->body.code_object) {
         MVMStaticFrame *sf = code->body.sf;
         if (sf->body.code_obj_sc_dep_idx > 0) {
+            MVMObject *resolved;
             MVMSerializationContext *sc = MVM_sc_get_sc(tc, sf->body.cu,
                 sf->body.code_obj_sc_dep_idx - 1);
             if (sc == NULL)
                 MVM_exception_throw_adhoc(tc,
                     "SC not yet resolved; lookup failed");
+
+            MVMROOT(tc, code, {
+                resolved = MVM_sc_get_object(tc, sc, sf->body.code_obj_sc_idx);
+            });
+
             MVM_ASSIGN_REF(tc, &(code->common.header), code->body.code_object,
-                MVM_sc_get_object(tc, sc, sf->body.code_obj_sc_idx));
+                resolved);
         }
     }
     return code->body.code_object;
@@ -1038,6 +1032,36 @@ void MVM_frame_capturelex(MVMThreadContext *tc, MVMObject *code) {
     MVM_ASSIGN_REF(tc, &(code->header), code_obj->body.outer, captured);
 }
 
+/* This is used for situations in Perl 6 like:
+ * supply {
+ *     my $x = something();
+ *     whenever $supply {
+ *         QUIT { $x.foo() }
+ *     }
+ * }
+ * Here, the QUIT may be called without an invocation of the whenever ever
+ * having taken place. At the point we closure-clone the whenever block, we
+ * will capture_inner the QUIT phaser. This creates a fake outer for the
+ * QUIT, but makes *its* outer point to the nearest instance of the relevant
+ * static frame on the call stack, so that the QUIT will disocver the correct
+ * $x.
+ */
+void MVM_frame_capture_inner(MVMThreadContext *tc, MVMObject *code) {
+    MVMCode *code_obj = (MVMCode *)code;
+    MVMFrame *outer;
+    MVMROOT(tc, code, {
+        MVMStaticFrame *sf_outer = code_obj->body.sf->body.outer;
+        MVMROOT(tc, sf_outer, {
+            outer = create_context_only(tc, sf_outer, (MVMObject *)sf_outer->body.static_code, 1);
+        });
+        MVMROOT(tc, outer, {
+            MVMFrame *outer_outer = autoclose(tc, sf_outer->body.outer);
+            MVM_ASSIGN_REF(tc, &(outer->header), outer->outer, outer_outer);
+        });
+    });
+    MVM_ASSIGN_REF(tc, &(code->header), code_obj->body.outer, outer);
+}
+
 /* Given the specified code object, copies it and returns a copy which
  * captures a closure over the current scope. */
 MVMObject * MVM_frame_takeclosure(MVMThreadContext *tc, MVMObject *code) {
@@ -1070,7 +1094,7 @@ MVMObject * MVM_frame_vivify_lexical(MVMThreadContext *tc, MVMFrame *f, MVMuint1
     MVMuint8       *flags;
     MVMint16        flag;
     MVMRegister    *static_env;
-    MVMuint16       effective_idx;
+    MVMuint16       effective_idx = 0;
     MVMStaticFrame *effective_sf;
     if (idx < f->static_info->body.num_lexicals) {
         flags         = f->static_info->body.static_env_flags;
@@ -1097,17 +1121,20 @@ MVMObject * MVM_frame_vivify_lexical(MVMThreadContext *tc, MVMFrame *f, MVMuint1
     }
     flag  = flags ? flags[effective_idx] : -1;
     if (flag != -1 && static_env[effective_idx].o == NULL) {
-        MVMStaticFrameBody *static_frame_body = &(f->static_info->body);
         MVMint32 scid, objid;
         if (MVM_bytecode_find_static_lexical_scref(tc, effective_sf->body.cu,
                 effective_sf, effective_idx, &scid, &objid)) {
             MVMSerializationContext *sc = MVM_sc_get_sc(tc, effective_sf->body.cu, scid);
+            MVMObject *resolved;
             if (sc == NULL)
                 MVM_exception_throw_adhoc(tc,
                     "SC not yet resolved; lookup failed");
+            MVMROOT(tc, f, {
+                resolved = MVM_sc_get_object(tc, sc, objid);
+            });
             MVM_ASSIGN_REF(tc, &(f->static_info->common.header),
-                static_frame_body->static_env[effective_idx].o,
-                MVM_sc_get_object(tc, sc, objid));
+                f->static_info->body.static_env[effective_idx].o,
+                resolved);
         }
     }
     if (flag == 0) {
@@ -1136,15 +1163,12 @@ MVMObject * MVM_frame_vivify_lexical(MVMThreadContext *tc, MVMFrame *f, MVMuint1
  * if it does not exist. Incorrect type always throws. */
 MVMRegister * MVM_frame_find_lexical_by_name(MVMThreadContext *tc, MVMString *name, MVMuint16 type) {
     MVMFrame *cur_frame = tc->cur_frame;
-    MVM_string_flatten(tc, name);
     while (cur_frame != NULL) {
         MVMLexicalRegistry *lexical_names = cur_frame->static_info->body.lexical_names;
         if (lexical_names) {
             /* Indexes were formerly stored off-by-one to avoid semi-predicate issue. */
             MVMLexicalRegistry *entry;
-
             MVM_HASH_GET(tc, lexical_names, name, entry)
-
             if (entry) {
                 if (cur_frame->static_info->body.lexical_types[entry->value] == type) {
                     MVMRegister *result = &cur_frame->env[entry->value];
@@ -1176,7 +1200,6 @@ MVMRegister * MVM_frame_find_lexical_by_name(MVMThreadContext *tc, MVMString *na
  * chain. */
 MVM_PUBLIC void MVM_frame_bind_lexical_by_name(MVMThreadContext *tc, MVMString *name, MVMuint16 type, MVMRegister *value) {
     MVMFrame *cur_frame = tc->cur_frame;
-    MVM_string_flatten(tc, name);
     while (cur_frame != NULL) {
         MVMLexicalRegistry *lexical_names = cur_frame->static_info->body.lexical_names;
         if (lexical_names) {
@@ -1228,15 +1251,12 @@ MVMObject * MVM_frame_find_lexical_by_name_outer(MVMThreadContext *tc, MVMString
 /* Looks up the address of the lexical with the specified name, starting with
  * the specified frame. Only works if it's an object lexical.  */
 MVMRegister * MVM_frame_find_lexical_by_name_rel(MVMThreadContext *tc, MVMString *name, MVMFrame *cur_frame) {
-    MVM_string_flatten(tc, name);
     while (cur_frame != NULL) {
         MVMLexicalRegistry *lexical_names = cur_frame->static_info->body.lexical_names;
         if (lexical_names) {
             /* Indexes were formerly stored off-by-one to avoid semi-predicate issue. */
             MVMLexicalRegistry *entry;
-
             MVM_HASH_GET(tc, lexical_names, name, entry)
-
             if (entry) {
                 if (cur_frame->static_info->body.lexical_types[entry->value] == MVM_reg_obj) {
                     MVMRegister *result = &cur_frame->env[entry->value];
@@ -1261,7 +1281,6 @@ MVMRegister * MVM_frame_find_lexical_by_name_rel(MVMThreadContext *tc, MVMString
 /* Looks up the address of the lexical with the specified name, starting with
  * the specified frame. It checks all outer frames of the caller frame chain.  */
 MVMRegister * MVM_frame_find_lexical_by_name_rel_caller(MVMThreadContext *tc, MVMString *name, MVMFrame *cur_caller_frame) {
-    MVM_string_flatten(tc, name);
     while (cur_caller_frame != NULL) {
         MVMFrame *cur_frame = cur_caller_frame;
         while (cur_frame != NULL) {
@@ -1269,9 +1288,7 @@ MVMRegister * MVM_frame_find_lexical_by_name_rel_caller(MVMThreadContext *tc, MV
             if (lexical_names) {
                 /* Indexes were formerly stored off-by-one to avoid semi-predicate issue. */
                 MVMLexicalRegistry *entry;
-
                 MVM_HASH_GET(tc, lexical_names, name, entry)
-
                 if (entry) {
                     if (cur_frame->static_info->body.lexical_types[entry->value] == MVM_reg_obj) {
                         MVMRegister *result = &cur_frame->env[entry->value];
@@ -1298,6 +1315,7 @@ MVMRegister * MVM_frame_find_lexical_by_name_rel_caller(MVMThreadContext *tc, MV
 /* Looks up the address of the lexical with the specified name and the
  * specified type. Returns null if it does not exist. */
 static void try_cache_dynlex(MVMThreadContext *tc, MVMFrame *from, MVMFrame *to, MVMString *name, MVMRegister *reg, MVMuint16 type, MVMuint32 fcost, MVMuint32 icost) {
+#if MVM_DYNLEX_CACHE_ENABLED
     MVMint32 next = 0;
     MVMint32 frames = 0;
     MVMuint32 desperation = 0;
@@ -1324,6 +1342,7 @@ static void try_cache_dynlex(MVMThreadContext *tc, MVMFrame *from, MVMFrame *to,
         }
         from = from->caller;
     }
+#endif
 }
 MVMRegister * MVM_frame_find_contextual_by_name(MVMThreadContext *tc, MVMString *name, MVMuint16 *type, MVMFrame *cur_frame, MVMint32 vivify, MVMFrame **found_frame) {
     FILE *dlog = tc->instance->dynvar_log_fh;
@@ -1344,7 +1363,6 @@ MVMRegister * MVM_frame_find_contextual_by_name(MVMThreadContext *tc, MVMString
         last_time = tc->instance->dynvar_log_lasttime;
     }
 
-    MVM_string_flatten(tc, name);
     while (cur_frame != NULL) {
         MVMLexicalRegistry *lexical_names;
         MVMSpeshCandidate  *cand     = cur_frame->spesh_cand;
@@ -1383,7 +1401,7 @@ MVMRegister * MVM_frame_find_contextual_by_name(MVMThreadContext *tc, MVMString
                                 if (fcost+icost > 1)
                                   try_cache_dynlex(tc, initial_frame, cur_frame, name, result, *type, fcost, icost);
                                 if (dlog) {
-                                    fprintf(dlog, "I %s %d %d %d %d %llu %llu %llu\n", c_name, fcost, icost, ecost, xcost, last_time, start_time, uv_hrtime());
+                                    fprintf(dlog, "I %s %d %d %d %d %"PRIu64" %"PRIu64" %"PRIu64"\n", c_name, fcost, icost, ecost, xcost, last_time, start_time, uv_hrtime());
                                     fflush(dlog);
                                     MVM_free(c_name);
                                     tc->instance->dynvar_log_lasttime = uv_hrtime();
@@ -1420,7 +1438,7 @@ MVMRegister * MVM_frame_find_contextual_by_name(MVMThreadContext *tc, MVMString
                                 if (fcost+icost > 1)
                                   try_cache_dynlex(tc, initial_frame, cur_frame, name, result, *type, fcost, icost);
                                 if (dlog) {
-                                    fprintf(dlog, "I %s %d %d %d %d %llu %llu %llu\n", c_name, fcost, icost, ecost, xcost, last_time, start_time, uv_hrtime());
+                                    fprintf(dlog, "I %s %d %d %d %d %"PRIu64" %"PRIu64" %"PRIu64"\n", c_name, fcost, icost, ecost, xcost, last_time, start_time, uv_hrtime());
                                     fflush(dlog);
                                     MVM_free(c_name);
                                     tc->instance->dynvar_log_lasttime = uv_hrtime();
@@ -1442,7 +1460,7 @@ MVMRegister * MVM_frame_find_contextual_by_name(MVMThreadContext *tc, MVMString
                 if (fcost+icost > 5)
                     try_cache_dynlex(tc, initial_frame, cur_frame, name, result, *type, fcost, icost);
                 if (dlog) {
-                    fprintf(dlog, "C %s %d %d %d %d %llu %llu %llu\n", c_name, fcost, icost, ecost, xcost, last_time, start_time, uv_hrtime());
+                    fprintf(dlog, "C %s %d %d %d %d %"PRIu64" %"PRIu64" %"PRIu64"\n", c_name, fcost, icost, ecost, xcost, last_time, start_time, uv_hrtime());
                     fflush(dlog);
                     MVM_free(c_name);
                     tc->instance->dynvar_log_lasttime = uv_hrtime();
@@ -1473,7 +1491,7 @@ MVMRegister * MVM_frame_find_contextual_by_name(MVMThreadContext *tc, MVMString
                     });
                 }
                 if (dlog) {
-                    fprintf(dlog, "F %s %d %d %d %d %llu %llu %llu\n", c_name, fcost, icost, ecost, xcost, last_time, start_time, uv_hrtime());
+                    fprintf(dlog, "F %s %d %d %d %d %"PRIu64" %"PRIu64" %"PRIu64"\n", c_name, fcost, icost, ecost, xcost, last_time, start_time, uv_hrtime());
                     fflush(dlog);
                     MVM_free(c_name);
                     tc->instance->dynvar_log_lasttime = uv_hrtime();
@@ -1488,7 +1506,7 @@ MVMRegister * MVM_frame_find_contextual_by_name(MVMThreadContext *tc, MVMString
         cur_frame = cur_frame->caller;
     }
     if (dlog) {
-        fprintf(dlog, "N %s %d %d %d %d %llu %llu %llu\n", c_name, fcost, icost, ecost, xcost, last_time, start_time, uv_hrtime());
+        fprintf(dlog, "N %s %d %d %d %d %"PRIu64" %"PRIu64" %"PRIu64"\n", c_name, fcost, icost, ecost, xcost, last_time, start_time, uv_hrtime());
         fflush(dlog);
         MVM_free(c_name);
         tc->instance->dynvar_log_lasttime = uv_hrtime();
@@ -1587,7 +1605,6 @@ MVMRegister * MVM_frame_lexical(MVMThreadContext *tc, MVMFrame *f, MVMString *na
     MVMLexicalRegistry *lexical_names = f->static_info->body.lexical_names;
     if (lexical_names) {
         MVMLexicalRegistry *entry;
-        MVM_string_flatten(tc, name);
         MVM_HASH_GET(tc, lexical_names, name, entry)
         if (entry)
             return &f->env[entry->value];
@@ -1606,7 +1623,6 @@ MVMRegister * MVM_frame_try_get_lexical(MVMThreadContext *tc, MVMFrame *f, MVMSt
     MVMLexicalRegistry *lexical_names = f->static_info->body.lexical_names;
     if (lexical_names) {
         MVMLexicalRegistry *entry;
-        MVM_string_flatten(tc, name);
         MVM_HASH_GET(tc, lexical_names, name, entry)
         if (entry && f->static_info->body.lexical_types[entry->value] == type) {
             MVMRegister *result = &f->env[entry->value];
@@ -1623,7 +1639,6 @@ MVMuint16 MVM_frame_lexical_primspec(MVMThreadContext *tc, MVMFrame *f, MVMStrin
     MVMLexicalRegistry *lexical_names = f->static_info->body.lexical_names;
     if (lexical_names) {
         MVMLexicalRegistry *entry;
-        MVM_string_flatten(tc, name);
         MVM_HASH_GET(tc, lexical_names, name, entry)
         if (entry) {
             switch (f->static_info->body.lexical_types[entry->value]) {
@@ -1674,7 +1689,7 @@ static MVMObject * find_invokee_internal(MVMThreadContext *tc, MVMObject *code,
                 *tweak_cs = orig->with_invocant;
             }
             else {
-                MVMCallsite *new   = MVM_malloc(sizeof(MVMCallsite));
+                MVMCallsite *new   = MVM_calloc(1, sizeof(MVMCallsite));
                 MVMint32     fsize = orig->flag_count;
                 new->flag_count    = fsize + 1;
                 new->arg_flags     = MVM_malloc(new->flag_count * sizeof(MVMCallsiteEntry));
@@ -1760,35 +1775,5 @@ MVMObject * MVM_frame_context_wrapper(MVMThreadContext *tc, MVMFrame *f) {
         ctx = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTContext);
         MVM_ASSIGN_REF(tc, &(ctx->header), ((MVMContext *)ctx)->body.context, f);
     });
-    f->keep_caller = 1;
     return ctx;
 }
-
-/* Creates a shallow clone of a frame. Used by continuations. We rely on this
- * not allocating with the GC; update continuation clone code if it comes to
-  * do so. */
-MVMFrame * MVM_frame_clone(MVMThreadContext *tc, MVMFrame *f) {
-    /* First, just grab a copy of everything. */
-    MVMFrame *clone;
-    MVMROOT(tc, f, {
-        clone = MVM_gc_allocate_frame(tc);
-    });
-    memcpy(
-        (char *)clone + sizeof(MVMCollectable),
-        (char *)f + sizeof(MVMCollectable),
-        sizeof(MVMFrame) - sizeof(MVMCollectable));
-
-    /* Need fresh env and work. */
-    if (f->static_info->body.env_size) {
-        clone->env = MVM_fixed_size_alloc(tc, tc->instance->fsa, f->static_info->body.env_size);
-        clone->allocd_env = f->static_info->body.env_size;
-        memcpy(clone->env, f->env, f->static_info->body.env_size);
-    }
-    if (f->static_info->body.work_size) {
-        clone->work = MVM_malloc(f->static_info->body.work_size);
-        memcpy(clone->work, f->work, f->static_info->body.work_size);
-        clone->args = clone->work + f->static_info->body.num_locals;
-    }
-
-    return clone;
-}
