@@ -13,7 +13,7 @@ MVMHLLConfig *MVM_hll_get_config_for(MVMThreadContext *tc, MVMString *name) {
     }
 
     if (!entry) {
-        entry = MVM_calloc(sizeof(MVMHLLConfig), 1);
+        entry = MVM_calloc(1, sizeof(MVMHLLConfig));
         entry->name = name;
         entry->int_box_type = tc->instance->boot_types.BOOTInt;
         entry->num_box_type = tc->instance->boot_types.BOOTNum;
@@ -117,11 +117,11 @@ MVMObject * MVM_hll_set_config(MVMThreadContext *tc, MVMString *name, MVMObject
             check_config_key(tc, config_hash, "method_not_found_error", method_not_found_error, config);
             check_config_key(tc, config_hash, "lexical_handler_not_found_error", lexical_handler_not_found_error, config);
             check_config_key_reftype(tc, config_hash, "int_lex_ref", int_lex_ref,
-                config, MVM_STORAGE_SPEC_BP_INT, MVM_NATIVEREF_REG_OR_LEX);
+                config, MVM_STORAGE_SPEC_BP_INT, MVM_NATIVEREF_LEX);
             check_config_key_reftype(tc, config_hash, "num_lex_ref", num_lex_ref,
-                config, MVM_STORAGE_SPEC_BP_NUM, MVM_NATIVEREF_REG_OR_LEX);
+                config, MVM_STORAGE_SPEC_BP_NUM, MVM_NATIVEREF_LEX);
             check_config_key_reftype(tc, config_hash, "str_lex_ref", str_lex_ref,
-                config, MVM_STORAGE_SPEC_BP_STR, MVM_NATIVEREF_REG_OR_LEX);
+                config, MVM_STORAGE_SPEC_BP_STR, MVM_NATIVEREF_LEX);
             check_config_key_reftype(tc, config_hash, "int_attr_ref", int_attr_ref,
                 config, MVM_STORAGE_SPEC_BP_INT, MVM_NATIVEREF_ATTRIBUTE);
             check_config_key_reftype(tc, config_hash, "num_attr_ref", num_attr_ref,
