@@ -62,12 +62,14 @@ struct MVMStringConsts {
     MVMString *auto_viv_container;
     MVMString *done;
     MVMString *error;
-    MVMString *stdout_chars;
     MVMString *stdout_bytes;
-    MVMString *stderr_chars;
     MVMString *stderr_bytes;
+    MVMString *merge_bytes;
     MVMString *buf_type;
     MVMString *write;
+    MVMString *stdin_fd;
+    MVMString *stdout_fd;
+    MVMString *stderr_fd;
     MVMString *nativeref;
     MVMString *refkind;
     MVMString *positional;
@@ -79,6 +81,8 @@ struct MVMStringConsts {
     MVMString *kind;
     MVMString *instrumented;
     MVMString *heap;
+    MVMString *translate_newlines;
+    MVMString *platform_newline;
 };
 
 /* An entry in the representations registry. */
@@ -365,6 +369,10 @@ struct MVMInstance {
     MVMuint32  cross_thread_write_logging_include_locked;
     uv_mutex_t mutex_cross_thread_write_logging;
 
+    MVMuint32  coverage_logging;
+    /* Log file for coverage logging. */
+    FILE *coverage_log_fh;
+
     /* Cached backend config hash. */
     MVMObject *cached_backend_config;
 };
