@@ -337,7 +337,7 @@ void * MVM_nativecall_unmarshal_carray(MVMThreadContext *tc, MVMObject *value) {
 void * MVM_nativecall_unmarshal_vmarray(MVMThreadContext *tc, MVMObject *value) {
     if (!IS_CONCRETE(value))
         return NULL;
-    else if (REPR(value)->ID == MVM_REPR_ID_MVMArray) {
+    else if (REPR(value)->ID == MVM_REPR_ID_VMArray) {
         MVMArrayBody *body          = &((MVMArray *)value)->body;
         MVMArrayREPRData *repr_data = (MVMArrayREPRData *)STABLE(value)->REPR_data;
         size_t start_pos            = body->start * repr_data->elem_size;
@@ -375,8 +375,11 @@ void MVM_nativecall_build(MVMThreadContext *tc, MVMObject *site, MVMString *lib,
         MVMString *sym, MVMString *conv, MVMObject *arg_info, MVMObject *ret_info) {
     char *lib_name = MVM_string_utf8_c8_encode_C_string(tc, lib);
     char *sym_name = MVM_string_utf8_c8_encode_C_string(tc, sym);
+    MVMint8  keep_sym_name = 0;
     MVMint16 i;
 
+    unsigned int interval_id = MVM_telemetry_interval_start(tc, "building native call");
+
     MVMObject *entry_point_o = (MVMObject *)MVM_repr_at_key_o(tc, ret_info,
         tc->instance->str_consts.entry_point);
 
@@ -390,6 +393,7 @@ void MVM_nativecall_build(MVMThreadContext *tc, MVMObject *site, MVMString *lib,
     if (!body->lib_handle) {
         char *waste[] = { lib_name, NULL };
         MVM_free(sym_name);
+        MVM_telemetry_interval_stop(tc, interval_id, "error building native call");
         MVM_exception_throw_adhoc_free(tc, waste, "Cannot locate native library '%s': %s", lib_name, dlerror());
     }
 
@@ -397,16 +401,25 @@ void MVM_nativecall_build(MVMThreadContext *tc, MVMObject *site, MVMString *lib,
     if (entry_point_o) {
         body->entry_point = MVM_nativecall_unmarshal_cpointer(tc, entry_point_o);
         body->sym_name    = sym_name;
+        keep_sym_name     = 1;
     }
 
     if (!body->entry_point) {
         body->entry_point = MVM_nativecall_find_sym(body->lib_handle, sym_name);
         if (!body->entry_point) {
             char *waste[] = { sym_name, lib_name, NULL };
+            MVM_telemetry_interval_stop(tc, interval_id, "error building native call");
             MVM_exception_throw_adhoc_free(tc, waste, "Cannot locate symbol '%s' in native library '%s'",
                 sym_name, lib_name);
         }
         body->sym_name = sym_name;
+        keep_sym_name     = 1;
+    }
+
+    MVM_telemetry_interval_annotate_dynamic((uintptr_t)body->entry_point, interval_id, body->sym_name);
+
+    if (keep_sym_name == 0) {
+        MVM_free(sym_name);
     }
 
     /* Set calling convention, if any. */
@@ -439,6 +452,8 @@ void MVM_nativecall_build(MVMThreadContext *tc, MVMObject *site, MVMString *lib,
 #ifdef HAVE_LIBFFI
     body->ffi_ret_type = MVM_nativecall_get_ffi_type(tc, body->ret_type);
 #endif
+
+    MVM_telemetry_interval_stop(tc, interval_id, "nativecall built");
 }
 
 static MVMObject * nativecall_cast(MVMThreadContext *tc, MVMObject *target_spec, MVMObject *target_type, void *cpointer_body) {
@@ -639,7 +654,7 @@ MVMObject * MVM_nativecall_cast(MVMThreadContext *tc, MVMObject *target_spec, MV
         data_body = MVM_nativecall_unmarshal_cpointer(tc, source);
     else if (REPR(source)->ID == MVM_REPR_ID_MVMCArray)
         data_body = MVM_nativecall_unmarshal_carray(tc, source);
-    else if (REPR(source)->ID == MVM_REPR_ID_MVMArray)
+    else if (REPR(source)->ID == MVM_REPR_ID_VMArray)
         data_body = MVM_nativecall_unmarshal_vmarray(tc, source);
     else
         MVM_exception_throw_adhoc(tc,
