@@ -22,16 +22,6 @@ typedef enum {
     MVMGCStatus_STOLEN = 3
 } MVMGCStatus;
 
-/* To manage memory more efficiently, we cache MVMFrame instances.
- * The initial frame pool table size sets the initial guess at the
- * number of different types of frame (that is, an MVMStaticFrame)
- * that we'll encounter and cache. If we do deep recursion, we run
- * the risk of caching an enormous number of frames, so the length
- * limit sets how many frames of a given static frame type we will
- * keep around. */
-#define MVMInitialFramePoolTableSize    64
-#define MVMFramePoolLengthLimit         64
-
 /* Information associated with an executing thread. */
 struct MVMThreadContext {
     /* The current allocation pointer, where the next object to be allocated
@@ -54,6 +44,9 @@ struct MVMThreadContext {
     /* Thread object representing the thread. */
     MVMThread *thread_obj;
 
+    /* Per-thread fixed size allocator state. */
+    MVMFixedSizeAllocThread *thread_fsa;
+
     /* Pointer to where the interpreter's current opcode is stored. */
     MVMuint8 **interp_cur_op;
 
@@ -95,6 +88,9 @@ struct MVMThreadContext {
     /* Result object of the last-run exception handler. */
     MVMObject *last_handler_result;
 
+    /* Last payload made available in a payload-goto exception handler. */
+    MVMObject *last_payload;
+
     /* Mutex that must be released if we throw an exception. Used in places
      * like I/O, which grab a mutex but may throw an exception. */
     uv_mutex_t *ex_release_mutex;
@@ -174,8 +170,10 @@ struct MVMThreadContext {
      * index 0. */
     MVMObject     *compiling_scs;
 
-    /* Dispatcher set for next invocation to take. */
+    /* Dispatcher for next invocation that matches _for to take. If _for is
+     * NULL then anything matches. */
     MVMObject     *cur_dispatcher;
+    MVMObject     *cur_dispatcher_for;
 
     /* Cache of native code callback data. */
     MVMNativeCallbackCacheHead *native_callback_cache;
@@ -212,9 +210,15 @@ struct MVMThreadContext {
      * in the call stack */
     MVMint32 current_frame_nr;
     MVMint32 next_frame_nr;
+
+#if MVM_GC_DEBUG
+    /* Whether we are currently in the specializer. Used to catch GC runs that
+     * take place at times they never should. */
+    MVMint32 in_spesh;
+#endif
 };
 
-MVMThreadContext * MVM_tc_create(MVMInstance *instance);
+MVMThreadContext * MVM_tc_create(MVMThreadContext *parent, MVMInstance *instance);
 void MVM_tc_destroy(MVMThreadContext *tc);
 void MVM_tc_set_ex_release_mutex(MVMThreadContext *tc, uv_mutex_t *mutex);
 void MVM_tc_release_ex_release_mutex(MVMThreadContext *tc);
