@@ -215,6 +215,24 @@ static void finish_gc(MVMThreadContext *tc, MVMuint8 gen, MVMuint8 is_coordinato
             MVM_store(&thread_obj->body.stage, MVM_thread_stage_destroyed);
         }
         else {
+            /* Free gen2 unmarked if full collection. */
+            if (gen == MVMGCGenerations_Both) {
+                GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE,
+                    "Thread %d run %d : freeing gen2 of thread %d\n",
+                    other->thread_id);
+                MVM_gc_collect_free_gen2_unmarked(other, 0);
+            }
+
+            /* Contribute this thread's promoted bytes. */
+            MVM_add(&tc->instance->gc_promoted_bytes_since_last_full, other->gc_promoted_bytes);
+
+            /* Collect nursery. */
+            GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE,
+                "Thread %d run %d : collecting nursery uncopied of thread %d\n",
+                other->thread_id);
+            MVM_gc_collect_free_nursery_uncopied(other, tc->gc_work[i].limit);
+
+            /* Handle exited threads. */
             if (MVM_load(&thread_obj->body.stage) == MVM_thread_stage_exited) {
                 /* Don't bother freeing gen2; we'll do it next time */
                 MVM_store(&thread_obj->body.stage, MVM_thread_stage_clearing_nursery);
@@ -222,6 +240,8 @@ static void finish_gc(MVMThreadContext *tc, MVMuint8 gen, MVMuint8 is_coordinato
                     "Thread %d run %d : set thread %d clearing nursery stage to %d\n",
                     other->thread_id, (int)MVM_load(&thread_obj->body.stage));
             }
+
+            /* Mark thread free to continue. */
             MVM_cas(&other->gc_status, MVMGCStatus_STOLEN, MVMGCStatus_UNABLE);
             MVM_cas(&other->gc_status, MVMGCStatus_INTERRUPT, MVMGCStatus_NONE);
         }
@@ -286,18 +306,24 @@ MVMint32 MVM_gc_is_thread_blocked(MVMThreadContext *tc) {
 }
 
 static MVMint32 is_full_collection(MVMThreadContext *tc) {
-    MVMuint64 percent_growth;
-    size_t rss, promoted;
+    MVMuint64 percent_growth, promoted;
+    size_t rss;
 
     /* If it's below the absolute minimum, quickly return. */
-    promoted = (size_t)MVM_load(&tc->instance->gc_promoted_bytes_since_last_full);
+    promoted = (MVMuint64)MVM_load(&tc->instance->gc_promoted_bytes_since_last_full);
     if (promoted < MVM_GC_GEN2_THRESHOLD_MINIMUM)
         return 0;
 
+    /* If we're heap profiling then don't consider the resident set size, as
+     * it will be hugely distorted by the profile data we record. */
+    if (MVM_profile_heap_profiling(tc))
+        return 1;
+
     /* Otherwise, consider percentage of resident set size. */
     if (uv_resident_set_memory(&rss) < 0 || rss == 0)
         rss = 50 * 1024 * 1024;
-    percent_growth = (100 * promoted) / rss;
+    percent_growth = (100 * promoted) / (MVMuint64)rss;
+
     return percent_growth >= MVM_GC_GEN2_THRESHOLD_PERCENT;
 }
 
@@ -305,6 +331,8 @@ static void run_gc(MVMThreadContext *tc, MVMuint8 what_to_do) {
     MVMuint8   gen;
     MVMuint32  i, n;
 
+    unsigned int interval_id;
+
 #if MVM_GC_DEBUG
     if (tc->in_spesh)
         MVM_panic(1, "Must not GC when in the specializer/JIT\n");
@@ -313,6 +341,12 @@ static void run_gc(MVMThreadContext *tc, MVMuint8 what_to_do) {
     /* Decide nursery or full collection. */
     gen = tc->instance->gc_full_collect ? MVMGCGenerations_Both : MVMGCGenerations_Nursery;
 
+    if (tc->instance->gc_full_collect) {
+        interval_id = MVM_telemetry_interval_start(tc, "start full collection");
+    } else {
+        interval_id = MVM_telemetry_interval_start(tc, "start minor collection");
+    }
+
     /* Do GC work for ourselves and any work threads. */
     for (i = 0, n = tc->gc_work_count ; i < n; i++) {
         MVMThreadContext *other = tc->gc_work[i].tc;
@@ -326,32 +360,7 @@ static void run_gc(MVMThreadContext *tc, MVMuint8 what_to_do) {
     /* Wait for everybody to agree we're done. */
     finish_gc(tc, gen, what_to_do == MVMGCWhatToDo_All);
 
-    /* Now we're all done, it's safe to finalize any objects that need it. */
-	/* XXX TODO explore the feasability of doing this in a background
-	 * finalizer/destructor thread and letting the main thread(s) continue
-	 * on their merry way(s). */
-    for (i = 0, n = tc->gc_work_count ; i < n; i++) {
-        MVMThreadContext *other = tc->gc_work[i].tc;
-
-        /* The thread might've been destroyed */
-        if (!other)
-            continue;
-
-        /* Contribute this thread's promoted bytes. */
-        MVM_add(&tc->instance->gc_promoted_bytes_since_last_full, other->gc_promoted_bytes);
-
-        /* Collect nursery and gen2 as needed. */
-        GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE,
-            "Thread %d run %d : collecting nursery uncopied of thread %d\n",
-            other->thread_id);
-        MVM_gc_collect_free_nursery_uncopied(other, tc->gc_work[i].limit);
-        if (gen == MVMGCGenerations_Both) {
-            GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE,
-                "Thread %d run %d : freeing gen2 of thread %d\n",
-                other->thread_id);
-            MVM_gc_collect_free_gen2_unmarked(other, 0);
-        }
-    }
+    MVM_telemetry_interval_stop(tc, interval_id, "finished run_gc");
 }
 
 /* This is called when the allocator finds it has run out of memory and wants
@@ -361,6 +370,8 @@ static void run_gc(MVMThreadContext *tc, MVMuint8 what_to_do) {
 void MVM_gc_enter_from_allocator(MVMThreadContext *tc) {
     GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE, "Thread %d run %d : Entered from allocate\n");
 
+    MVM_telemetry_timestamp(tc, "gc_enter_from_allocator");
+
     /* Try to start the GC run. */
     if (MVM_trycas(&tc->instance->gc_start, 0, 1)) {
         MVMThread *last_starter = NULL;
@@ -384,6 +395,8 @@ void MVM_gc_enter_from_allocator(MVMThreadContext *tc) {
         /* Decide if it will be a full collection. */
         tc->instance->gc_full_collect = is_full_collection(tc);
 
+        MVM_telemetry_timestamp(tc, "won the gc starting race");
+
         /* If profiling, record that GC is starting. */
         if (tc->instance->profiling)
             MVM_profiler_log_gc_start(tc, tc->instance->gc_full_collect);
@@ -460,6 +473,8 @@ void MVM_gc_enter_from_allocator(MVMThreadContext *tc) {
         if (tc->instance->profiling)
             MVM_profiler_log_gc_end(tc);
 
+        MVM_telemetry_timestamp(tc, "gc finished");
+
         GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE, "Thread %d run %d : GC complete (cooridnator)\n");
     }
     else {
@@ -478,6 +493,8 @@ void MVM_gc_enter_from_interrupt(MVMThreadContext *tc) {
 
     GCDEBUG_LOG(tc, MVM_GC_DEBUG_ORCHESTRATE, "Thread %d run %d : Entered from interrupt\n");
 
+    MVM_telemetry_timestamp(tc, "gc_enter_from_interrupt");
+
     /* If profiling, record that GC is starting. */
     if (tc->instance->profiling)
         MVM_profiler_log_gc_start(tc, is_full_collection(tc));
