@@ -156,6 +156,7 @@ void MVM_gc_root_add_tc_roots_to_worklist(MVMThreadContext *tc, MVMGCWorklist *w
 
     /* Current dispatcher. */
     add_collectable(tc, worklist, snapshot, tc->cur_dispatcher, "Current dispatcher");
+    add_collectable(tc, worklist, snapshot, tc->cur_dispatcher_for, "Current dispatcher for");
 
     /* Callback cache. */
     HASH_ITER(hash_handle, tc->native_callback_cache, current_cbceh, tmp_cbceh, bucket_tmp) {
@@ -387,8 +388,9 @@ void MVM_gc_root_add_frame_registers_to_worklist(MVMThreadContext *tc, MVMGCWork
     MVMuint16 *type_map;
     MVMuint8  *flag_map;
 
-    /* Scan locals. */
-    if (frame->work && frame->tc) {
+    /* We only need to do any of this work if the frame is in dynamic scope. */
+    if (frame->work) {
+        /* Scan locals. */
         if (frame->spesh_cand && frame->spesh_log_idx == -1 && frame->spesh_cand->local_types) {
             type_map = frame->spesh_cand->local_types;
             count    = frame->spesh_cand->num_locals;
@@ -400,37 +402,37 @@ void MVM_gc_root_add_frame_registers_to_worklist(MVMThreadContext *tc, MVMGCWork
         for (i = 0; i < count; i++)
             if (type_map[i] == MVM_reg_str || type_map[i] == MVM_reg_obj)
                 MVM_gc_worklist_add(tc, worklist, &frame->work[i].o);
-    }
 
-    /* Scan arg buffer if needed. */
-    if (frame->args && frame->cur_args_callsite) {
-        flag_map = frame->cur_args_callsite->arg_flags;
-        count = frame->cur_args_callsite->arg_count;
-        for (i = 0, flag = 0; i < count; i++, flag++) {
-            if (flag_map[flag] & MVM_CALLSITE_ARG_NAMED) {
-                /* Current position is name, then next is value. */
-                MVM_gc_worklist_add(tc, worklist, &frame->args[i].s);
-                i++;
+        /* Scan arg buffer if needed. */
+        if (frame->cur_args_callsite) {
+            flag_map = frame->cur_args_callsite->arg_flags;
+            count = frame->cur_args_callsite->arg_count;
+            for (i = 0, flag = 0; i < count; i++, flag++) {
+                if (flag_map[flag] & MVM_CALLSITE_ARG_NAMED) {
+                    /* Current position is name, then next is value. */
+                    MVM_gc_worklist_add(tc, worklist, &frame->args[i].s);
+                    i++;
+                }
+                if (flag_map[flag] & MVM_CALLSITE_ARG_STR || flag_map[flag] & MVM_CALLSITE_ARG_OBJ)
+                    MVM_gc_worklist_add(tc, worklist, &frame->args[i].o);
             }
-            if (flag_map[flag] & MVM_CALLSITE_ARG_STR || flag_map[flag] & MVM_CALLSITE_ARG_OBJ)
-                MVM_gc_worklist_add(tc, worklist, &frame->args[i].o);
         }
-    }
 
-    /* Scan arguments in case there was a flattening. Don't need to if
-     * there wasn't a flattening because orig args is a subset of locals. */
-    if (frame->params.arg_flags && frame->params.callsite->has_flattening) {
-        MVMArgProcContext *ctx = &frame->params;
-        flag_map = ctx->arg_flags;
-        count = ctx->arg_count;
-        for (i = 0, flag = 0; i < count; i++, flag++) {
-            if (flag_map[flag] & MVM_CALLSITE_ARG_NAMED) {
-                /* Current position is name, then next is value. */
-                MVM_gc_worklist_add(tc, worklist, &ctx->args[i].s);
-                i++;
+        /* Scan arguments in case there was a flattening. Don't need to if
+         * there wasn't a flattening because orig args is a subset of locals. */
+        if (frame->params.arg_flags && frame->params.callsite->has_flattening) {
+            MVMArgProcContext *ctx = &frame->params;
+            flag_map = ctx->arg_flags;
+            count = ctx->arg_count;
+            for (i = 0, flag = 0; i < count; i++, flag++) {
+                if (flag_map[flag] & MVM_CALLSITE_ARG_NAMED) {
+                    /* Current position is name, then next is value. */
+                    MVM_gc_worklist_add(tc, worklist, &ctx->args[i].s);
+                    i++;
+                }
+                if (flag_map[flag] & MVM_CALLSITE_ARG_STR || flag_map[flag] & MVM_CALLSITE_ARG_OBJ)
+                    MVM_gc_worklist_add(tc, worklist, &ctx->args[i].o);
             }
-            if (flag_map[flag] & MVM_CALLSITE_ARG_STR || flag_map[flag] & MVM_CALLSITE_ARG_OBJ)
-                MVM_gc_worklist_add(tc, worklist, &ctx->args[i].o);
         }
     }
 }
