@@ -102,28 +102,31 @@ static void instrument_graph(MVMThreadContext *tc, MVMSpeshGraph *g) {
 /* Adds instrumented version of the unspecialized bytecode. */
 static void add_instrumentation(MVMThreadContext *tc, MVMStaticFrame *sf) {
     MVMSpeshCode  *sc;
-    MVMSpeshGraph *sg = MVM_spesh_graph_create(tc, sf, 1);
+    MVMStaticFrameInstrumentation *ins;
+    MVMSpeshGraph *sg = MVM_spesh_graph_create(tc, sf, 1, 0);
     instrument_graph(tc, sg);
     sc = MVM_spesh_codegen(tc, sg);
-    sf->body.instrumented_bytecode        = sc->bytecode;
-    sf->body.instrumented_handlers        = sc->handlers;
-    sf->body.instrumented_bytecode_size   = sc->bytecode_size;
-    sf->body.uninstrumented_bytecode      = sf->body.bytecode;
-    sf->body.uninstrumented_handlers      = sf->body.handlers;
-    sf->body.uninstrumented_bytecode_size = sf->body.bytecode_size;
+    ins = MVM_calloc(1, sizeof(MVMStaticFrameInstrumentation));
+    ins->instrumented_bytecode        = sc->bytecode;
+    ins->instrumented_handlers        = sc->handlers;
+    ins->instrumented_bytecode_size   = sc->bytecode_size;
+    ins->uninstrumented_bytecode      = sf->body.bytecode;
+    ins->uninstrumented_handlers      = sf->body.handlers;
+    ins->uninstrumented_bytecode_size = sf->body.bytecode_size;
+    sf->body.instrumentation = ins;
     MVM_spesh_graph_destroy(tc, sg);
     MVM_free(sc);
 }
 
 /* Instruments code with detection and reporting of cross-thread writes. */
 void MVM_cross_thread_write_instrument(MVMThreadContext *tc, MVMStaticFrame *sf) {
-    if (sf->body.bytecode != sf->body.instrumented_bytecode) {
+    if (!sf->body.instrumentation || sf->body.bytecode != sf->body.instrumentation->instrumented_bytecode) {
         /* Handle main, non-specialized, bytecode. */
-        if (!sf->body.instrumented_bytecode)
+        if (!sf->body.instrumentation)
             add_instrumentation(tc, sf);
-        sf->body.bytecode      = sf->body.instrumented_bytecode;
-        sf->body.handlers      = sf->body.instrumented_handlers;
-        sf->body.bytecode_size = sf->body.instrumented_bytecode_size;
+        sf->body.bytecode      = sf->body.instrumentation->instrumented_bytecode;
+        sf->body.handlers      = sf->body.instrumentation->instrumented_handlers;
+        sf->body.bytecode_size = sf->body.instrumentation->instrumented_bytecode_size;
 
         /* Throw away any specializations; we'll need to reproduce them as
          * instrumented versions. */
@@ -143,6 +146,18 @@ static MVMint64 filtered_out(MVMThreadContext *tc, MVMObject *written) {
     if (REPR(written)->ID == MVM_REPR_ID_ConcBlockingQueue)
         return 1;
 
+    /* Write on object from event loop thread is usually shift of invokable. */
+    if (tc->instance->event_loop_thread)
+        if (written->header.owner == tc->instance->event_loop_thread->thread_id)
+            return 1;
+
+    /* Filter out writes to Sub and Method, since these are almost always just
+     * multi-dispatch caches. */
+    if (memcmp(written->st->debug_name, "Method", 6) == 0)
+        return 1;
+    if (memcmp(written->st->debug_name, "Sub", 3) == 0)
+        return 1;
+
     /* Otherwise, may be relevant. */
     return 0;
 }
@@ -187,8 +202,8 @@ void MVM_cross_thread_write_check(MVMThreadContext *tc, MVMObject *written, MVMi
                 break;
         }
         uv_mutex_lock(&(tc->instance->mutex_cross_thread_write_logging));
-        fprintf(stderr, "Thread %d %s an object allocated by thread %d\n",
-            tc->thread_id, guilty_desc, written->header.owner);
+        fprintf(stderr, "Thread %d %s an object (%s) allocated by thread %d\n",
+            tc->thread_id, guilty_desc, STABLE(written)->debug_name, written->header.owner);
         MVM_dump_backtrace(tc);
         fprintf(stderr, "\n");
         uv_mutex_unlock(&(tc->instance->mutex_cross_thread_write_logging));
