@@ -4,15 +4,11 @@
 typedef struct {
     /* The libuv handle to the socket. */
     uv_stream_t *handle;
-
-    /* Decode stream, for turning bytes into strings. */
-    MVMDecodeStream *ds;
 } MVMIOAsyncSocketData;
 
 /* Info we convey about a read task. */
 typedef struct {
     MVMOSHandle      *handle;
-    MVMDecodeStream  *ds;
     MVMObject        *buf_type;
     int               seq_number;
     MVMThreadContext *tc;
@@ -36,34 +32,25 @@ static void on_read(uv_stream_t *handle, ssize_t nread, const uv_buf_t *buf) {
     ReadInfo         *ri  = (ReadInfo *)handle->data;
     MVMThreadContext *tc  = ri->tc;
     MVMObject        *arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
-    MVMAsyncTask     *t   = (MVMAsyncTask *)MVM_repr_at_pos_o(tc,
-        tc->instance->event_loop_active, ri->work_idx);
+    MVMAsyncTask     *t   = MVM_io_eventloop_get_active_work(tc, ri->work_idx);
     MVM_repr_push_o(tc, arr, t->body.schedulee);
     if (nread >= 0) {
         MVMROOT(tc, t, {
         MVMROOT(tc, arr, {
+            MVMArray *res_buf;
+
             /* Push the sequence number. */
             MVMObject *seq_boxed = MVM_repr_box_int(tc,
                 tc->instance->boot_types.BOOTInt, ri->seq_number++);
             MVM_repr_push_o(tc, arr, seq_boxed);
 
-            /* Either need to produce a buffer or decode characters. */
-            if (ri->ds) {
-                MVMString *str;
-                MVMObject *boxed_str;
-                MVM_string_decodestream_add_bytes(tc, ri->ds, buf->base, nread);
-                str = MVM_string_decodestream_get_all(tc, ri->ds);
-                boxed_str = MVM_repr_box_str(tc, tc->instance->boot_types.BOOTStr, str);
-                MVM_repr_push_o(tc, arr, boxed_str);
-            }
-            else {
-                MVMArray *res_buf      = (MVMArray *)MVM_repr_alloc_init(tc, ri->buf_type);
-                res_buf->body.slots.i8 = (MVMint8 *)buf->base;
-                res_buf->body.start    = 0;
-                res_buf->body.ssize    = nread;
-                res_buf->body.elems    = nread;
-                MVM_repr_push_o(tc, arr, (MVMObject *)res_buf);
-            }
+            /* Produce a buffer and push it. */
+            res_buf      = (MVMArray *)MVM_repr_alloc_init(tc, ri->buf_type);
+            res_buf->body.slots.i8 = (MVMint8 *)buf->base;
+            res_buf->body.start    = 0;
+            res_buf->body.ssize    = buf->len;
+            res_buf->body.elems    = nread;
+            MVM_repr_push_o(tc, arr, (MVMObject *)res_buf);
 
             /* Finally, no error. */
             MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
@@ -83,6 +70,7 @@ static void on_read(uv_stream_t *handle, ssize_t nread, const uv_buf_t *buf) {
         if (buf->base)
             MVM_free(buf->base);
         uv_read_stop(handle);
+        MVM_io_eventloop_remove_active_work(tc, &(ri->work_idx));
     }
     else {
         MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTInt);
@@ -99,6 +87,7 @@ static void on_read(uv_stream_t *handle, ssize_t nread, const uv_buf_t *buf) {
         if (buf->base)
             MVM_free(buf->base);
         uv_read_stop(handle);
+        MVM_io_eventloop_remove_active_work(tc, &(ri->work_idx));
     }
     MVM_repr_push_o(tc, t->body.queue, arr);
 }
@@ -106,16 +95,35 @@ static void on_read(uv_stream_t *handle, ssize_t nread, const uv_buf_t *buf) {
 /* Does setup work for setting up asynchronous reads. */
 static void read_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_task, void *data) {
     MVMIOAsyncSocketData *handle_data;
+    ReadInfo             *ri;
     int                   r;
 
+    /* Ensure not closed. */
+    ri = (ReadInfo *)data;
+    handle_data = (MVMIOAsyncSocketData *)ri->handle->body.data;
+    if (!handle_data->handle || uv_is_closing((uv_handle_t *)handle_data->handle)) {
+        /* Closed, so immediately send done. */
+        MVMAsyncTask *t = (MVMAsyncTask *)async_task;
+        MVMROOT(tc, t, {
+            MVMObject *arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
+            MVM_repr_push_o(tc, arr, t->body.schedulee);
+            MVMROOT(tc, arr, {
+                MVMObject *final = MVM_repr_box_int(tc,
+                    tc->instance->boot_types.BOOTInt, ri->seq_number);
+                MVM_repr_push_o(tc, arr, final);
+            });
+            MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
+            MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
+            MVM_repr_push_o(tc, t->body.queue, arr);
+        });
+        return;
+    }
+
     /* Add to work in progress. */
-    ReadInfo *ri  = (ReadInfo *)data;
     ri->tc        = tc;
-    ri->work_idx  = MVM_repr_elems(tc, tc->instance->event_loop_active);
-    MVM_repr_push_o(tc, tc->instance->event_loop_active, async_task);
+    ri->work_idx  = MVM_io_eventloop_add_active_work(tc, async_task);
 
     /* Start reading the stream. */
-    handle_data = (MVMIOAsyncSocketData *)ri->handle->body.data;
     handle_data->handle->data = data;
     if ((r = uv_read_start(handle_data->handle, on_alloc, on_read)) < 0) {
         /* Error; need to notify. */
@@ -134,6 +142,7 @@ static void read_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_t
             });
             MVM_repr_push_o(tc, t->body.queue, arr);
         });
+        MVM_io_eventloop_remove_active_work(tc, &(ri->work_idx));
     }
 }
 
@@ -146,12 +155,8 @@ static void read_gc_mark(MVMThreadContext *tc, void *data, MVMGCWorklist *workli
 
 /* Frees info for a read task. */
 static void read_gc_free(MVMThreadContext *tc, MVMObject *t, void *data) {
-    if (data) {
-        ReadInfo *ri = (ReadInfo *)data;
-        if (ri->ds)
-            MVM_string_decodestream_destory(tc, ri->ds);
+    if (data)
         MVM_free(data);
-    }
 }
 
 /* Operations table for async read task. */
@@ -162,43 +167,6 @@ static const MVMAsyncTaskOps read_op_table = {
     read_gc_free
 };
 
-static MVMAsyncTask * read_chars(MVMThreadContext *tc, MVMOSHandle *h, MVMObject *queue,
-                                 MVMObject *schedulee, MVMObject *async_type) {
-    MVMAsyncTask *task;
-    ReadInfo    *ri;
-
-    /* Validate REPRs. */
-    if (REPR(queue)->ID != MVM_REPR_ID_ConcBlockingQueue)
-        MVM_exception_throw_adhoc(tc,
-            "asyncreadchars target queue must have ConcBlockingQueue REPR");
-    if (REPR(async_type)->ID != MVM_REPR_ID_MVMAsyncTask)
-        MVM_exception_throw_adhoc(tc,
-            "asyncreadchars result type must have REPR AsyncTask");
-
-    /* Create async task handle. */
-    MVMROOT(tc, queue, {
-    MVMROOT(tc, schedulee, {
-    MVMROOT(tc, h, {
-        task = (MVMAsyncTask *)MVM_repr_alloc_init(tc, async_type);
-    });
-    });
-    });
-    MVM_ASSIGN_REF(tc, &(task->common.header), task->body.queue, queue);
-    MVM_ASSIGN_REF(tc, &(task->common.header), task->body.schedulee, schedulee);
-    task->body.ops  = &read_op_table;
-    ri              = MVM_calloc(1, sizeof(ReadInfo));
-    ri->ds          = MVM_string_decodestream_create(tc, MVM_encoding_type_utf8, 0, 0);
-    MVM_ASSIGN_REF(tc, &(task->common.header), ri->handle, h);
-    task->body.data = ri;
-
-    /* Hand the task off to the event loop. */
-    MVMROOT(tc, task, {
-        MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
-    });
-
-    return task;
-}
-
 static MVMAsyncTask * read_bytes(MVMThreadContext *tc, MVMOSHandle *h, MVMObject *queue,
                                  MVMObject *schedulee, MVMObject *buf_type, MVMObject *async_type) {
     MVMAsyncTask *task;
@@ -211,7 +179,7 @@ static MVMAsyncTask * read_bytes(MVMThreadContext *tc, MVMOSHandle *h, MVMObject
     if (REPR(async_type)->ID != MVM_REPR_ID_MVMAsyncTask)
         MVM_exception_throw_adhoc(tc,
             "asyncreadbytes result type must have REPR AsyncTask");
-    if (REPR(buf_type)->ID == MVM_REPR_ID_MVMArray) {
+    if (REPR(buf_type)->ID == MVM_REPR_ID_VMArray) {
         MVMint32 slot_type = ((MVMArrayREPRData *)STABLE(buf_type)->REPR_data)->slot_type;
         if (slot_type != MVM_ARRAY_U8 && slot_type != MVM_ARRAY_I8)
             MVM_exception_throw_adhoc(tc, "asyncreadbytes buffer type must be an array of uint8 or int8");
@@ -249,7 +217,6 @@ static MVMAsyncTask * read_bytes(MVMThreadContext *tc, MVMOSHandle *h, MVMObject
 /* Info we convey about a write task. */
 typedef struct {
     MVMOSHandle      *handle;
-    MVMString        *str_data;
     MVMObject        *buf_data;
     uv_write_t       *req;
     uv_buf_t          buf;
@@ -262,8 +229,7 @@ static void on_write(uv_write_t *req, int status) {
     WriteInfo        *wi  = (WriteInfo *)req->data;
     MVMThreadContext *tc  = wi->tc;
     MVMObject        *arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
-    MVMAsyncTask     *t   = (MVMAsyncTask *)MVM_repr_at_pos_o(tc,
-        tc->instance->event_loop_active, wi->work_idx);
+    MVMAsyncTask     *t   = MVM_io_eventloop_get_active_work(tc, wi->work_idx);
     MVM_repr_push_o(tc, arr, t->body.schedulee);
     if (status >= 0) {
         MVMROOT(tc, arr, {
@@ -289,43 +255,52 @@ static void on_write(uv_write_t *req, int status) {
         });
     }
     MVM_repr_push_o(tc, t->body.queue, arr);
-    if (wi->str_data)
-        MVM_free(wi->buf.base);
     MVM_free(wi->req);
+    MVM_io_eventloop_remove_active_work(tc, &(wi->work_idx));
 }
 
 /* Does setup work for an asynchronous write. */
 static void write_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_task, void *data) {
     MVMIOAsyncSocketData *handle_data;
+    MVMArray             *buffer;
+    WriteInfo            *wi;
     char                 *output;
     int                   output_size, r;
 
-    /* Add to work in progress. */
-    WriteInfo *wi = (WriteInfo *)data;
-    wi->tc        = tc;
-    wi->work_idx  = MVM_repr_elems(tc, tc->instance->event_loop_active);
-    MVM_repr_push_o(tc, tc->instance->event_loop_active, async_task);
+    /* Ensure not closed. */
+    wi = (WriteInfo *)data;
+    handle_data = (MVMIOAsyncSocketData *)wi->handle->body.data;
+    if (!handle_data->handle || uv_is_closing((uv_handle_t *)handle_data->handle)) {
+        MVMROOT(tc, async_task, {
+            MVMObject    *arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
+            MVMAsyncTask *t   = (MVMAsyncTask *)async_task;
+            MVM_repr_push_o(tc, arr, t->body.schedulee);
+            MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTInt);
+            MVMROOT(tc, arr, {
+                MVMString *msg_str = MVM_string_ascii_decode_nt(tc,
+                    tc->instance->VMString, "Cannot write to a closed socket");
+                MVMObject *msg_box = MVM_repr_box_str(tc,
+                    tc->instance->boot_types.BOOTStr, msg_str);
+                MVM_repr_push_o(tc, arr, msg_box);
+            });
+            MVM_repr_push_o(tc, t->body.queue, arr);
+        });
+        return;
+    }
 
-    /* Encode the string, or extract buf data. */
-    if (wi->str_data) {
-        MVMuint64 output_size_64;
-        output = MVM_string_utf8_encode(tc, wi->str_data, &output_size_64, 0);
-        output_size = (int)output_size_64;
-    }
-    else {
-        MVMArray *buffer = (MVMArray *)wi->buf_data;
-        output = (char *)(buffer->body.slots.i8 + buffer->body.start);
-        output_size = (int)buffer->body.elems;
-    }
+    /* Add to work in progress. */
+    wi->tc = tc;
+    wi->work_idx = MVM_io_eventloop_add_active_work(tc, async_task);
+
+    /* Extract buf data. */
+    buffer = (MVMArray *)wi->buf_data;
+    output = (char *)(buffer->body.slots.i8 + buffer->body.start);
+    output_size = (int)buffer->body.elems;
 
     /* Create and initialize write request. */
     wi->req           = MVM_malloc(sizeof(uv_write_t));
     wi->buf           = uv_buf_init(output, output_size);
     wi->req->data     = data;
-    handle_data       = (MVMIOAsyncSocketData *)wi->handle->body.data;
-
-    if (uv_is_closing((uv_handle_t *)handle_data->handle))
-        MVM_exception_throw_adhoc(tc, "cannot write to a closed socket");
 
     if ((r = uv_write(wi->req, handle_data->handle, &(wi->buf), 1, on_write)) < 0) {
         /* Error; need to notify. */
@@ -347,6 +322,7 @@ static void write_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
         /* Cleanup handle. */
         MVM_free(wi->req);
         wi->req = NULL;
+        MVM_io_eventloop_remove_active_work(tc, &(wi->work_idx));
     }
 }
 
@@ -354,7 +330,6 @@ static void write_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
 static void write_gc_mark(MVMThreadContext *tc, void *data, MVMGCWorklist *worklist) {
     WriteInfo *wi = (WriteInfo *)data;
     MVM_gc_worklist_add(tc, worklist, &wi->handle);
-    MVM_gc_worklist_add(tc, worklist, &wi->str_data);
     MVM_gc_worklist_add(tc, worklist, &wi->buf_data);
 }
 
@@ -372,45 +347,6 @@ static const MVMAsyncTaskOps write_op_table = {
     write_gc_free
 };
 
-static MVMAsyncTask * write_str(MVMThreadContext *tc, MVMOSHandle *h, MVMObject *queue,
-                                MVMObject *schedulee, MVMString *s, MVMObject *async_type) {
-    MVMAsyncTask *task;
-    WriteInfo    *wi;
-
-    /* Validate REPRs. */
-    if (REPR(queue)->ID != MVM_REPR_ID_ConcBlockingQueue)
-        MVM_exception_throw_adhoc(tc,
-            "asyncwritestr target queue must have ConcBlockingQueue REPR");
-    if (REPR(async_type)->ID != MVM_REPR_ID_MVMAsyncTask)
-        MVM_exception_throw_adhoc(tc,
-            "asyncwritestr result type must have REPR AsyncTask");
-
-    /* Create async task handle. */
-    MVMROOT(tc, queue, {
-    MVMROOT(tc, schedulee, {
-    MVMROOT(tc, h, {
-    MVMROOT(tc, s, {
-        task = (MVMAsyncTask *)MVM_repr_alloc_init(tc, async_type);
-    });
-    });
-    });
-    });
-    MVM_ASSIGN_REF(tc, &(task->common.header), task->body.queue, queue);
-    MVM_ASSIGN_REF(tc, &(task->common.header), task->body.schedulee, schedulee);
-    task->body.ops  = &write_op_table;
-    wi              = MVM_calloc(1, sizeof(WriteInfo));
-    MVM_ASSIGN_REF(tc, &(task->common.header), wi->handle, h);
-    MVM_ASSIGN_REF(tc, &(task->common.header), wi->str_data, s);
-    task->body.data = wi;
-
-    /* Hand the task off to the event loop. */
-    MVMROOT(tc, task, {
-        MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
-    });
-
-    return task;
-}
-
 static MVMAsyncTask * write_bytes(MVMThreadContext *tc, MVMOSHandle *h, MVMObject *queue,
                                   MVMObject *schedulee, MVMObject *buffer, MVMObject *async_type) {
     MVMAsyncTask *task;
@@ -423,7 +359,7 @@ static MVMAsyncTask * write_bytes(MVMThreadContext *tc, MVMOSHandle *h, MVMObjec
     if (REPR(async_type)->ID != MVM_REPR_ID_MVMAsyncTask)
         MVM_exception_throw_adhoc(tc,
             "asyncwritebytes result type must have REPR AsyncTask");
-    if (!IS_CONCRETE(buffer) || REPR(buffer)->ID != MVM_REPR_ID_MVMArray)
+    if (!IS_CONCRETE(buffer) || REPR(buffer)->ID != MVM_REPR_ID_VMArray)
         MVM_exception_throw_adhoc(tc, "asyncwritebytes requires a native array to read from");
     if (((MVMArrayREPRData *)STABLE(buffer)->REPR_data)->slot_type != MVM_ARRAY_U8
         && ((MVMArrayREPRData *)STABLE(buffer)->REPR_data)->slot_type != MVM_ARRAY_I8)
@@ -455,56 +391,68 @@ static MVMAsyncTask * write_bytes(MVMThreadContext *tc, MVMOSHandle *h, MVMObjec
     return task;
 }
 
+/* Info we convey about a socket close task. */
+typedef struct {
+    MVMOSHandle *handle;
+} CloseInfo;
+
 /* Does an asynchronous close (since it must run on the event loop). */
 static void close_perform(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_task, void *data) {
-    uv_handle_t *handle = (uv_handle_t *)data;
+    CloseInfo *ci = (CloseInfo *)data;
+    MVMIOAsyncSocketData *handle_data = (MVMIOAsyncSocketData *)ci->handle->body.data;
+    uv_handle_t *handle = (uv_handle_t *)handle_data->handle;
+    if (handle && !uv_is_closing(handle)) {
+        handle_data->handle = NULL;
+        uv_close(handle, free_on_close_cb);
+    }
+}
 
-    if (uv_is_closing(handle))
-        MVM_exception_throw_adhoc(tc, "cannot close a closed socket");
+/* Marks objects for a close task. */
+static void close_gc_mark(MVMThreadContext *tc, void *data, MVMGCWorklist *worklist) {
+    CloseInfo *ci = (CloseInfo *)data;
+    MVM_gc_worklist_add(tc, worklist, &ci->handle);
+}
 
-    uv_close(handle, free_on_close_cb);
+/* Frees info for a close task. */
+static void close_gc_free(MVMThreadContext *tc, MVMObject *t, void *data) {
+    if (data)
+        MVM_free(data);
 }
 
 /* Operations table for async close task. */
 static const MVMAsyncTaskOps close_op_table = {
     close_perform,
     NULL,
-    NULL,
-    NULL
+    close_gc_mark,
+    close_gc_free
 };
 
 static MVMint64 close_socket(MVMThreadContext *tc, MVMOSHandle *h) {
     MVMIOAsyncSocketData *data = (MVMIOAsyncSocketData *)h->body.data;
     MVMAsyncTask *task;
+    CloseInfo *ci;
 
     MVMROOT(tc, h, {
         task = (MVMAsyncTask *)MVM_repr_alloc_init(tc,
             tc->instance->boot_types.BOOTAsync);
     });
-    task->body.ops  = &close_op_table;
-    task->body.data = data->handle;
+    task->body.ops = &close_op_table;
+    ci = MVM_calloc(1, sizeof(CloseInfo));
+    MVM_ASSIGN_REF(tc, &(task->common.header), ci->handle, h);
+    task->body.data = ci;
     MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
 
     return 0;
 }
 
-static void gc_free(MVMThreadContext *tc, MVMObject *h, void *d) {
-    MVMIOAsyncSocketData *data = (MVMIOAsyncSocketData *)d;
-    if (data->ds) {
-        MVM_string_decodestream_destory(tc, data->ds);
-        data->ds = NULL;
-    }
-}
-
 /* IO ops table, populated with functions. */
 static const MVMIOClosable      closable       = { close_socket };
-static const MVMIOAsyncReadable async_readable = { read_chars, read_bytes };
-static const MVMIOAsyncWritable async_writable = { write_str, write_bytes };
+static const MVMIOAsyncReadable async_readable = { read_bytes };
+static const MVMIOAsyncWritable async_writable = { write_bytes };
 static const MVMIOOps op_table = {
     &closable,
     NULL,
     NULL,
-    NULL,
     &async_readable,
     &async_writable,
     NULL,
@@ -514,7 +462,7 @@ static const MVMIOOps op_table = {
     NULL,
     NULL,
     NULL,
-    gc_free
+    NULL
 };
 
 /* Info we convey about a connection attempt task. */
@@ -526,13 +474,12 @@ typedef struct {
     int               work_idx;
 } ConnectInfo;
 
-/* When a connection takes place, need to keep the appropriate promise. */
+/* When a connection takes place, need to send result. */
 static void on_connect(uv_connect_t* req, int status) {
     ConnectInfo      *ci  = (ConnectInfo *)req->data;
     MVMThreadContext *tc  = ci->tc;
     MVMObject        *arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
-    MVMAsyncTask     *t   = (MVMAsyncTask *)MVM_repr_at_pos_o(tc,
-        tc->instance->event_loop_active, ci->work_idx);
+    MVMAsyncTask     *t   = MVM_io_eventloop_get_active_work(tc, ci->work_idx);
     MVM_repr_push_o(tc, arr, t->body.schedulee);
     if (status >= 0) {
         /* Allocate and set up handle. */
@@ -562,6 +509,7 @@ static void on_connect(uv_connect_t* req, int status) {
     }
     MVM_repr_push_o(tc, t->body.queue, arr);
     MVM_free(req);
+    MVM_io_eventloop_remove_active_work(tc, &(ci->work_idx));
 }
 
 /* Initilalize the connection on the event loop. */
@@ -571,8 +519,7 @@ static void connect_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *asyn
     /* Add to work in progress. */
     ConnectInfo *ci = (ConnectInfo *)data;
     ci->tc        = tc;
-    ci->work_idx  = MVM_repr_elems(tc, tc->instance->event_loop_active);
-    MVM_repr_push_o(tc, tc->instance->event_loop_active, async_task);
+    ci->work_idx  = MVM_io_eventloop_add_active_work(tc, async_task);
 
     /* Create and initialize socket and connection. */
     ci->socket        = MVM_malloc(sizeof(uv_tcp_t));
@@ -601,6 +548,7 @@ static void connect_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *asyn
         ci->connect = NULL;
         uv_close((uv_handle_t *)ci->socket, free_on_close_cb);
         ci->socket = NULL;
+        MVM_io_eventloop_remove_active_work(tc, &(ci->work_idx));
     }
 }
 
@@ -675,8 +623,7 @@ static void on_connection(uv_stream_t *server, int status) {
     ListenInfo       *li     = (ListenInfo *)server->data;
     MVMThreadContext *tc     = li->tc;
     MVMObject        *arr    = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
-    MVMAsyncTask     *t      = (MVMAsyncTask *)MVM_repr_at_pos_o(tc,
-        tc->instance->event_loop_active, li->work_idx);
+    MVMAsyncTask     *t      = MVM_io_eventloop_get_active_work(tc, li->work_idx);
 
     uv_tcp_t         *client = MVM_malloc(sizeof(uv_tcp_t));
     int               r;
@@ -721,8 +668,7 @@ static void listen_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async
     /* Add to work in progress. */
     ListenInfo *li = (ListenInfo *)data;
     li->tc         = tc;
-    li->work_idx   = MVM_repr_elems(tc, tc->instance->event_loop_active);
-    MVM_repr_push_o(tc, tc->instance->event_loop_active, async_task);
+    li->work_idx   = MVM_io_eventloop_add_active_work(tc, async_task);
 
     /* Create and initialize socket and connection, and start listening. */
     li->socket        = MVM_malloc(sizeof(uv_tcp_t));
@@ -747,6 +693,7 @@ static void listen_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async
         });
         uv_close((uv_handle_t *)li->socket, free_on_close_cb);
         li->socket = NULL;
+        MVM_io_eventloop_remove_active_work(tc, &(li->work_idx));
         return;
     }
 }
@@ -756,8 +703,8 @@ static void on_listen_cancelled(uv_handle_t *handle) {
     ListenInfo       *li = (ListenInfo *)handle->data;
     MVMThreadContext *tc = li->tc;
     MVM_io_eventloop_send_cancellation_notification(tc,
-        (MVMAsyncTask *)MVM_repr_at_pos_o(tc, tc->instance->event_loop_active,
-            li->work_idx));
+        MVM_io_eventloop_get_active_work(tc, li->work_idx));
+    MVM_io_eventloop_remove_active_work(tc, &(li->work_idx));
 }
 static void listen_cancel(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_task, void *data) {
     ListenInfo *li = (ListenInfo *)data;
