@@ -7,15 +7,11 @@
 typedef struct {
     /* The libuv handle to the socket. */
     uv_udp_t *handle;
-
-    /* Decode stream, for turning bytes into strings. */
-    MVMDecodeStream *ds;
 } MVMIOAsyncUDPSocketData;
 
 /* Info we convey about a read task. */
 typedef struct {
     MVMOSHandle      *handle;
-    MVMDecodeStream  *ds;
     MVMObject        *buf_type;
     int               seq_number;
     MVMThreadContext *tc;
@@ -39,34 +35,25 @@ static void on_read(uv_udp_t *handle, ssize_t nread, const uv_buf_t *buf, const
     ReadInfo         *ri  = (ReadInfo *)handle->data;
     MVMThreadContext *tc  = ri->tc;
     MVMObject        *arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
-    MVMAsyncTask     *t   = (MVMAsyncTask *)MVM_repr_at_pos_o(tc,
-        tc->instance->event_loop_active, ri->work_idx);
+    MVMAsyncTask     *t   = MVM_io_eventloop_get_active_work(tc, ri->work_idx);
     MVM_repr_push_o(tc, arr, t->body.schedulee);
     if (nread >= 0) {
         MVMROOT(tc, t, {
         MVMROOT(tc, arr, {
+            MVMArray *res_buf;
+
             /* Push the sequence number. */
             MVMObject *seq_boxed = MVM_repr_box_int(tc,
                 tc->instance->boot_types.BOOTInt, ri->seq_number++);
             MVM_repr_push_o(tc, arr, seq_boxed);
 
-            /* Either need to produce a buffer or decode characters. */
-            if (ri->ds) {
-                MVMString *str;
-                MVMObject *boxed_str;
-                MVM_string_decodestream_add_bytes(tc, ri->ds, buf->base, nread);
-                str = MVM_string_decodestream_get_all(tc, ri->ds);
-                boxed_str = MVM_repr_box_str(tc, tc->instance->boot_types.BOOTStr, str);
-                MVM_repr_push_o(tc, arr, boxed_str);
-            }
-            else {
-                MVMArray *res_buf      = (MVMArray *)MVM_repr_alloc_init(tc, ri->buf_type);
-                res_buf->body.slots.i8 = (MVMint8 *)buf->base;
-                res_buf->body.start    = 0;
-                res_buf->body.ssize    = nread;
-                res_buf->body.elems    = nread;
-                MVM_repr_push_o(tc, arr, (MVMObject *)res_buf);
-            }
+            /* Produce a buffer and push it. */
+            res_buf      = (MVMArray *)MVM_repr_alloc_init(tc, ri->buf_type);
+            res_buf->body.slots.i8 = (MVMint8 *)buf->base;
+            res_buf->body.start    = 0;
+            res_buf->body.ssize    = buf->len;
+            res_buf->body.elems    = nread;
+            MVM_repr_push_o(tc, arr, (MVMObject *)res_buf);
 
             /* Finally, no error. */
             MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
@@ -86,6 +73,7 @@ static void on_read(uv_udp_t *handle, ssize_t nread, const uv_buf_t *buf, const
         if (buf->base)
             MVM_free(buf->base);
         uv_udp_recv_stop(handle);
+        MVM_io_eventloop_remove_active_work(tc, &(ri->work_idx));
     }
     else {
         MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTInt);
@@ -102,6 +90,7 @@ static void on_read(uv_udp_t *handle, ssize_t nread, const uv_buf_t *buf, const
         if (buf->base)
             MVM_free(buf->base);
         uv_udp_recv_stop(handle);
+        MVM_io_eventloop_remove_active_work(tc, &(ri->work_idx));
     }
     MVM_repr_push_o(tc, t->body.queue, arr);
 }
@@ -114,8 +103,7 @@ static void read_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_t
     /* Add to work in progress. */
     ReadInfo *ri  = (ReadInfo *)data;
     ri->tc        = tc;
-    ri->work_idx  = MVM_repr_elems(tc, tc->instance->event_loop_active);
-    MVM_repr_push_o(tc, tc->instance->event_loop_active, async_task);
+    ri->work_idx  = MVM_io_eventloop_add_active_work(tc, async_task);
 
     /* Start reading the stream. */
     handle_data = (MVMIOAsyncUDPSocketData *)ri->handle->body.data;
@@ -149,12 +137,8 @@ static void read_gc_mark(MVMThreadContext *tc, void *data, MVMGCWorklist *workli
 
 /* Frees info for a read task. */
 static void read_gc_free(MVMThreadContext *tc, MVMObject *t, void *data) {
-    if (data) {
-        ReadInfo *ri = (ReadInfo *)data;
-        if (ri->ds)
-            MVM_string_decodestream_destory(tc, ri->ds);
+    if (data)
         MVM_free(data);
-    }
 }
 
 /* Operations table for async read task. */
@@ -165,43 +149,6 @@ static const MVMAsyncTaskOps read_op_table = {
     read_gc_free
 };
 
-static MVMAsyncTask * read_chars(MVMThreadContext *tc, MVMOSHandle *h, MVMObject *queue,
-                                 MVMObject *schedulee, MVMObject *async_type) {
-    MVMAsyncTask *task;
-    ReadInfo    *ri;
-
-    /* Validate REPRs. */
-    if (REPR(queue)->ID != MVM_REPR_ID_ConcBlockingQueue)
-        MVM_exception_throw_adhoc(tc,
-            "asyncreadchars target queue must have ConcBlockingQueue REPR");
-    if (REPR(async_type)->ID != MVM_REPR_ID_MVMAsyncTask)
-        MVM_exception_throw_adhoc(tc,
-            "asyncreadchars result type must have REPR AsyncTask");
-
-    /* Create async task handle. */
-    MVMROOT(tc, queue, {
-    MVMROOT(tc, schedulee, {
-    MVMROOT(tc, h, {
-        task = (MVMAsyncTask *)MVM_repr_alloc_init(tc, async_type);
-    });
-    });
-    });
-    MVM_ASSIGN_REF(tc, &(task->common.header), task->body.queue, queue);
-    MVM_ASSIGN_REF(tc, &(task->common.header), task->body.schedulee, schedulee);
-    task->body.ops  = &read_op_table;
-    ri              = MVM_calloc(1, sizeof(ReadInfo));
-    ri->ds          = MVM_string_decodestream_create(tc, MVM_encoding_type_utf8, 0, 0);
-    MVM_ASSIGN_REF(tc, &(task->common.header), ri->handle, h);
-    task->body.data = ri;
-
-    /* Hand the task off to the event loop. */
-    MVMROOT(tc, task, {
-        MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
-    });
-
-    return task;
-}
-
 static MVMAsyncTask * read_bytes(MVMThreadContext *tc, MVMOSHandle *h, MVMObject *queue,
                                  MVMObject *schedulee, MVMObject *buf_type, MVMObject *async_type) {
     MVMAsyncTask *task;
@@ -214,7 +161,7 @@ static MVMAsyncTask * read_bytes(MVMThreadContext *tc, MVMOSHandle *h, MVMObject
     if (REPR(async_type)->ID != MVM_REPR_ID_MVMAsyncTask)
         MVM_exception_throw_adhoc(tc,
             "asyncreadbytes result type must have REPR AsyncTask");
-    if (REPR(buf_type)->ID == MVM_REPR_ID_MVMArray) {
+    if (REPR(buf_type)->ID == MVM_REPR_ID_VMArray) {
         MVMint32 slot_type = ((MVMArrayREPRData *)STABLE(buf_type)->REPR_data)->slot_type;
         if (slot_type != MVM_ARRAY_U8 && slot_type != MVM_ARRAY_I8)
             MVM_exception_throw_adhoc(tc, "asyncreadbytes buffer type must be an array of uint8 or int8");
@@ -252,7 +199,6 @@ static MVMAsyncTask * read_bytes(MVMThreadContext *tc, MVMOSHandle *h, MVMObject
 /* Info we convey about a write task. */
 typedef struct {
     MVMOSHandle      *handle;
-    MVMString        *str_data;
     MVMObject        *buf_data;
     uv_udp_send_t    *req;
     uv_buf_t          buf;
@@ -266,8 +212,7 @@ static void on_write(uv_udp_send_t *req, int status) {
     WriteInfo        *wi  = (WriteInfo *)req->data;
     MVMThreadContext *tc  = wi->tc;
     MVMObject        *arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
-    MVMAsyncTask     *t   = (MVMAsyncTask *)MVM_repr_at_pos_o(tc,
-        tc->instance->event_loop_active, wi->work_idx);
+    MVMAsyncTask     *t   = MVM_io_eventloop_get_active_work(tc, wi->work_idx);
     MVM_repr_push_o(tc, arr, t->body.schedulee);
     if (status >= 0) {
         MVMROOT(tc, arr, {
@@ -293,34 +238,26 @@ static void on_write(uv_udp_send_t *req, int status) {
         });
     }
     MVM_repr_push_o(tc, t->body.queue, arr);
-    if (wi->str_data)
-        MVM_free(wi->buf.base);
     MVM_free(wi->req);
+    MVM_io_eventloop_remove_active_work(tc, &(wi->work_idx));
 }
 
 /* Does setup work for an asynchronous write. */
 static void write_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_task, void *data) {
     MVMIOAsyncUDPSocketData *handle_data;
-    char                 *output;
-    int                   output_size, r;
+    MVMArray                *buffer;
+    char                    *output;
+    int                      output_size, r;
 
     /* Add to work in progress. */
     WriteInfo *wi = (WriteInfo *)data;
     wi->tc        = tc;
-    wi->work_idx  = MVM_repr_elems(tc, tc->instance->event_loop_active);
-    MVM_repr_push_o(tc, tc->instance->event_loop_active, async_task);
+    wi->work_idx  = MVM_io_eventloop_add_active_work(tc, async_task);
 
-    /* Encode the string, or extract buf data. */
-    if (wi->str_data) {
-        MVMuint64 output_size_64;
-        output = MVM_string_utf8_encode(tc, wi->str_data, &output_size_64, 0);
-        output_size = (int)output_size_64;
-    }
-    else {
-        MVMArray *buffer = (MVMArray *)wi->buf_data;
-        output = (char *)(buffer->body.slots.i8 + buffer->body.start);
-        output_size = (int)buffer->body.elems;
-    }
+    /* Extract buf data. */
+    buffer = (MVMArray *)wi->buf_data;
+    output = (char *)(buffer->body.slots.i8 + buffer->body.start);
+    output_size = (int)buffer->body.elems;
 
     /* Create and initialize write request. */
     wi->req           = MVM_malloc(sizeof(uv_udp_send_t));
@@ -351,6 +288,7 @@ static void write_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
         /* Cleanup handle. */
         MVM_free(wi->req);
         wi->req = NULL;
+        MVM_io_eventloop_remove_active_work(tc, &(wi->work_idx));
     }
 }
 
@@ -358,7 +296,6 @@ static void write_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
 static void write_gc_mark(MVMThreadContext *tc, void *data, MVMGCWorklist *worklist) {
     WriteInfo *wi = (WriteInfo *)data;
     MVM_gc_worklist_add(tc, worklist, &wi->handle);
-    MVM_gc_worklist_add(tc, worklist, &wi->str_data);
     MVM_gc_worklist_add(tc, worklist, &wi->buf_data);
 }
 
@@ -380,51 +317,6 @@ static const MVMAsyncTaskOps write_op_table = {
     write_gc_free
 };
 
-static MVMAsyncTask * write_str_to(MVMThreadContext *tc, MVMOSHandle *h, MVMObject *queue,
-                                   MVMObject *schedulee, MVMString *s, MVMObject *async_type,
-                                   MVMString *host, MVMint64 port) {
-    MVMAsyncTask    *task;
-    WriteInfo       *wi;
-    struct sockaddr *dest_addr;
-
-    /* Validate REPRs. */
-    if (REPR(queue)->ID != MVM_REPR_ID_ConcBlockingQueue)
-        MVM_exception_throw_adhoc(tc,
-            "asyncwritestrto target queue must have ConcBlockingQueue REPR");
-    if (REPR(async_type)->ID != MVM_REPR_ID_MVMAsyncTask)
-        MVM_exception_throw_adhoc(tc,
-            "asyncwritestrto result type must have REPR AsyncTask");
-
-    /* Resolve destination. */
-    dest_addr = MVM_io_resolve_host_name(tc, host, port);
-
-    /* Create async task handle. */
-    MVMROOT(tc, queue, {
-    MVMROOT(tc, schedulee, {
-    MVMROOT(tc, h, {
-    MVMROOT(tc, s, {
-        task = (MVMAsyncTask *)MVM_repr_alloc_init(tc, async_type);
-    });
-    });
-    });
-    });
-    MVM_ASSIGN_REF(tc, &(task->common.header), task->body.queue, queue);
-    MVM_ASSIGN_REF(tc, &(task->common.header), task->body.schedulee, schedulee);
-    task->body.ops  = &write_op_table;
-    wi              = MVM_calloc(1, sizeof(WriteInfo));
-    MVM_ASSIGN_REF(tc, &(task->common.header), wi->handle, h);
-    MVM_ASSIGN_REF(tc, &(task->common.header), wi->str_data, s);
-    wi->dest_addr = dest_addr;
-    task->body.data = wi;
-
-    /* Hand the task off to the event loop. */
-    MVMROOT(tc, task, {
-        MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
-    });
-
-    return task;
-}
-
 static MVMAsyncTask * write_bytes_to(MVMThreadContext *tc, MVMOSHandle *h, MVMObject *queue,
                                      MVMObject *schedulee, MVMObject *buffer, MVMObject *async_type,
                                      MVMString *host, MVMint64 port) {
@@ -439,7 +331,7 @@ static MVMAsyncTask * write_bytes_to(MVMThreadContext *tc, MVMOSHandle *h, MVMOb
     if (REPR(async_type)->ID != MVM_REPR_ID_MVMAsyncTask)
         MVM_exception_throw_adhoc(tc,
             "asyncwritebytesto result type must have REPR AsyncTask");
-    if (!IS_CONCRETE(buffer) || REPR(buffer)->ID != MVM_REPR_ID_MVMArray)
+    if (!IS_CONCRETE(buffer) || REPR(buffer)->ID != MVM_REPR_ID_VMArray)
         MVM_exception_throw_adhoc(tc, "asyncwritebytesto requires a native array to read from");
     if (((MVMArrayREPRData *)STABLE(buffer)->REPR_data)->slot_type != MVM_ARRAY_U8
         && ((MVMArrayREPRData *)STABLE(buffer)->REPR_data)->slot_type != MVM_ARRAY_I8)
@@ -508,23 +400,14 @@ static MVMint64 close_socket(MVMThreadContext *tc, MVMOSHandle *h) {
     return 0;
 }
 
-static void gc_free(MVMThreadContext *tc, MVMObject *h, void *d) {
-    MVMIOAsyncUDPSocketData *data = (MVMIOAsyncUDPSocketData *)d;
-    if (data->ds) {
-        MVM_string_decodestream_destory(tc, data->ds);
-        data->ds = NULL;
-    }
-}
-
 /* IO ops table, populated with functions. */
 static const MVMIOClosable        closable          = { close_socket };
-static const MVMIOAsyncReadable   async_readable    = { read_chars, read_bytes };
-static const MVMIOAsyncWritableTo async_writable_to = { write_str_to, write_bytes_to };
+static const MVMIOAsyncReadable   async_readable    = { read_bytes };
+static const MVMIOAsyncWritableTo async_writable_to = { write_bytes_to };
 static const MVMIOOps op_table = {
     &closable,
     NULL,
     NULL,
-    NULL,
     &async_readable,
     NULL,
     &async_writable_to,
@@ -534,7 +417,7 @@ static const MVMIOOps op_table = {
     NULL,
     NULL,
     NULL,
-    gc_free
+    NULL
 };
 
 /* Info we convey about a socket setup task. */
