@@ -72,10 +72,16 @@ static uv_loop_t *get_or_vivify_loop(MVMThreadContext *tc) {
 
     if (!instance->event_loop_thread) {
         /* Grab starting mutex and ensure we didn't lose the race. */
+        MVM_telemetry_timestamp(tc, "hoping to start an event loop thread");
+        MVM_gc_mark_thread_blocked(tc);
         uv_mutex_lock(&instance->mutex_event_loop_start);
+        MVM_gc_mark_thread_unblocked(tc);
         if (!instance->event_loop_thread) {
             MVMObject *thread, *loop_runner;
             int r;
+            unsigned int interval_id;
+
+            interval_id = MVM_telemetry_interval_start(tc, "creating the event loop thread");
 
             /* Create various bits of state the async event loop thread needs. */
             instance->event_loop_todo_queue   = MVM_repr_alloc_init(tc,
@@ -102,12 +108,16 @@ static uv_loop_t *get_or_vivify_loop(MVMThreadContext *tc) {
                 MVM_thread_run(tc, thread);
 
                 /* Block until we know it's fully started and initialized. */
+                MVM_gc_mark_thread_blocked(tc);
                 uv_sem_wait(&(instance->sem_event_loop_started));
+                MVM_gc_mark_thread_unblocked(tc);
                 uv_sem_destroy(&(instance->sem_event_loop_started));
 
                 /* Make the started event loop thread visible to others. */
                 instance->event_loop_thread = ((MVMThread *)thread)->body.tc;
             });
+
+            MVM_telemetry_interval_stop(tc, interval_id, "created the event loop thread");
         }
         uv_mutex_unlock(&instance->mutex_event_loop_start);
     }
@@ -153,3 +163,38 @@ void MVM_io_eventloop_send_cancellation_notification(MVMThreadContext *tc, MVMAs
     if (notify_queue && notify_schedulee)
         MVM_repr_push_o(tc, notify_queue, notify_schedulee);
 }
+
+/* Adds a work item to the active async task set. */
+int MVM_io_eventloop_add_active_work(MVMThreadContext *tc, MVMObject *async_task) {
+    int work_idx = MVM_repr_elems(tc, tc->instance->event_loop_active);
+    MVM_repr_push_o(tc, tc->instance->event_loop_active, async_task);
+    return work_idx;
+}
+
+/* Gets an active work item from the active work eventloop. */
+MVMAsyncTask * MVM_io_eventloop_get_active_work(MVMThreadContext *tc, int work_idx) {
+    if (work_idx >= 0 && work_idx < MVM_repr_elems(tc, tc->instance->event_loop_active)) {
+        MVMObject *task_obj = MVM_repr_at_pos_o(tc, tc->instance->event_loop_active, work_idx);
+        if (REPR(task_obj)->ID != MVM_REPR_ID_MVMAsyncTask)
+            MVM_panic(1, "non-AsyncTask fetched from eventloop active work list");
+        return (MVMAsyncTask *)task_obj;
+    }
+    else {
+        MVM_panic(1, "use of invalid eventloop work item index %d", work_idx);
+    }
+}
+
+/* Removes an active work index from the active work list, enabling any
+ * memory associated with it to be collected. Replaces the work index with -1
+ * so that any future use of the task will be a failed lookup. */
+void MVM_io_eventloop_remove_active_work(MVMThreadContext *tc, int *work_idx_to_clear) {
+    int work_idx = *work_idx_to_clear;
+    if (work_idx >= 0 && work_idx < MVM_repr_elems(tc, tc->instance->event_loop_active)) {
+        *work_idx_to_clear = -1;
+        MVM_repr_bind_pos_o(tc, tc->instance->event_loop_active, work_idx, tc->instance->VMNull);
+        /* TODO: start to re-use the indices */
+    }
+    else {
+        MVM_panic(1, "cannot remove invalid eventloop work item index %d", work_idx);
+    }
+}
