@@ -12,8 +12,7 @@ static void on_changed(uv_fs_event_t *handle, const char *filename, int events,
     WatchInfo        *wi  = (WatchInfo *)handle->data;
     MVMThreadContext *tc  = wi->tc;
     MVMObject        *arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
-    MVMAsyncTask     *t   = (MVMAsyncTask *)MVM_repr_at_pos_o(tc,
-        tc->instance->event_loop_active, wi->work_idx);
+    MVMAsyncTask     *t   = MVM_io_eventloop_get_active_work(tc, wi->work_idx);
     MVM_repr_push_o(tc, arr, t->body.schedulee);
     MVMROOT(tc, t, {
     MVMROOT(tc, arr, {
@@ -46,10 +45,9 @@ static void setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_task,
     int        r;
 
     /* Add task to active list. */
-    wi->work_idx    = MVM_repr_elems(tc, tc->instance->event_loop_active);
+    wi->work_idx    = MVM_io_eventloop_add_active_work(tc, async_task);
     wi->tc          = tc;
     wi->handle.data = wi;
-    MVM_repr_push_o(tc, tc->instance->event_loop_active, async_task);
 
     /* Start watching. */
     uv_fs_event_init(loop, &wi->handle);
