@@ -5,7 +5,7 @@
 
 static MVMOSHandle * verify_is_handle(MVMThreadContext *tc, MVMObject *oshandle, const char *op) {
     if (REPR(oshandle)->ID != MVM_REPR_ID_MVMOSHandle)
-        MVM_exception_throw_adhoc(tc, "%s requires an object with REPR MVMOSHandle", op);
+        MVM_exception_throw_adhoc(tc, "%s requires an object with REPR MVMOSHandle (got %s with REPR %s)", op, STABLE(oshandle)->debug_name, REPR(oshandle)->name);
     return (MVMOSHandle *)oshandle;
 }
 
@@ -35,7 +35,8 @@ MVMint64 MVM_io_close(MVMThreadContext *tc, MVMObject *oshandle) {
 
 MVMint64 MVM_io_is_tty(MVMThreadContext *tc, MVMObject *oshandle) {
     MVMOSHandle *handle = verify_is_handle(tc, oshandle, "istty");
-    if (handle->body.ops->introspection) {
+    /* We need the extra check on is_tty because it is NULL for pipes. */
+    if (handle->body.ops->introspection && handle->body.ops->introspection->is_tty) {
         uv_mutex_t *mutex = acquire_mutex(tc, handle);
         MVMint64 ret = handle->body.ops->introspection->is_tty(tc, handle);
         release_mutex(tc, mutex);
@@ -59,20 +60,6 @@ MVMint64 MVM_io_fileno(MVMThreadContext *tc, MVMObject *oshandle) {
     }
 }
 
-void MVM_io_set_encoding(MVMThreadContext *tc, MVMObject *oshandle, MVMString *encoding_name) {
-    MVMOSHandle *handle = verify_is_handle(tc, oshandle, "set encoding");
-    MVMROOT(tc, handle, {
-        const MVMuint8 encoding_flag = MVM_string_find_encoding(tc, encoding_name);
-        if (handle->body.ops->encodable) {
-            uv_mutex_t *mutex = acquire_mutex(tc, handle);
-            handle->body.ops->encodable->set_encoding(tc, handle, encoding_flag);
-            release_mutex(tc, mutex);
-        }
-        else
-            MVM_exception_throw_adhoc(tc, "Cannot set encoding on this kind of handle");
-    });
-}
-
 void MVM_io_seek(MVMThreadContext *tc, MVMObject *oshandle, MVMint64 offset, MVMint64 flag) {
     MVMOSHandle *handle = verify_is_handle(tc, oshandle, "seek");
     if (handle->body.ops->seekable) {
@@ -96,92 +83,29 @@ MVMint64 MVM_io_tell(MVMThreadContext *tc, MVMObject *oshandle) {
         MVM_exception_throw_adhoc(tc, "Cannot tell this kind of handle");
 }
 
-void MVM_io_set_separator(MVMThreadContext *tc, MVMObject *oshandle, MVMString *sep) {
-    MVMOSHandle *handle = verify_is_handle(tc, oshandle, "set separator");
-    if (handle->body.ops->sync_readable) {
-        uv_mutex_t *mutex = acquire_mutex(tc, handle);
-        handle->body.ops->sync_readable->set_separator(tc, handle, &sep, 1);
-        release_mutex(tc, mutex);
-    }
-    else
-        MVM_exception_throw_adhoc(tc, "Cannot set a separator on this kind of handle");
-}
-
-void MVM_io_set_separators(MVMThreadContext *tc, MVMObject *oshandle, MVMObject *seps) {
-    MVMOSHandle *handle = verify_is_handle(tc, oshandle, "set separators");
-    if (handle->body.ops->sync_readable) {
-        MVMint32 is_str_array = REPR(seps)->pos_funcs.get_elem_storage_spec(tc,
-            STABLE(seps)).boxed_primitive == MVM_STORAGE_SPEC_BP_STR;
-        if (is_str_array) {
-            uv_mutex_t *mutex;
-            MVMString **c_seps;
-            MVMuint64 i;
-
-            MVMuint64 num_seps = MVM_repr_elems(tc, seps);
-            if (num_seps > 0xFFFFFF)
-                MVM_exception_throw_adhoc(tc, "Too many line separators");
-            c_seps = MVM_malloc((num_seps ? num_seps : 1) * sizeof(MVMString *));
-            for (i = 0; i < num_seps; i++)
-                c_seps[i] = MVM_repr_at_pos_s(tc, seps, i);
-
-            mutex = acquire_mutex(tc, handle);
-            handle->body.ops->sync_readable->set_separator(tc, handle, c_seps, (MVMint32)num_seps);
-            release_mutex(tc, mutex);
-
-            MVM_free(c_seps);
-        }
-        else {
-            MVM_exception_throw_adhoc(tc, "Set separators requires a native string array");
-        }
-    }
-    else {
-        MVM_exception_throw_adhoc(tc, "Cannot set separators on this kind of handle");
-    }
-}
-
-MVMString * MVM_io_readline(MVMThreadContext *tc, MVMObject *oshandle, MVMint32 chomp) {
-    MVMOSHandle *handle = verify_is_handle(tc, oshandle, "readline");
-    if (handle->body.ops->sync_readable) {
-        uv_mutex_t *mutex = acquire_mutex(tc, handle);
-        MVMString *result = handle->body.ops->sync_readable->read_line(tc, handle, chomp);
-        release_mutex(tc, mutex);
-        return result;
-    }
-    else
-        MVM_exception_throw_adhoc(tc, "Cannot read lines from this kind of handle");
-}
-
-MVMString * MVM_io_read_string(MVMThreadContext *tc, MVMObject *oshandle, MVMint64 chars) {
-    MVMOSHandle *handle = verify_is_handle(tc, oshandle, "read string");
-    if (handle->body.ops->sync_readable) {
-        uv_mutex_t *mutex = acquire_mutex(tc, handle);
-        MVMString *result = handle->body.ops->sync_readable->read_chars(tc, handle, chars);
-        release_mutex(tc, mutex);
-        return result;
-    }
-    else
-        MVM_exception_throw_adhoc(tc, "Cannot read characters from this kind of handle");
-}
-
 void MVM_io_read_bytes(MVMThreadContext *tc, MVMObject *oshandle, MVMObject *result, MVMint64 length) {
     MVMOSHandle *handle = verify_is_handle(tc, oshandle, "read bytes");
     MVMint64 bytes_read;
     char *buf;
 
     /* Ensure the target is in the correct form. */
-    if (!IS_CONCRETE(result) || REPR(result)->ID != MVM_REPR_ID_MVMArray)
+    if (!IS_CONCRETE(result) || REPR(result)->ID != MVM_REPR_ID_VMArray)
         MVM_exception_throw_adhoc(tc, "read_fhb requires a native array to write to");
     if (((MVMArrayREPRData *)STABLE(result)->REPR_data)->slot_type != MVM_ARRAY_U8
         && ((MVMArrayREPRData *)STABLE(result)->REPR_data)->slot_type != MVM_ARRAY_I8)
         MVM_exception_throw_adhoc(tc, "read_fhb requires a native array of uint8 or int8");
 
-    if (length < 1 || length > 99999999)
+    if (length < 1)
         MVM_exception_throw_adhoc(tc, "Out of range: attempted to read %"PRId64" bytes from filehandle", length);
 
     if (handle->body.ops->sync_readable) {
-        uv_mutex_t *mutex = acquire_mutex(tc, handle);
-        bytes_read = handle->body.ops->sync_readable->read_bytes(tc, handle, &buf, length);
-        release_mutex(tc, mutex);
+        MVMROOT(tc, handle, {
+        MVMROOT(tc, result, {
+            uv_mutex_t *mutex = acquire_mutex(tc, handle);
+            bytes_read = handle->body.ops->sync_readable->read_bytes(tc, handle, &buf, length);
+            release_mutex(tc, mutex);
+        });
+        });
     }
     else
         MVM_exception_throw_adhoc(tc, "Cannot read characters from this kind of handle");
@@ -193,39 +117,13 @@ void MVM_io_read_bytes(MVMThreadContext *tc, MVMObject *oshandle, MVMObject *res
     ((MVMArray *)result)->body.elems    = bytes_read;
 }
 
-MVMString * MVM_io_slurp(MVMThreadContext *tc, MVMObject *oshandle) {
-    MVMOSHandle *handle = verify_is_handle(tc, oshandle, "slurp");
-    if (handle->body.ops->sync_readable) {
-        uv_mutex_t *mutex = acquire_mutex(tc, handle);
-        MVMString *result = handle->body.ops->sync_readable->slurp(tc, handle);
-        release_mutex(tc, mutex);
-        return result;
-    }
-    else
-        MVM_exception_throw_adhoc(tc, "Cannot slurp this kind of handle");
-}
-
-MVMint64 MVM_io_write_string(MVMThreadContext *tc, MVMObject *oshandle, MVMString *str, MVMint8 addnl) {
-    MVMOSHandle *handle = verify_is_handle(tc, oshandle, "write string");
-    if (str == NULL)
-        MVM_exception_throw_adhoc(tc, "Failed to write to filehandle: NULL string given");
-    if (handle->body.ops->sync_writable) {
-        uv_mutex_t *mutex = acquire_mutex(tc, handle);
-        MVMint64 result = handle->body.ops->sync_writable->write_str(tc, handle, str, addnl);
-        release_mutex(tc, mutex);
-        return result;
-    }
-    else
-        MVM_exception_throw_adhoc(tc, "Cannot write a string to this kind of handle");
-}
-
 void MVM_io_write_bytes(MVMThreadContext *tc, MVMObject *oshandle, MVMObject *buffer) {
     MVMOSHandle *handle = verify_is_handle(tc, oshandle, "write bytes");
     char *output;
     MVMint64 output_size;
 
     /* Ensure the target is in the correct form. */
-    if (!IS_CONCRETE(buffer) || REPR(buffer)->ID != MVM_REPR_ID_MVMArray)
+    if (!IS_CONCRETE(buffer) || REPR(buffer)->ID != MVM_REPR_ID_VMArray)
         MVM_exception_throw_adhoc(tc, "write_fhb requires a native array to read from");
     if (((MVMArrayREPRData *)STABLE(buffer)->REPR_data)->slot_type != MVM_ARRAY_U8
         && ((MVMArrayREPRData *)STABLE(buffer)->REPR_data)->slot_type != MVM_ARRAY_I8)
@@ -243,18 +141,16 @@ void MVM_io_write_bytes(MVMThreadContext *tc, MVMObject *oshandle, MVMObject *bu
         MVM_exception_throw_adhoc(tc, "Cannot write bytes to this kind of handle");
 }
 
-MVMObject * MVM_io_read_chars_async(MVMThreadContext *tc, MVMObject *oshandle, MVMObject *queue,
-                                    MVMObject *schedulee, MVMObject *async_type) {
-    MVMOSHandle *handle = verify_is_handle(tc, oshandle, "read chars asynchronously");
-    if (handle->body.ops->async_readable) {
+void MVM_io_write_bytes_c(MVMThreadContext *tc, MVMObject *oshandle, char *output,
+                          MVMuint64 output_size) {
+    MVMOSHandle *handle = verify_is_handle(tc, oshandle, "write bytes");
+    if (handle->body.ops->sync_writable) {
         uv_mutex_t *mutex = acquire_mutex(tc, handle);
-        MVMObject *result = (MVMObject *)handle->body.ops->async_readable->read_chars(tc,
-            handle, queue, schedulee, async_type);
+        handle->body.ops->sync_writable->write_bytes(tc, handle, output, output_size);
         release_mutex(tc, mutex);
-        return result;
     }
     else
-        MVM_exception_throw_adhoc(tc, "Cannot read chars asynchronously from this kind of handle");
+        MVM_exception_throw_adhoc(tc, "Cannot write bytes to this kind of handle");
 }
 
 MVMObject * MVM_io_read_bytes_async(MVMThreadContext *tc, MVMObject *oshandle, MVMObject *queue,
@@ -271,22 +167,6 @@ MVMObject * MVM_io_read_bytes_async(MVMThreadContext *tc, MVMObject *oshandle, M
         MVM_exception_throw_adhoc(tc, "Cannot read bytes asynchronously from this kind of handle");
 }
 
-MVMObject * MVM_io_write_string_async(MVMThreadContext *tc, MVMObject *oshandle, MVMObject *queue,
-                                      MVMObject *schedulee, MVMString *str, MVMObject *async_type) {
-    MVMOSHandle *handle = verify_is_handle(tc, oshandle, "write string asynchronously");
-    if (str == NULL)
-        MVM_exception_throw_adhoc(tc, "Failed to write to filehandle: NULL string given");
-    if (handle->body.ops->async_writable) {
-        uv_mutex_t *mutex = acquire_mutex(tc, handle);
-        MVMObject *result = (MVMObject *)handle->body.ops->async_writable->write_str(tc,
-            handle, queue, schedulee, str, async_type);
-        release_mutex(tc, mutex);
-        return result;
-    }
-    else
-        MVM_exception_throw_adhoc(tc, "Cannot write a string asynchronously to this kind of handle");
-}
-
 MVMObject * MVM_io_write_bytes_async(MVMThreadContext *tc, MVMObject *oshandle, MVMObject *queue,
                                      MVMObject *schedulee, MVMObject *buffer, MVMObject *async_type) {
     MVMOSHandle *handle = verify_is_handle(tc, oshandle, "write buffer asynchronously");
@@ -303,23 +183,6 @@ MVMObject * MVM_io_write_bytes_async(MVMThreadContext *tc, MVMObject *oshandle,
         MVM_exception_throw_adhoc(tc, "Cannot write bytes asynchronously to this kind of handle");
 }
 
-MVMObject * MVM_io_write_string_to_async(MVMThreadContext *tc, MVMObject *oshandle, MVMObject *queue,
-                                         MVMObject *schedulee, MVMString *str, MVMObject *async_type,
-                                         MVMString *host, MVMint64 port) {
-    MVMOSHandle *handle = verify_is_handle(tc, oshandle, "write string asynchronously to destination");
-    if (str == NULL)
-        MVM_exception_throw_adhoc(tc, "Failed to write to filehandle: NULL string given");
-    if (handle->body.ops->async_writable_to) {
-        uv_mutex_t *mutex = acquire_mutex(tc, handle);
-        MVMObject *result = (MVMObject *)handle->body.ops->async_writable_to->write_str_to(tc,
-            handle, queue, schedulee, str, async_type, host, port);
-        release_mutex(tc, mutex);
-        return result;
-    }
-    else
-        MVM_exception_throw_adhoc(tc, "Cannot write a string to a destination asynchronously to this kind of handle");
-}
-
 MVMObject * MVM_io_write_bytes_to_async(MVMThreadContext *tc, MVMObject *oshandle, MVMObject *queue,
                                         MVMObject *schedulee, MVMObject *buffer, MVMObject *async_type,
                                         MVMString *host, MVMint64 port) {
@@ -416,6 +279,18 @@ void MVM_io_bind(MVMThreadContext *tc, MVMObject *oshandle, MVMString *host, MVM
         MVM_exception_throw_adhoc(tc, "Cannot bind this kind of handle");
 }
 
+MVMint64 MVM_io_getport(MVMThreadContext *tc, MVMObject *oshandle) {
+    MVMOSHandle *handle = verify_is_handle(tc, oshandle, "getport");
+    if (handle->body.ops->sockety) {
+        uv_mutex_t *mutex = acquire_mutex(tc, handle);
+        MVMint64 result = handle->body.ops->sockety->getport(tc, handle);
+        release_mutex(tc, mutex);
+        return result;
+    }
+    else
+        MVM_exception_throw_adhoc(tc, "Cannot getport for this kind of handle");
+}
+
 MVMObject * MVM_io_accept(MVMThreadContext *tc, MVMObject *oshandle) {
     MVMOSHandle *handle = verify_is_handle(tc, oshandle, "accept");
     if (handle->body.ops->sockety) {
