@@ -16,7 +16,7 @@
 extern char **environ;
 #  endif
 #else
-#  include <process.h>
+#include <stdlib.h>
 #endif
 
 #ifdef _WIN32
@@ -33,7 +33,7 @@ static wchar_t * ANSIToUnicode(MVMuint16 acp, const char *str)
 static char * UnicodeToUTF8(const wchar_t *str)
 {
      const int       len = WideCharToMultiByte(CP_UTF8, 0, str, -1, NULL, 0, NULL, NULL);
-     char * const result = (char *)MVM_malloc(len * sizeof(char));
+     char * const result = (char *)MVM_malloc(len + 1);
 
      WideCharToMultiByte(CP_UTF8, 0, str, -1, result, len, NULL, NULL);
 
@@ -49,29 +49,45 @@ static char * ANSIToUTF8(MVMuint16 acp, const char * str)
     return result;
 }
 
+MVM_PUBLIC char **
+MVM_UnicodeToUTF8_argv(const int argc, wchar_t **wargv)
+{
+    int i;
+    char **argv = MVM_malloc((argc + 1) * sizeof(*argv));
+    for (i = 0; i < argc; ++i)
+    {
+        argv[i] = UnicodeToUTF8(wargv[i]);
+    }
+    argv[i] = NULL;
+    return argv;
+}
+
 #endif
 
 MVMObject * MVM_proc_getenvhash(MVMThreadContext *tc) {
     MVMInstance * const instance = tc->instance;
     MVMObject   *       env_hash;
 
-#ifdef _WIN32
-    const MVMuint16 acp = GetACP(); /* We should get ACP at runtime. */
-#endif
     MVMuint32     pos = 0;
     MVMString *needle = MVM_string_ascii_decode(tc, instance->VMString, STR_WITH_LEN("="));
+#ifndef _WIN32
     char      *env;
+#else
+    wchar_t   *env;
+    (void) _wgetenv(L"windows"); /* populate _wenviron */
+#endif
 
     MVM_gc_root_temp_push(tc, (MVMCollectable **)&needle);
 
     env_hash = MVM_repr_alloc_init(tc,  MVM_hll_current(tc)->slurpy_hash_type);
     MVM_gc_root_temp_push(tc, (MVMCollectable **)&env_hash);
 
-    while ((env = environ[pos++]) != NULL) {
 #ifndef _WIN32
+    while ((env = environ[pos++]) != NULL) {
         MVMString    *str = MVM_string_utf8_c8_decode(tc, instance->VMString, env, strlen(env));
 #else
-        char * const _env = ANSIToUTF8(acp, env);
+    while ((env = _wenviron[pos++]) != NULL) {
+        char * const _env = UnicodeToUTF8(env);
         MVMString    *str = MVM_string_utf8_c8_decode(tc, instance->VMString, _env, strlen(_env));
 #endif
 
@@ -139,7 +155,7 @@ static void setup_process_stdio(MVMThreadContext *tc, MVMObject *handle, uv_proc
         MVMIOSyncPipeData *pipedata;
 
         if (REPR(handle)->ID != MVM_REPR_ID_MVMOSHandle)
-            MVM_exception_throw_adhoc(tc, "%s requires an object with REPR MVMOSHandle", op);
+            MVM_exception_throw_adhoc(tc, "%s requires an object with REPR MVMOSHandle (got %s with REPR %s)", op, STABLE(handle)->debug_name, REPR(handle)->name);
 
         pipedata           = (MVMIOSyncPipeData *)((MVMOSHandle *)handle)->body.data;
         pipedata->process  = process;
@@ -156,9 +172,9 @@ static void setup_process_stdio(MVMThreadContext *tc, MVMObject *handle, uv_proc
             MVMOSHandleBody body = ((MVMOSHandle *)handle)->body;
 
             if (REPR(handle)->ID != MVM_REPR_ID_MVMOSHandle)
-                MVM_exception_throw_adhoc(tc, "%s requires an object with REPR MVMOSHandle", op);
+                MVM_exception_throw_adhoc(tc, "%s requires an object with REPR MVMOSHandle (got %s with REPR %s)", op, STABLE(handle)->debug_name, REPR(handle)->name);
 
-            body.ops->pipeable->bind_stdio_handle(tc, ((MVMOSHandle *)handle), stdio, process);
+            body.ops->pipeable->bind_stdio_handle(tc, ((MVMOSHandle *)handle), stdio);
         }
     }
     else
@@ -361,18 +377,17 @@ typedef struct {
     char              *cwd;
     char             **env;
     char             **args;
-    MVMDecodeStream   *ds_stdout;
-    MVMDecodeStream   *ds_stderr;
     MVMuint32          seq_stdout;
     MVMuint32          seq_stderr;
+    MVMuint32          seq_merge;
     uv_stream_t       *stdin_handle;
-    ProcessState      state;
+    ProcessState       state;
+    int                using;
 } SpawnInfo;
 
 /* Info we convey about a write task. */
 typedef struct {
     MVMOSHandle      *handle;
-    MVMString        *str_data;
     MVMObject        *buf_data;
     uv_write_t       *req;
     uv_buf_t          buf;
@@ -385,8 +400,7 @@ static void on_write(uv_write_t *req, int status) {
     SpawnWriteInfo   *wi  = (SpawnWriteInfo *)req->data;
     MVMThreadContext *tc  = wi->tc;
     MVMObject        *arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
-    MVMAsyncTask     *t   = (MVMAsyncTask *)MVM_repr_at_pos_o(tc,
-        tc->instance->event_loop_active, wi->work_idx);
+    MVMAsyncTask     *t   = MVM_io_eventloop_get_active_work(tc, wi->work_idx);
     MVM_repr_push_o(tc, arr, t->body.schedulee);
     if (status >= 0) {
         MVMROOT(tc, arr, {
@@ -412,8 +426,7 @@ static void on_write(uv_write_t *req, int status) {
         });
     }
     MVM_repr_push_o(tc, t->body.queue, arr);
-    if (wi->str_data)
-        MVM_free(wi->buf.base);
+    MVM_io_eventloop_remove_active_work(tc, &(wi->work_idx));
     MVM_free(wi->req);
 }
 
@@ -421,6 +434,7 @@ static void on_write(uv_write_t *req, int status) {
 static void write_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_task, void *data) {
     MVMIOAsyncProcessData *handle_data;
     MVMAsyncTask          *spawn_task;
+    MVMArray              *buffer;
     SpawnInfo             *si;
     char                  *output;
     int                    output_size, r;
@@ -428,20 +442,12 @@ static void write_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
     /* Add to work in progress. */
     SpawnWriteInfo *wi = (SpawnWriteInfo *)data;
     wi->tc             = tc;
-    wi->work_idx       = MVM_repr_elems(tc, tc->instance->event_loop_active);
-    MVM_repr_push_o(tc, tc->instance->event_loop_active, async_task);
+    wi->work_idx       = MVM_io_eventloop_add_active_work(tc, async_task);
 
-    /* Encode the string, or extract buf data. */
-    if (wi->str_data) {
-        MVMuint64 output_size_64;
-        output = MVM_string_utf8_encode(tc, wi->str_data, &output_size_64, 1);
-        output_size = (int)output_size_64;
-    }
-    else {
-        MVMArray *buffer = (MVMArray *)wi->buf_data;
-        output = (char *)(buffer->body.slots.i8 + buffer->body.start);
-        output_size = (int)buffer->body.elems;
-    }
+    /* Extract buf data. */
+    buffer = (MVMArray *)wi->buf_data;
+    output = (char *)(buffer->body.slots.i8 + buffer->body.start);
+    output_size = (int)buffer->body.elems;
 
     /* Create and initialize write request. */
     wi->req           = MVM_malloc(sizeof(uv_write_t));
@@ -479,7 +485,6 @@ static void write_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
 static void write_gc_mark(MVMThreadContext *tc, void *data, MVMGCWorklist *worklist) {
     SpawnWriteInfo *wi = (SpawnWriteInfo *)data;
     MVM_gc_worklist_add(tc, worklist, &wi->handle);
-    MVM_gc_worklist_add(tc, worklist, &wi->str_data);
     MVM_gc_worklist_add(tc, worklist, &wi->buf_data);
 }
 
@@ -497,45 +502,6 @@ static const MVMAsyncTaskOps write_op_table = {
     write_gc_free
 };
 
-static MVMAsyncTask * write_str(MVMThreadContext *tc, MVMOSHandle *h, MVMObject *queue,
-                                MVMObject *schedulee, MVMString *s, MVMObject *async_type) {
-    MVMAsyncTask *task;
-    SpawnWriteInfo    *wi;
-
-    /* Validate REPRs. */
-    if (REPR(queue)->ID != MVM_REPR_ID_ConcBlockingQueue)
-        MVM_exception_throw_adhoc(tc,
-            "asyncwritestr target queue must have ConcBlockingQueue REPR");
-    if (REPR(async_type)->ID != MVM_REPR_ID_MVMAsyncTask)
-        MVM_exception_throw_adhoc(tc,
-            "asyncwritestr result type must have REPR AsyncTask");
-
-    /* Create async task handle. */
-    MVMROOT(tc, queue, {
-    MVMROOT(tc, schedulee, {
-    MVMROOT(tc, h, {
-    MVMROOT(tc, s, {
-        task = (MVMAsyncTask *)MVM_repr_alloc_init(tc, async_type);
-    });
-    });
-    });
-    });
-    MVM_ASSIGN_REF(tc, &(task->common.header), task->body.queue, queue);
-    MVM_ASSIGN_REF(tc, &(task->common.header), task->body.schedulee, schedulee);
-    task->body.ops  = &write_op_table;
-    wi              = MVM_calloc(1, sizeof(SpawnWriteInfo));
-    MVM_ASSIGN_REF(tc, &(task->common.header), wi->handle, h);
-    MVM_ASSIGN_REF(tc, &(task->common.header), wi->str_data, s);
-    task->body.data = wi;
-
-    /* Hand the task off to the event loop. */
-    MVMROOT(tc, task, {
-        MVM_io_eventloop_queue_work(tc, (MVMObject *)task);
-    });
-
-    return task;
-}
-
 static MVMAsyncTask * write_bytes(MVMThreadContext *tc, MVMOSHandle *h, MVMObject *queue,
                                   MVMObject *schedulee, MVMObject *buffer, MVMObject *async_type) {
     MVMAsyncTask *task;
@@ -548,7 +514,7 @@ static MVMAsyncTask * write_bytes(MVMThreadContext *tc, MVMOSHandle *h, MVMObjec
     if (REPR(async_type)->ID != MVM_REPR_ID_MVMAsyncTask)
         MVM_exception_throw_adhoc(tc,
             "asyncwritebytes result type must have REPR AsyncTask");
-    if (!IS_CONCRETE(buffer) || REPR(buffer)->ID != MVM_REPR_ID_MVMArray)
+    if (!IS_CONCRETE(buffer) || REPR(buffer)->ID != MVM_REPR_ID_VMArray)
         MVM_exception_throw_adhoc(tc, "asyncwritebytes requires a native array to read from");
     if (((MVMArrayREPRData *)STABLE(buffer)->REPR_data)->slot_type != MVM_ARRAY_U8
         && ((MVMArrayREPRData *)STABLE(buffer)->REPR_data)->slot_type != MVM_ARRAY_I8)
@@ -662,14 +628,13 @@ static void deferred_close_perform(MVMThreadContext *tc, uv_loop_t *loop, MVMObj
 }
 
 /* IO ops table, for async process, populated with functions. */
-static const MVMIOAsyncWritable proc_async_writable = { write_str, write_bytes };
+static const MVMIOAsyncWritable proc_async_writable = { write_bytes };
 static const MVMIOClosable      closable            = { close_stdin };
 static const MVMIOOps proc_op_table = {
     &closable,
     NULL,
     NULL,
     NULL,
-    NULL,
     &proc_async_writable,
     NULL,
     NULL,
@@ -699,8 +664,7 @@ static void async_spawn_on_exit(uv_process_t *req, MVMint64 exit_status, int ter
 
             /* Get what we'll need to build and convey the result. */
             MVMObject        *arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
-            MVMAsyncTask     *t   = (MVMAsyncTask *)MVM_repr_at_pos_o(tc,
-                tc->instance->event_loop_active, si->work_idx);
+            MVMAsyncTask     *t   = MVM_io_eventloop_get_active_work(tc, si->work_idx);
 
             /* Box and send along status. */
             MVM_repr_push_o(tc, arr, done_cb);
@@ -725,6 +689,8 @@ static void async_spawn_on_exit(uv_process_t *req, MVMint64 exit_status, int ter
     /* Close handle. */
     uv_close((uv_handle_t *)req, spawn_async_close);
     ((MVMIOAsyncProcessData *)((MVMOSHandle *)si->handle)->body.data)->handle = NULL;
+    if (--si->using == 0)
+        MVM_io_eventloop_remove_active_work(tc, &(si->work_idx));
 }
 
 /* Allocates a buffer of the suggested size. */
@@ -734,15 +700,15 @@ static void on_alloc(uv_handle_t *handle, size_t suggested_size, uv_buf_t *buf)
     buf->len    = size;
 }
 
-/* Read functions for stdout/stderr. */
+/* Read functions for stdout/stderr/merged. */
 static void async_read(uv_stream_t *handle, ssize_t nread, const uv_buf_t *buf, SpawnInfo *si,
-                       MVMObject *callback, MVMDecodeStream *ds, MVMuint32 seq_number) {
+                       MVMObject *callback, MVMuint32 seq_number) {
     MVMThreadContext *tc  = si->tc;
     MVMObject *arr;
     MVMAsyncTask *t;
     MVMROOT(tc, callback, {
         arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
-        t = (MVMAsyncTask *)MVM_repr_at_pos_o(tc, tc->instance->event_loop_active, si->work_idx);
+        t = MVM_io_eventloop_get_active_work(tc, si->work_idx);
     });
     MVM_repr_push_o(tc, arr, callback);
     if (nread >= 0) {
@@ -753,22 +719,14 @@ static void async_read(uv_stream_t *handle, ssize_t nread, const uv_buf_t *buf,
                 tc->instance->boot_types.BOOTInt, seq_number);
             MVM_repr_push_o(tc, arr, seq_boxed);
 
-            /* Either need to produce a buffer or decode characters. */
-            if (ds) {
-                MVMString *str;
-                MVMObject *boxed_str;
-                MVM_string_decodestream_add_bytes(tc, ds, buf->base, nread);
-                str = MVM_string_decodestream_get_all(tc, ds);
-                boxed_str = MVM_repr_box_str(tc, tc->instance->boot_types.BOOTStr, str);
-                MVM_repr_push_o(tc, arr, boxed_str);
-            }
-            else {
+            /* Push buffer of data. */
+            {
                 MVMObject *buf_type    = MVM_repr_at_key_o(tc, si->callbacks,
                                             tc->instance->str_consts.buf_type);
                 MVMArray  *res_buf     = (MVMArray *)MVM_repr_alloc_init(tc, buf_type);
                 res_buf->body.slots.i8 = (MVMint8 *)buf->base;
                 res_buf->body.start    = 0;
-                res_buf->body.ssize    = nread;
+                res_buf->body.ssize    = buf->len;
                 res_buf->body.elems    = nread;
                 MVM_repr_push_o(tc, arr, (MVMObject *)res_buf);
             }
@@ -791,6 +749,8 @@ static void async_read(uv_stream_t *handle, ssize_t nread, const uv_buf_t *buf,
         if (buf->base)
             MVM_free(buf->base);
         uv_close((uv_handle_t *) handle, NULL);
+        if (--si->using == 0)
+            MVM_io_eventloop_remove_active_work(tc, &(si->work_idx));
     }
     else {
         MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTInt);
@@ -807,32 +767,28 @@ static void async_read(uv_stream_t *handle, ssize_t nread, const uv_buf_t *buf,
         if (buf->base)
             MVM_free(buf->base);
         uv_close((uv_handle_t *) handle, NULL);
+        if (--si->using == 0)
+            MVM_io_eventloop_remove_active_work(tc, &(si->work_idx));
     }
     MVM_repr_push_o(tc, t->body.queue, arr);
 }
-static void async_spawn_stdout_chars_read(uv_stream_t *handle, ssize_t nread, const uv_buf_t *buf) {
-    SpawnInfo *si = (SpawnInfo *)handle->data;
-    MVMObject *cb = MVM_repr_at_key_o(si->tc, si->callbacks,
-        si->tc->instance->str_consts.stdout_chars);
-    async_read(handle, nread, buf, si, cb, si->ds_stdout, si->seq_stdout++);
-}
 static void async_spawn_stdout_bytes_read(uv_stream_t *handle, ssize_t nread, const uv_buf_t *buf) {
     SpawnInfo *si = (SpawnInfo *)handle->data;
     MVMObject *cb = MVM_repr_at_key_o(si->tc, si->callbacks,
         si->tc->instance->str_consts.stdout_bytes);
-    async_read(handle, nread, buf, si, cb, NULL, si->seq_stdout++);
-}
-static void async_spawn_stderr_chars_read(uv_stream_t *handle, ssize_t nread, const uv_buf_t *buf) {
-    SpawnInfo *si = (SpawnInfo *)handle->data;
-    MVMObject *cb = MVM_repr_at_key_o(si->tc, si->callbacks,
-        si->tc->instance->str_consts.stderr_chars);
-    async_read(handle, nread, buf, si, cb, si->ds_stderr, si->seq_stderr++);
+    async_read(handle, nread, buf, si, cb, si->seq_stdout++);
 }
 static void async_spawn_stderr_bytes_read(uv_stream_t *handle, ssize_t nread, const uv_buf_t *buf) {
     SpawnInfo *si = (SpawnInfo *)handle->data;
     MVMObject *cb = MVM_repr_at_key_o(si->tc, si->callbacks,
         si->tc->instance->str_consts.stderr_bytes);
-    async_read(handle, nread, buf, si, cb, NULL, si->seq_stderr++);
+    async_read(handle, nread, buf, si, cb, si->seq_stderr++);
+}
+static void async_spawn_merge_bytes_read(uv_stream_t *handle, ssize_t nread, const uv_buf_t *buf) {
+    SpawnInfo *si = (SpawnInfo *)handle->data;
+    MVMObject *cb = MVM_repr_at_key_o(si->tc, si->callbacks,
+        si->tc->instance->str_consts.merge_bytes);
+    async_read(handle, nread, buf, si, cb, si->seq_merge++);
 }
 
 /* Actually spawns an async task. This runs in the event loop thread. */
@@ -850,8 +806,8 @@ static void spawn_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
     /* Add to work in progress. */
     SpawnInfo *si = (SpawnInfo *)data;
     si->tc        = tc;
-    si->work_idx  = MVM_repr_elems(tc, tc->instance->event_loop_active);
-    MVM_repr_push_o(tc, tc->instance->event_loop_active, async_task);
+    si->work_idx  = MVM_io_eventloop_add_active_work(tc, async_task);
+    si->using     = 1;
 
     /* Create input/output handles as needed. */
     if (MVM_repr_exists_key(tc, si->callbacks, tc->instance->str_consts.write)) {
@@ -862,55 +818,69 @@ static void spawn_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
         process_stdio[0].data.stream = (uv_stream_t *)pipe;
         si->stdin_handle             = (uv_stream_t *)pipe;
     }
+    else if (MVM_repr_exists_key(tc, si->callbacks, tc->instance->str_consts.stdin_fd)) {
+        process_stdio[0].flags   = UV_INHERIT_FD;
+        process_stdio[0].data.fd = (int)MVM_repr_get_int(tc,
+            MVM_repr_at_key_o(tc, si->callbacks, tc->instance->str_consts.stdin_fd));
+    }
     else {
         process_stdio[0].flags   = UV_INHERIT_FD;
         process_stdio[0].data.fd = 0;
     }
-    if (MVM_repr_exists_key(tc, si->callbacks, tc->instance->str_consts.stdout_chars)) {
-        uv_pipe_t *pipe = MVM_malloc(sizeof(uv_pipe_t));
-        uv_pipe_init(tc->loop, pipe, 0);
-        pipe->data = si;
+    if (MVM_repr_exists_key(tc, si->callbacks, tc->instance->str_consts.merge_bytes)) {
+        stdout_pipe = MVM_malloc(sizeof(uv_pipe_t));
+        uv_pipe_init(tc->loop, stdout_pipe, 0);
+        stdout_pipe->data = si;
         process_stdio[1].flags       = UV_CREATE_PIPE | UV_WRITABLE_PIPE;
-        process_stdio[1].data.stream = (uv_stream_t *)pipe;
-        si->ds_stdout                = MVM_string_decodestream_create(tc, MVM_encoding_type_utf8, 0, 1);
-        stdout_pipe                  = pipe;
-        stdout_cb                    = async_spawn_stdout_chars_read;
-    }
-    else if (MVM_repr_exists_key(tc, si->callbacks, tc->instance->str_consts.stdout_bytes)) {
-        uv_pipe_t *pipe = MVM_malloc(sizeof(uv_pipe_t));
-        uv_pipe_init(tc->loop, pipe, 0);
-        pipe->data = si;
-        process_stdio[1].flags       = UV_CREATE_PIPE | UV_WRITABLE_PIPE;
-        process_stdio[1].data.stream = (uv_stream_t *)pipe;
-        stdout_pipe                  = pipe;
-        stdout_cb                    = async_spawn_stdout_bytes_read;
+        process_stdio[1].data.stream = (uv_stream_t *)stdout_pipe;
+        stdout_cb                    = async_spawn_merge_bytes_read;
+        stderr_pipe = MVM_malloc(sizeof(uv_pipe_t));
+        uv_pipe_init(tc->loop, stderr_pipe, 0);
+        stderr_pipe->data = si;
+        process_stdio[2].flags       = UV_CREATE_PIPE | UV_WRITABLE_PIPE;
+        process_stdio[2].data.stream = (uv_stream_t *)stderr_pipe;
+        stderr_cb                    = async_spawn_merge_bytes_read;
+        si->using += 2;
     }
     else {
-        process_stdio[1].flags   = UV_INHERIT_FD;
-        process_stdio[1].data.fd = 1;
-    }
-    if (MVM_repr_exists_key(tc, si->callbacks, tc->instance->str_consts.stderr_chars)) {
-        uv_pipe_t *pipe = MVM_malloc(sizeof(uv_pipe_t));
-        uv_pipe_init(tc->loop, pipe, 0);
-        pipe->data = si;
-        process_stdio[2].flags       = UV_CREATE_PIPE | UV_WRITABLE_PIPE;
-        process_stdio[2].data.stream = (uv_stream_t *)pipe;
-        si->ds_stderr                = MVM_string_decodestream_create(tc, MVM_encoding_type_utf8, 0, 1);
-        stderr_pipe                  = pipe;
-        stderr_cb                    = async_spawn_stderr_chars_read;
-    }
-    else if (MVM_repr_exists_key(tc, si->callbacks, tc->instance->str_consts.stderr_bytes)) {
-        uv_pipe_t *pipe = MVM_malloc(sizeof(uv_pipe_t));
-        uv_pipe_init(tc->loop, pipe, 0);
-        pipe->data = si;
-        process_stdio[2].flags       = UV_CREATE_PIPE | UV_WRITABLE_PIPE;
-        process_stdio[2].data.stream = (uv_stream_t *)pipe;
-        stderr_pipe                  = pipe;
-        stderr_cb                    = async_spawn_stderr_bytes_read;
-    }
-    else {
-        process_stdio[2].flags   = UV_INHERIT_FD;
-        process_stdio[2].data.fd = 2;
+        if (MVM_repr_exists_key(tc, si->callbacks, tc->instance->str_consts.stdout_bytes)) {
+            uv_pipe_t *pipe = MVM_malloc(sizeof(uv_pipe_t));
+            uv_pipe_init(tc->loop, pipe, 0);
+            pipe->data = si;
+            process_stdio[1].flags       = UV_CREATE_PIPE | UV_WRITABLE_PIPE;
+            process_stdio[1].data.stream = (uv_stream_t *)pipe;
+            stdout_pipe                  = pipe;
+            stdout_cb                    = async_spawn_stdout_bytes_read;
+            si->using++;
+        }
+        else if (MVM_repr_exists_key(tc, si->callbacks, tc->instance->str_consts.stdout_fd)) {
+            process_stdio[1].flags   = UV_INHERIT_FD;
+            process_stdio[1].data.fd = (int)MVM_repr_get_int(tc,
+                MVM_repr_at_key_o(tc, si->callbacks, tc->instance->str_consts.stdout_fd));
+        }
+        else {
+            process_stdio[1].flags   = UV_INHERIT_FD;
+            process_stdio[1].data.fd = 1;
+        }
+        if (MVM_repr_exists_key(tc, si->callbacks, tc->instance->str_consts.stderr_bytes)) {
+            uv_pipe_t *pipe = MVM_malloc(sizeof(uv_pipe_t));
+            uv_pipe_init(tc->loop, pipe, 0);
+            pipe->data = si;
+            process_stdio[2].flags       = UV_CREATE_PIPE | UV_WRITABLE_PIPE;
+            process_stdio[2].data.stream = (uv_stream_t *)pipe;
+            stderr_pipe                  = pipe;
+            stderr_cb                    = async_spawn_stderr_bytes_read;
+            si->using++;
+        }
+        else if (MVM_repr_exists_key(tc, si->callbacks, tc->instance->str_consts.stderr_fd)) {
+            process_stdio[2].flags   = UV_INHERIT_FD;
+            process_stdio[2].data.fd = (int)MVM_repr_get_int(tc,
+                MVM_repr_at_key_o(tc, si->callbacks, tc->instance->str_consts.stderr_fd));
+        }
+        else {
+            process_stdio[2].flags   = UV_INHERIT_FD;
+            process_stdio[2].data.fd = 2;
+        }
     }
 
     /* Set up process start info. */
@@ -918,7 +888,7 @@ static void spawn_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
     process_options.file        = si->prog;
     process_options.args        = si->args;
     process_options.cwd         = si->cwd;
-    process_options.flags       = UV_PROCESS_WINDOWS_VERBATIM_ARGUMENTS | UV_PROCESS_WINDOWS_HIDE;
+    process_options.flags       = UV_PROCESS_WINDOWS_HIDE;
     process_options.env         = si->env;
     process_options.stdio_count = 3;
     process_options.exit_cb     = async_spawn_on_exit;
@@ -927,25 +897,51 @@ static void spawn_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
     process->data = si;
     spawn_result  = uv_spawn(tc->loop, process, &process_options);
     if (spawn_result) {
-        MVMObject *error_cb = MVM_repr_at_key_o(tc, si->callbacks,
-            tc->instance->str_consts.error);
+        MVMObject *msg_box = NULL;
         si->state = STATE_DONE;
-        if (!MVM_is_null(tc, error_cb)) {
-            MVMROOT(tc, error_cb, {
-            MVMROOT(tc, async_task, {
-                MVMObject *arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
-                MVM_repr_push_o(tc, arr, error_cb);
-                MVMROOT(tc, arr, {
-                    MVMString *msg_str = MVM_string_ascii_decode_nt(tc,
-                        tc->instance->VMString, uv_strerror(spawn_result));
-                    MVMObject *msg_box = MVM_repr_box_str(tc,
-                        tc->instance->boot_types.BOOTStr, msg_str);
+        MVMROOT(tc, async_task, {
+        MVMROOT(tc, msg_box, {
+            MVMString *msg_str = MVM_string_ascii_decode_nt(tc,
+                tc->instance->VMString, uv_strerror(spawn_result));
+            MVMObject *msg_box = MVM_repr_box_str(tc,
+                tc->instance->boot_types.BOOTStr, msg_str);
+
+            MVMObject *error_cb = MVM_repr_at_key_o(tc, si->callbacks,
+                tc->instance->str_consts.error);
+            if (!MVM_is_null(tc, error_cb)) {
+                MVMROOT(tc, error_cb, {
+                    MVMObject *arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
+                    MVM_repr_push_o(tc, arr, error_cb);
                     MVM_repr_push_o(tc, arr, msg_box);
+                    MVM_repr_push_o(tc, ((MVMAsyncTask *)async_task)->body.queue, arr);
                 });
+            }
+
+            if (stdout_pipe) {
+                MVMObject *arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
+                MVMObject *cb = MVM_repr_at_key_o(tc, si->callbacks,
+                    tc->instance->str_consts.stdout_bytes);
+                MVM_repr_push_o(tc, arr, cb);
+                MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTInt);
+                MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
+                MVM_repr_push_o(tc, arr, msg_box);
                 MVM_repr_push_o(tc, ((MVMAsyncTask *)async_task)->body.queue, arr);
-            });
-            });
-        }
+            }
+
+            if (stderr_pipe) {
+                MVMObject *arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
+                MVMObject *cb = MVM_repr_at_key_o(tc, si->callbacks,
+                    tc->instance->str_consts.stderr_bytes);
+                MVM_repr_push_o(tc, arr, cb);
+                MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTInt);
+                MVM_repr_push_o(tc, arr, tc->instance->boot_types.BOOTStr);
+                MVM_repr_push_o(tc, arr, msg_box);
+                MVM_repr_push_o(tc, ((MVMAsyncTask *)async_task)->body.queue, arr);
+            }
+        });
+        });
+
+        MVM_io_eventloop_remove_active_work(tc, &(si->work_idx));
     }
     else {
         MVMOSHandle           *handle  = (MVMOSHandle *)si->handle;
@@ -966,13 +962,13 @@ static void spawn_setup(MVMThreadContext *tc, uv_loop_t *loop, MVMObject *async_
             });
             });
         }
-    }
 
-    /* Start any output readers. */
-    if (stdout_pipe)
-        uv_read_start((uv_stream_t *)stdout_pipe, on_alloc, stdout_cb);
-    if (stderr_pipe)
-        uv_read_start((uv_stream_t *)stderr_pipe, on_alloc, stderr_cb);
+        /* Start any output readers. */
+        if (stdout_pipe)
+            uv_read_start((uv_stream_t *)stdout_pipe, on_alloc, stdout_cb);
+        if (stderr_pipe)
+            uv_read_start((uv_stream_t *)stderr_pipe, on_alloc, stderr_cb);
+    }
 }
 
 /* On cancel, kill the process. */
@@ -1023,14 +1019,6 @@ static void spawn_gc_free(MVMThreadContext *tc, MVMObject *t, void *data) {
             MVM_free(si->args);
             si->args = NULL;
         }
-        if (si->ds_stdout) {
-            MVM_string_decodestream_destory(tc, si->ds_stdout);
-            si->ds_stdout = NULL;
-        }
-        if (si->ds_stderr) {
-            MVM_string_decodestream_destory(tc, si->ds_stderr);
-            si->ds_stderr = NULL;
-        }
         MVM_free(si);
     }
 }
@@ -1217,7 +1205,6 @@ MVMObject * MVM_proc_clargs(MVMThreadContext *tc) {
         });
 #else
         MVMROOT(tc, clargs, {
-            const MVMuint16 acp = GetACP();
             const MVMint64 num_clargs = instance->num_clargs;
             MVMint64 count;
 
@@ -1230,10 +1217,8 @@ MVMObject * MVM_proc_clargs(MVMThreadContext *tc) {
 
             for (count = 0; count < num_clargs; count++) {
                 char *raw_clarg = instance->raw_clargs[count];
-                char * const _tmp = ANSIToUTF8(acp, raw_clarg);
                 MVMString *string = MVM_string_utf8_c8_decode(tc,
-                    instance->VMString, _tmp, strlen(_tmp));
-                MVM_free(_tmp);
+                    instance->VMString, raw_clarg, strlen(raw_clarg));
                 boxed_str = MVM_repr_box_str(tc,
                     instance->boot_types.BOOTStr, string);
                 MVM_repr_push_o(tc, clargs, boxed_str);
