@@ -16,7 +16,7 @@
         fprintf(stderr, "MoarVM: Initialization of " name " mutex failed\n    %s\n", \
             uv_strerror(init_stat)); \
         exit(1); \
-	} \
+    } \
 } while (0)
 
 static void setup_std_handles(MVMThreadContext *tc);
@@ -76,7 +76,8 @@ MVMInstance * MVM_vm_create_instance(void) {
     instance = MVM_calloc(1, sizeof(MVMInstance));
 
     /* Create the main thread's ThreadContext and stash it. */
-    instance->main_thread = MVM_tc_create(instance);
+    instance->main_thread = MVM_tc_create(NULL, instance);
+
     instance->main_thread->thread_id = 1;
 
     /* Next thread to be created gets ID 2 (the main thread got ID 1). */
@@ -128,6 +129,11 @@ MVMInstance * MVM_vm_create_instance(void) {
     instance->int_const_cache = MVM_calloc(1, sizeof(MVMIntConstCache));
     instance->int_to_str_cache = MVM_calloc(MVM_INT_TO_STR_CACHE_SIZE, sizeof(MVMString *));
 
+    /* Initialize Unicode database and NFG. */
+    MVM_unicode_init(instance->main_thread);
+    MVM_string_cclass_init(instance->main_thread);
+    MVM_nfg_init(instance->main_thread);
+
     /* Bootstrap 6model. It is assumed the GC will not be called during this. */
     MVM_6model_bootstrap(instance->main_thread);
 
@@ -161,12 +167,6 @@ MVMInstance * MVM_vm_create_instance(void) {
     /* Set up hll symbol tables mutex. */
     init_mutex(instance->mutex_hll_syms, "hll syms");
 
-    /* Initialize Unicode database */
-    MVM_unicode_init(instance->main_thread);
-
-    /* Initialize string cclass handling. */
-    MVM_string_cclass_init(instance->main_thread);
-
     /* Create callsite intern pool. */
     instance->callsite_interns = MVM_calloc(1, sizeof(MVMCallsiteInterns));
     init_mutex(instance->mutex_callsite_interns, "callsite interns");
@@ -233,7 +233,7 @@ MVMInstance * MVM_vm_create_instance(void) {
     dynvar_log = getenv("MVM_DYNVAR_LOG");
     if (dynvar_log && strlen(dynvar_log)) {
         instance->dynvar_log_fh = fopen_perhaps_with_pid(dynvar_log, "w");
-        fprintf(instance->dynvar_log_fh, "+ x 0 0 0 0 0 %llu\n", uv_hrtime());
+        fprintf(instance->dynvar_log_fh, "+ x 0 0 0 0 0 %"PRIu64"\n", uv_hrtime());
         fflush(instance->dynvar_log_fh);
         instance->dynvar_log_lasttime = uv_hrtime();
     }
@@ -252,9 +252,18 @@ MVMInstance * MVM_vm_create_instance(void) {
         instance->cross_thread_write_logging = 0;
     }
 
-    /* Set up NFG state mutation mutex. */
-    instance->nfg = calloc(1, sizeof(MVMNFGState));
-    init_mutex(instance->nfg->update_mutex, "NFG update mutex");
+    if (getenv("MVM_COVERAGE_LOG")) {
+        char *coverage_log = getenv("MVM_COVERAGE_LOG");
+        instance->coverage_logging = 1;
+        instance->instrumentation_level++;
+        if (strlen(coverage_log))
+            instance->coverage_log_fh = fopen_perhaps_with_pid(coverage_log, "a");
+        else
+            instance->coverage_log_fh = stderr;
+    }
+    else {
+        instance->coverage_logging = 0;
+    }
 
     /* Create std[in/out/err]. */
     setup_std_handles(instance->main_thread);
@@ -267,15 +276,15 @@ MVMInstance * MVM_vm_create_instance(void) {
 
 /* Set up some standard file handles. */
 static void setup_std_handles(MVMThreadContext *tc) {
-    tc->instance->stdin_handle  = MVM_file_get_stdstream(tc, 0, 1);
+    tc->instance->stdin_handle  = MVM_file_get_stdstream(tc, 0);
     MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&tc->instance->stdin_handle,
         "stdin handle");
 
-    tc->instance->stdout_handle = MVM_file_get_stdstream(tc, 1, 0);
+    tc->instance->stdout_handle = MVM_file_get_stdstream(tc, 1);
     MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&tc->instance->stdout_handle,
         "stdout handle");
 
-    tc->instance->stderr_handle = MVM_file_get_stdstream(tc, 2, 0);
+    tc->instance->stderr_handle = MVM_file_get_stdstream(tc, 2);
     MVM_gc_root_add_permanent_desc(tc, (MVMCollectable **)&tc->instance->stderr_handle,
         "stderr handle");
 }
@@ -350,7 +359,7 @@ void MVM_vm_exit(MVMInstance *instance) {
     if (instance->jit_bytecode_map)
         fclose(instance->jit_bytecode_map);
     if (instance->dynvar_log_fh) {
-        fprintf(instance->dynvar_log_fh, "- x 0 0 0 0 %lld %llu %llu\n", instance->dynvar_log_lasttime, uv_hrtime(), uv_hrtime());
+        fprintf(instance->dynvar_log_fh, "- x 0 0 0 0 %"PRId64" %"PRIu64" %"PRIu64"\n", instance->dynvar_log_lasttime, uv_hrtime(), uv_hrtime());
         fclose(instance->dynvar_log_fh);
     }
 
