@@ -136,6 +136,7 @@ MVM_PUBLIC const MVMint32 MVM_jit_support(void);
 #include "gc/roots.h"
 #include "gc/objectid.h"
 #include "gc/finalize.h"
+#include "core/regionalloc.h"
 #include "spesh/dump.h"
 #include "spesh/graph.h"
 #include "spesh/codegen.h"
@@ -150,13 +151,14 @@ MVM_PUBLIC const MVMint32 MVM_jit_support(void);
 #include "spesh/inline.h"
 #include "spesh/osr.h"
 #include "spesh/lookup.h"
+#include "strings/nfg.h"
 #include "strings/normalize.h"
 #include "strings/decode_stream.h"
 #include "strings/ascii.h"
+#include "strings/parse_num.h"
 #include "strings/utf8.h"
 #include "strings/utf8_c8.h"
 #include "strings/utf16.h"
-#include "strings/nfg.h"
 #include "strings/iter.h"
 #include "strings/ops.h"
 #include "strings/unicode_gen.h"
@@ -188,7 +190,9 @@ MVM_PUBLIC const MVMint32 MVM_jit_support(void);
 #include "profiler/log.h"
 #include "profiler/profile.h"
 #include "profiler/heapsnapshot.h"
+#include "profiler/telemeh.h"
 #include "instrument/crossthreadwrite.h"
+#include "instrument/line_coverage.h"
 
 MVMObject *MVM_backend_config(MVMThreadContext *tc);
 
