@@ -183,7 +183,7 @@ static MVMuint64 get_collectable_idx(MVMThreadContext *tc,
     return idx;
 }
 
-/* Gets the index of a frame, either returning an existing inde if we've seen
+/* Gets the index of a frame, either returning an existing index if we've seen
  * it before or adding it if not. */
 static MVMuint64 get_frame_idx(MVMThreadContext *tc, MVMHeapSnapshotState *ss,
         MVMFrame *frame) {
@@ -243,10 +243,13 @@ static void set_static_frame_index(MVMThreadContext *tc, MVMHeapSnapshotState *s
         s = &(ss->col->static_frames[i]);
         if (s->name == name_idx && s->cuid == cuid_idx && s->line == line && s->file == file_idx) {
             col->type_or_frame_index = i;
+            MVM_free(ann);
             return;
         }
     }
 
+    MVM_free(ann);
+
     grow_storage(&(ss->col->static_frames), &(ss->col->num_static_frames),
         &(ss->col->alloc_static_frames), sizeof(MVMHeapSnapshotStaticFrame));
     s = &(ss->col->static_frames[ss->col->num_static_frames]);
@@ -261,7 +264,7 @@ static void set_static_frame_index(MVMThreadContext *tc, MVMHeapSnapshotState *s
 /* Processes the work items, until we've none left. */
 static void process_collectable(MVMThreadContext *tc, MVMHeapSnapshotState *ss,
         MVMHeapSnapshotCollectable *col, MVMCollectable *c) {
-    MVMuint32 sc_idx = MVM_get_idx_of_sc(c);
+    MVMuint32 sc_idx = MVM_sc_get_idx_of_sc(c);
     if (sc_idx > 0)
         add_reference_const_cstr(tc, ss, "<SC>",
             get_collectable_idx(tc, ss,
@@ -270,14 +273,13 @@ static void process_collectable(MVMThreadContext *tc, MVMHeapSnapshotState *ss,
 }
 static void process_gc_worklist(MVMThreadContext *tc, MVMHeapSnapshotState *ss, char *desc) {
     MVMCollectable **c_ptr;
-    MVMFrame *f;
     MVMuint16 ref_kind = desc
         ? MVM_SNAPSHOT_REF_KIND_STRING
         : MVM_SNAPSHOT_REF_KIND_UNKNOWN;
     MVMuint16 ref_index = desc
         ? get_string_index(tc, ss, desc, STR_MODE_CONST)
         : 0;
-    while (c_ptr = MVM_gc_worklist_get(tc, ss->gcwl)) {
+    while (( c_ptr = MVM_gc_worklist_get(tc, ss->gcwl) )) {
         MVMCollectable *c = *c_ptr;
         if (c)
             add_reference(tc, ss, ref_kind, ref_index,
@@ -310,7 +312,7 @@ static void process_workitems(MVMThreadContext *tc, MVMHeapSnapshotState *ss) {
         MVMHeapSnapshotWorkItem item = pop_workitem(tc, ss);
 
         /* We take our own working copy of the collectable info, since the
-         * collectables array can grow and be reallocated. */ 
+         * collectables array can grow and be reallocated. */
         MVMHeapSnapshotCollectable col;
         set_ref_from(tc, ss, item.col_idx);
         col = ss->hs->collectables[item.col_idx];
@@ -447,8 +449,6 @@ static void process_workitems(MVMThreadContext *tc, MVMHeapSnapshotState *ss) {
                     (MVMCollectable *)frame->code_ref, "Code reference");
                 MVM_profile_heap_add_collectable_rel_const_cstr(tc, ss,
                     (MVMCollectable *)frame->static_info, "Static frame");
-                MVM_profile_heap_add_collectable_rel_const_cstr(tc, ss,
-                    (MVMCollectable *)frame->context_object, "Context object");
                 MVM_profile_heap_add_collectable_rel_const_cstr(tc, ss,
                     (MVMCollectable *)frame->dynlex_cache_name,
                     "Dynamic lexical cache name");
@@ -482,7 +482,7 @@ static void process_workitems(MVMThreadContext *tc, MVMHeapSnapshotState *ss) {
             case MVM_SNAPSHOT_COL_KIND_THREAD_ROOTS: {
                 MVMThreadContext *thread_tc = (MVMThreadContext *)item.target;
                 MVM_gc_root_add_tc_roots_to_worklist(thread_tc, NULL, ss);
-                if (!MVM_FRAME_IS_ON_CALLSTACK(tc, tc->cur_frame))
+                if (thread_tc->cur_frame && !MVM_FRAME_IS_ON_CALLSTACK(thread_tc, tc->cur_frame))
                     add_reference_const_cstr(tc, ss, "Current frame",
                         get_frame_idx(tc, ss, thread_tc->cur_frame));
                  break;
@@ -581,9 +581,7 @@ void MVM_profile_heap_add_collectable_rel_idx(MVMThreadContext *tc,
 
 /* Drives the overall process of recording a snapshot of the heap. */
 static void record_snapshot(MVMThreadContext *tc, MVMHeapSnapshotCollection *col, MVMHeapSnapshot *hs) {
-    MVMuint64 perm_root_synth;
-
-    /* Iinitialize state for taking a snapshot. */
+    /* Initialize state for taking a snapshot. */
     MVMHeapSnapshotState ss;
     memset(&ss, 0, sizeof(MVMHeapSnapshotState));
     ss.col = col;
@@ -661,7 +659,7 @@ MVMObject * types_str(MVMThreadContext *tc, MVMHeapSnapshotCollection *col) {
      MVMuint64 i;
      for (i = 0; i < col->num_types; i++) {
          char tmp[256];
-         size_t item_chars = snprintf(tmp, 256,
+         int item_chars = snprintf(tmp, 256,
             "%"PRIu64",%"PRIu64";",
             col->types[i].repr_name,
             col->types[i].type_name);
@@ -695,7 +693,7 @@ MVMObject * static_frames_str(MVMThreadContext *tc, MVMHeapSnapshotCollection *c
      MVMuint64 i;
      for (i = 0; i < col->num_static_frames; i++) {
          char tmp[256];
-         size_t item_chars = snprintf(tmp, 256,
+         int item_chars = snprintf(tmp, 256,
             "%"PRId64",%"PRId64",%"PRId64",%"PRId64";",
             col->static_frames[i].name,
             col->static_frames[i].cuid,
@@ -731,7 +729,7 @@ MVMObject * collectables_str(MVMThreadContext *tc, MVMHeapSnapshot *s) {
      MVMuint64 i;
      for (i = 0; i < s->num_collectables; i++) {
          char tmp[256];
-         size_t item_chars = snprintf(tmp, 256,
+         int item_chars = snprintf(tmp, 256,
             "%"PRIu16",%"PRId32",%"PRIu16",%"PRIu64",%"PRIu64",%"PRIu32";",
             s->collectables[i].kind,
             s->collectables[i].type_or_frame_index,
@@ -769,7 +767,7 @@ MVMObject * references_str(MVMThreadContext *tc, MVMHeapSnapshot *s) {
     MVMuint64 i;
     for (i = 0; i < s->num_references; i++) {
         char tmp[128];
-        size_t item_chars = snprintf(tmp, 128, "%lu,%lu,%lu;",
+        int item_chars = snprintf(tmp, 128, "%"PRIu64",%"PRIu64",%"PRIu64";",
             s->references[i].description & ((1 << MVM_SNAPSHOT_REF_KIND_BITS) - 1),
             s->references[i].description >> MVM_SNAPSHOT_REF_KIND_BITS,
             s->references[i].collectable_index);
