@@ -35,7 +35,7 @@ static void appendf(DumpStr *ds, const char *fmt, ...) {
     char *c_message = MVM_malloc(1024);
     va_list args;
     va_start(args, fmt);
-    c_message[vsnprintf(c_message, 1023, fmt, args)] = 0;
+    vsnprintf(c_message, 1023, fmt, args);
     append(ds, c_message);
     MVM_free(c_message);
     va_end(args);
@@ -138,6 +138,14 @@ static void dump_bb(MVMThreadContext *tc, DumpStr *ds, MVMSpeshGraph *g, MVMSpes
                     appendf(ds, "      [Annotation: INS Deopt OSR (idx %d -> pc %d); line %d]\n",
                         ann->data.deopt_idx, g->deopt_addrs[2 * ann->data.deopt_idx], line_number);
                     break;
+                case MVM_SPESH_ANN_LINENO: {
+                    char *cstr = MVM_string_utf8_encode_C_string(tc,
+                        MVM_cu_string(tc, g->sf->body.cu, ann->data.lineno.filename_string_index));
+                    appendf(ds, "      [Annotation: Line Number: %s:%d]\n",
+                        cstr, ann->data.lineno.line_number);
+                    MVM_free(cstr);
+                    break;
+                }
                 default:
                     appendf(ds, "      [Annotation: %d (unknown)]\n", ann->type);
             }
@@ -271,7 +279,7 @@ static void dump_bb(MVMThreadContext *tc, DumpStr *ds, MVMSpeshGraph *g, MVMSpes
                 /* We can try to find out what the debug_name of this thing is. */
                 MVMint16 dep = cur_ins->operands[1].lit_i16;
                 MVMint64 idx;
-                MVMCollectable *result;
+                MVMCollectable *result = NULL;
                 MVMSerializationContext *sc;
                 char *debug_name = NULL;
                 const char *repr_name = NULL;
@@ -372,6 +380,9 @@ static void dump_facts(MVMThreadContext *tc, DumpStr *ds, MVMSpeshGraph *g) {
             if (flags & 8192) {
                 append(ds, " KRWCn");
             }
+            if (g->facts[i][j].dead_writer) {
+                append(ds, " DeadWriter");
+            }
             if (g->facts[i][j].writer && g->facts[i][j].writer->info->opcode == MVM_SSA_PHI) {
                 appendf(ds, " (merged from %d regs)", g->facts[i][j].writer->info->num_operands - 1);
             }
