@@ -187,7 +187,7 @@ static void iter_facts(MVMThreadContext *tc, MVMSpeshGraph *g,
 
     if (in_facts->flags & MVM_SPESH_FACT_KNOWN_TYPE) {
         switch (REPR(in_facts->type)->ID) {
-            case MVM_REPR_ID_MVMArray:
+            case MVM_REPR_ID_VMArray:
                 out_facts->type = g->sf->body.cu->body.hll_config->array_iterator_type;
                 out_facts->flags |= MVM_SPESH_FACT_ARRAY_ITER;
                 break;
@@ -568,33 +568,23 @@ static void add_bb_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
                 tc->instance->boot_types.BOOTException);
             break;
-        case MVM_OP_getregref_i:
         case MVM_OP_getlexref_i:
-        case MVM_OP_getregref_i32:
         case MVM_OP_getlexref_i32:
-        case MVM_OP_getregref_i16:
         case MVM_OP_getlexref_i16:
-        case MVM_OP_getregref_i8:
         case MVM_OP_getlexref_i8:
-        case MVM_OP_getregref_u32:
         case MVM_OP_getlexref_u32:
-        case MVM_OP_getregref_u16:
         case MVM_OP_getlexref_u16:
-        case MVM_OP_getregref_u8:
         case MVM_OP_getlexref_u8:
             create_facts_with_type(tc, g,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
                 g->sf->body.cu->body.hll_config->int_lex_ref);
             break;
-        case MVM_OP_getregref_n:
         case MVM_OP_getlexref_n:
-        case MVM_OP_getregref_n32:
         case MVM_OP_getlexref_n32:
             create_facts_with_type(tc, g,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
                 g->sf->body.cu->body.hll_config->num_lex_ref);
             break;
-        case MVM_OP_getregref_s:
         case MVM_OP_getlexref_s:
             create_facts_with_type(tc, g,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
