@@ -14,42 +14,10 @@
 #define GET_UI32(pc, idx)   *((MVMuint32 *)((pc) + (idx)))
 #define GET_N32(pc, idx)    *((MVMnum32 *)((pc) + (idx)))
 
-/* Allocate a piece of memory from the spesh graph's buffer. Deallocated when
- * the spesh graph is. */
+/* Allocate a piece of memory from the spesh graph's region
+ * allocator. Deallocated when the spesh graph is. */
 void * MVM_spesh_alloc(MVMThreadContext *tc, MVMSpeshGraph *g, size_t bytes) {
-    char *result = NULL;
-
-#if !defined(MVM_CAN_UNALIGNED_INT64) || !defined(MVM_CAN_UNALIGNED_NUM64)
-    /* Round up size to next multiple of 8, to ensure alignment. */
-    bytes = (bytes + 7) & ~7;
-#endif
-
-    if (g->mem_block) {
-        MVMSpeshMemBlock *block = g->mem_block;
-        if (block->alloc + bytes < block->limit) {
-            result = block->alloc;
-            block->alloc += bytes;
-        }
-    }
-    if (!result) {
-        /* No block, or block was full. Add another. */
-        MVMSpeshMemBlock *block = MVM_malloc(sizeof(MVMSpeshMemBlock));
-        size_t buffer_size = g->mem_block
-            ? MVM_SPESH_MEMBLOCK_SIZE
-            : MVM_SPESH_FIRST_MEMBLOCK_SIZE;
-        if (buffer_size < bytes)
-            buffer_size = bytes;
-        block->buffer = MVM_calloc(buffer_size, 1);
-        block->alloc  = block->buffer;
-        block->limit  = block->buffer + buffer_size;
-        block->prev   = g->mem_block;
-        g->mem_block  = block;
-
-        /* Now allocate out of it. */
-        result = block->alloc;
-        block->alloc += bytes;
-    }
-    return result;
+    return MVM_region_alloc(tc, &g->region_alloc, bytes);
 }
 
 /* Looks up op info; doesn't sanity check, since we should be working on code
@@ -103,6 +71,7 @@ MVMSpeshBB * MVM_spesh_graph_linear_prev(MVMThreadContext *tc, MVMSpeshGraph *g,
  * with it. This also makes nodes for all of the instruction. */
 #define MVM_CFG_BB_START    1
 #define MVM_CFG_BB_END      2
+#define MVM_CFG_BB_JUMPLIST 4
 static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf,
                       MVMint32 *existing_deopts, MVMint32 num_existing_deopts) {
     MVMSpeshBB  *cur_bb, *prev_bb;
@@ -118,10 +87,11 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
      * integer. This is used for two things:
      * A) When we make the MVMSpeshIns for an instruction starting at the
      *    byte, we put the instruction index (into ins_flat) in the slot,
-     *    shifting it by 2 bits to the left. We will use this to do fixups.
+     *    shifting it by 3 bits to the left. We will use this to do fixups.
      * B) The first bit is "I have an incoming branch" - that is, start of
      *    a basic block. The second bit is "I can branch" - that is, end of
-     *    a basic block. It's possible to have both bits set.
+     *    a basic block. It's possible to have both bits set. If it's part
+     *    of a jumplist, it gets the third bit set also.
      * Anything that's just a zero has no instruction starting there. */
     MVMuint32 *byte_to_ins_flags = MVM_calloc(g->bytecode_size, sizeof(MVMuint32));
 
@@ -136,6 +106,10 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
     MVMuint8    *end      = g->bytecode + g->bytecode_size;
     MVMuint32    ins_idx  = 0;
     MVMuint8     next_bbs = 1; /* Next iteration (here, first) starts a BB. */
+    MVMuint32    lineno_ann_offs = 0;
+
+    MVMBytecodeAnnotation *ann_ptr = MVM_bytecode_resolve_annotation(tc, &sf->body, sf->body.bytecode - pc);
+
     for (i = 0; i < g->num_handlers; i++)
         byte_to_ins_flags[g->handlers[i].goto_offset] |= MVM_CFG_BB_START;
     while (pc < end) {
@@ -148,7 +122,7 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
         /* Create an instruction node, add it, and record its position. */
         MVMSpeshIns *ins_node = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshIns));
         ins_flat[ins_idx] = ins_node;
-        byte_to_ins_flags[pc - g->bytecode] |= ins_idx << 2;
+        byte_to_ins_flags[pc - g->bytecode] |= ins_idx << 3;
 
         /* Did previous instruction end a basic block? */
         if (next_bbs) {
@@ -170,6 +144,18 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
         /* Store opcode */
         ins_node->info = info;
 
+        /* Let's see if we have a line-number annotation */
+        if (ann_ptr && pc - sf->body.bytecode == ann_ptr->bytecode_offset) {
+            MVMSpeshAnn *lineno_ann = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshAnn));
+            lineno_ann->next = ins_node->annotations;
+            lineno_ann->type = MVM_SPESH_ANN_LINENO;
+            lineno_ann->data.lineno.filename_string_index = ann_ptr->filename_string_heap_index;
+            lineno_ann->data.lineno.line_number = ann_ptr->line_number;
+            ins_node->annotations = lineno_ann;
+
+            MVM_bytecode_advance_annotation(tc, &sf->body, ann_ptr);
+        }
+
         /* Go over operands. */
         ins_node->operands = MVM_spesh_alloc(tc, g, info->num_operands * sizeof(MVMSpeshOperand));
         for (i = 0; i < info->num_operands; i++) {
@@ -272,7 +258,8 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
         if (opcode == MVM_OP_jumplist) {
             MVMint64 n = MVM_BC_get_I64(args, 0);
             for (i = 0; i <= n; i++)
-                byte_to_ins_flags[(pc - g->bytecode) + 12 + i * 6] |= MVM_CFG_BB_START;
+                byte_to_ins_flags[(pc - g->bytecode) + 12 + i * 6] |=
+                    MVM_CFG_BB_START | MVM_CFG_BB_JUMPLIST;
             byte_to_ins_flags[pc - g->bytecode] |= MVM_CFG_BB_END;
         }
 
@@ -325,9 +312,9 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
 
     /* Annotate instructions that are handler-significant. */
     for (i = 0; i < g->num_handlers; i++) {
-        MVMSpeshIns *start_ins = ins_flat[byte_to_ins_flags[g->handlers[i].start_offset] >> 2];
-        MVMSpeshIns *end_ins   = ins_flat[byte_to_ins_flags[g->handlers[i].end_offset] >> 2];
-        MVMSpeshIns *goto_ins  = ins_flat[byte_to_ins_flags[g->handlers[i].goto_offset] >> 2];
+        MVMSpeshIns *start_ins = ins_flat[byte_to_ins_flags[g->handlers[i].start_offset] >> 3];
+        MVMSpeshIns *end_ins   = ins_flat[byte_to_ins_flags[g->handlers[i].end_offset] >> 3];
+        MVMSpeshIns *goto_ins  = ins_flat[byte_to_ins_flags[g->handlers[i].goto_offset] >> 3];
         MVMSpeshAnn *start_ann = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshAnn));
         MVMSpeshAnn *end_ann   = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshAnn));
         MVMSpeshAnn *goto_ann  = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshAnn));
@@ -350,8 +337,8 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
 
     /* Annotate instructions that are inline start/end points. */
     for (i = 0; i < g->num_inlines; i++) {
-        MVMSpeshIns *start_ins = ins_flat[byte_to_ins_flags[g->inlines[i].start] >> 2];
-        MVMSpeshIns *end_ins   = ins_flat[byte_to_ins_flags[g->inlines[i].end] >> 2];
+        MVMSpeshIns *start_ins = ins_flat[byte_to_ins_flags[g->inlines[i].start] >> 3];
+        MVMSpeshIns *end_ins   = ins_flat[byte_to_ins_flags[g->inlines[i].end] >> 3];
         MVMSpeshAnn *start_ann = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshAnn));
         MVMSpeshAnn *end_ann   = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshAnn));
 
@@ -390,7 +377,7 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
             continue;
 
         /* Get current instruction. */
-        cur_ins = ins_flat[byte_to_ins_flags[i] >> 2];
+        cur_ins = ins_flat[byte_to_ins_flags[i] >> 3];
 
         /* Start of a basic block? */
         if (byte_to_ins_flags[i] & MVM_CFG_BB_START) {
@@ -405,6 +392,7 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
             cur_bb->first_ins = cur_ins;
             cur_bb->idx = bb_idx;
             cur_bb->initial_pc = i;
+            cur_bb->jumplist = byte_to_ins_flags[i] & MVM_CFG_BB_JUMPLIST;
             bb_idx++;
 
             /* Record instruction -> BB start mapping. */
@@ -451,7 +439,7 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
             cur_bb->succ[0]  = cur_bb->linear_next;
             for (i = 0; i < g->num_handlers; i++) {
                 MVMuint32 offset = g->handlers[i].goto_offset;
-                cur_bb->succ[i + 1] = ins_to_bb[byte_to_ins_flags[offset] >> 2];
+                cur_bb->succ[i + 1] = ins_to_bb[byte_to_ins_flags[offset] >> 3];
             }
         }
 
@@ -473,7 +461,7 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
                 case MVM_OP_goto: {
                     /* Unconditional branch, so one successor. */
                     MVMuint32   offset = cur_bb->last_ins->operands[0].ins_offset;
-                    MVMSpeshBB *tgt    = ins_to_bb[byte_to_ins_flags[offset] >> 2];
+                    MVMSpeshBB *tgt    = ins_to_bb[byte_to_ins_flags[offset] >> 3];
                     cur_bb->succ       = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshBB *));
                     cur_bb->succ[0]    = tgt;
                     cur_bb->num_succ   = 1;
@@ -488,7 +476,7 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
                     for (i = 0; i < cur_bb->last_ins->info->num_operands; i++) {
                         if (cur_bb->last_ins->info->operands[i] == MVM_operand_ins) {
                             MVMuint32 offset = cur_bb->last_ins->operands[i].ins_offset;
-                            cur_bb->succ[0] = ins_to_bb[byte_to_ins_flags[offset] >> 2];
+                            cur_bb->succ[0] = ins_to_bb[byte_to_ins_flags[offset] >> 3];
                             cur_bb->num_succ++;
                             cur_bb->last_ins->operands[i].ins_bb = cur_bb->succ[0];
                         }
@@ -517,10 +505,10 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
     if (existing_deopts) {
         for (i = 0; i < num_existing_deopts; i ++) {
             if (existing_deopts[2 * i + 1] >= 0) {
-                MVMSpeshIns *post_ins     = ins_flat[byte_to_ins_flags[existing_deopts[2 * i + 1]] >> 2];
+                MVMSpeshIns *post_ins     = ins_flat[byte_to_ins_flags[existing_deopts[2 * i + 1]] >> 3];
                 MVMSpeshIns *deopt_ins    = post_ins->prev ? post_ins->prev :
                     MVM_spesh_graph_linear_prev(tc, g,
-                        ins_to_bb[byte_to_ins_flags[existing_deopts[2 * i + 1]] >> 2])->last_ins;
+                        ins_to_bb[byte_to_ins_flags[existing_deopts[2 * i + 1]] >> 3])->last_ins;
                 MVMSpeshAnn *deopt_ann    = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshAnn));
                 deopt_ann->next           = deopt_ins->annotations;
                 deopt_ann->type           = MVM_SPESH_ANN_DEOPT_INLINE;
@@ -534,6 +522,37 @@ static void build_cfg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMStaticFrame *sf
     MVM_free(byte_to_ins_flags);
     MVM_free(ins_flat);
     MVM_free(ins_to_bb);
+    MVM_free(ann_ptr);
+}
+
+/* Inserts nulling of object reigsters. A later stage of the optimizer will
+ * throw out any that are unrequired, leaving only those that cover (rare)
+ * "register read before assigned" cases. (We can thus just start off with
+ * them NULL, since zeroed memory is cheaper than copying a VMNulls in to
+ * place). */
+static MVMint32 is_handler_reg(MVMThreadContext *tc, MVMSpeshGraph *g, MVMuint16 reg) {
+    MVMuint32 num_handlers = g->num_handlers;
+    MVMuint32 i;
+    for (i = 0; i < num_handlers; i++)
+        if (g->handlers[i].action == MVM_EX_ACTION_INVOKE)
+            if (g->handlers[i].block_reg == reg)
+                return 1;
+    return 0;
+}
+static void insert_object_null_instructions(MVMThreadContext *tc, MVMSpeshGraph *g) {
+    MVMSpeshBB *insert_bb = g->entry->linear_next;
+    MVMuint16 *local_types = g->sf->body.local_types;
+    MVMuint16  num_locals = g->sf->body.num_locals;
+    MVMuint16 i;
+    for (i = 0; i < num_locals; i++) {
+        if (local_types[i] == MVM_reg_obj && !is_handler_reg(tc, g, i)) {
+            MVMSpeshIns *null_ins = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshIns));
+            null_ins->info = MVM_op_get_op(MVM_OP_null);
+            null_ins->operands = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshOperand));
+            null_ins->operands[0].reg.orig = i;
+            MVM_spesh_manipulate_insert_ins(tc, insert_bb, NULL, null_ins);
+        }
+    }
 }
 
 /* Eliminates any unreachable basic blocks (that is, dead code). Not having
@@ -592,7 +611,8 @@ static void add_predecessors(MVMThreadContext *tc, MVMSpeshGraph *g) {
             MVMSpeshBB  *tgt = cur_bb->succ[i];
             MVMSpeshBB **new_pred = MVM_spesh_alloc(tc, g,
                 (tgt->num_pred + 1) * sizeof(MVMSpeshBB *));
-            memcpy(new_pred, tgt->pred, tgt->num_pred * sizeof(MVMSpeshBB *));
+            if (tgt->num_pred)
+                memcpy(new_pred, tgt->pred, tgt->num_pred * sizeof(MVMSpeshBB *));
             new_pred[tgt->num_pred] = cur_bb;
             tgt->pred = new_pred;
             tgt->num_pred++;
@@ -731,7 +751,8 @@ static void add_child(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *target
 
     /* Nope, so insert. */
     new_children = MVM_spesh_alloc(tc, g, (target->num_children + 1) * sizeof(MVMSpeshBB *));
-    memcpy(new_children, target->children, target->num_children * sizeof(MVMSpeshBB *));
+    if (target->num_children)
+        memcpy(new_children, target->children, target->num_children * sizeof(MVMSpeshBB *));
     new_children[target->num_children] = to_add;
     target->children = new_children;
     target->num_children++;
@@ -758,7 +779,8 @@ static void add_to_frontier_set(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpesh
 
     /* Nope, so insert. */
     new_df = MVM_spesh_alloc(tc, g, (target->num_df + 1) * sizeof(MVMSpeshBB *));
-    memcpy(new_df, target->df, target->num_df * sizeof(MVMSpeshBB *));
+    if (target->num_df)
+        memcpy(new_df, target->df, target->num_df * sizeof(MVMSpeshBB *));
     new_df[target->num_df] = to_add;
     target->df = new_df;
     target->num_df++;
@@ -799,7 +821,7 @@ typedef struct {
 /* Creates an SSAVarInfo for each local, initializing it with a list of nodes
  * that assign to the local. */
 static SSAVarInfo * initialize_ssa_var_info(MVMThreadContext *tc, MVMSpeshGraph *g) {
-    SSAVarInfo *var_info = MVM_calloc(sizeof(SSAVarInfo), g->num_locals);
+    SSAVarInfo *var_info = MVM_calloc(g->num_locals, sizeof(SSAVarInfo));
     MVMint32 i;
 
     /* Visit all instructions, looking for local writes. */
@@ -1077,7 +1099,8 @@ static void ssa(MVMThreadContext *tc, MVMSpeshGraph *g) {
 }
 
 /* Takes a static frame and creates a spesh graph for it. */
-MVMSpeshGraph * MVM_spesh_graph_create(MVMThreadContext *tc, MVMStaticFrame *sf, MVMuint32 cfg_only) {
+MVMSpeshGraph * MVM_spesh_graph_create(MVMThreadContext *tc, MVMStaticFrame *sf,
+        MVMuint32 cfg_only, MVMuint32 insert_object_nulls) {
     /* Create top-level graph object. */
     MVMSpeshGraph *g = MVM_calloc(1, sizeof(MVMSpeshGraph));
     g->sf            = sf;
@@ -1097,6 +1120,8 @@ MVMSpeshGraph * MVM_spesh_graph_create(MVMThreadContext *tc, MVMStaticFrame *sf,
 
     /* Build the CFG out of the static frame, and transform it to SSA. */
     build_cfg(tc, g, sf, NULL, 0);
+    if (insert_object_nulls)
+        insert_object_null_instructions(tc, g);
     if (!cfg_only) {
         eliminate_dead(tc, g);
         add_predecessors(tc, g);
@@ -1116,7 +1141,7 @@ MVMSpeshGraph * MVM_spesh_graph_create_from_cand(MVMThreadContext *tc, MVMStatic
     g->bytecode          = cand->bytecode;
     g->bytecode_size     = cand->bytecode_size;
     g->handlers          = cand->handlers;
-    g->num_handlers      = sf->body.num_handlers;
+    g->num_handlers      = cand->num_handlers;
     g->num_locals        = cand->num_locals;
     g->num_lexicals      = cand->num_lexicals;
     g->inlines           = cand->inlines;
@@ -1129,6 +1154,7 @@ MVMSpeshGraph * MVM_spesh_graph_create_from_cand(MVMThreadContext *tc, MVMStatic
     g->spesh_slots       = cand->spesh_slots;
     g->num_spesh_slots   = cand->num_spesh_slots;
     g->phi_infos         = MVM_spesh_alloc(tc, g, MVMPhiNodeCacheSize * sizeof(MVMOpInfo));
+    g->cand              = cand;
 
     /* Ensure the frame is validated, since we'll rely on this. */
     if (sf->body.instrumentation_level == 0) {
@@ -1179,13 +1205,10 @@ void MVM_spesh_graph_mark(MVMThreadContext *tc, MVMSpeshGraph *g, MVMGCWorklist
 /* Destroys a spesh graph, deallocating all its associated memory. */
 void MVM_spesh_graph_destroy(MVMThreadContext *tc, MVMSpeshGraph *g) {
     /* Free all of the allocated node memory. */
-    MVMSpeshMemBlock *cur_block = g->mem_block;
-    while (cur_block) {
-        MVMSpeshMemBlock *prev = cur_block->prev;
-        MVM_free(cur_block->buffer);
-        MVM_free(cur_block);
-        cur_block = prev;
-    }
+    MVM_region_destroy(tc, &g->region_alloc);
+    /* Free handlers array, if different from the static frame. */
+    if (!g->cand && g->handlers != g->sf->body.handlers)
+        MVM_free(g->handlers);
 
     /* Free the graph itself. */
     MVM_free(g);
