@@ -44,9 +44,8 @@ struct MVMSpeshGraph {
     /* Number of log-based guards we have. */
     MVMint32 num_log_guards;
 
-    /* Memory blocks we allocate to store spesh nodes, and which we free along
-     * with the graph. Contains a link to previous blocks. */
-    MVMSpeshMemBlock *mem_block;
+    /* Region allocator for spesh nodes */
+    MVMRegionAlloc region_alloc;
 
     /* Values placed in spesh slots. */
     MVMCollectable **spesh_slots;
@@ -108,26 +107,6 @@ struct MVMSpeshGraph {
     MVMSpeshCandidate *cand;
 };
 
-/* The default allocation chunk size for memory blocks used to store spesh
- * graph nodes. Power of two is best; we start small also. */
-#define MVM_SPESH_FIRST_MEMBLOCK_SIZE 32768
-#define MVM_SPESH_MEMBLOCK_SIZE       8192
-
-/* A block of bump-pointer allocated memory. */
-struct MVMSpeshMemBlock {
-    /* The memory buffer itself. */
-    char *buffer;
-
-    /* Current allocation position. */
-    char *alloc;
-
-    /* Allocation limit. */
-    char *limit;
-
-    /* Previous, now full, memory block. */
-    MVMSpeshMemBlock *prev;
-};
-
 /* A temporary register, added to support transformations. */
 struct MVMSpeshTemporary {
     /* The number of the local along with the current SSA index. */
@@ -182,7 +161,10 @@ struct MVMSpeshBB {
     MVMuint32 initial_pc;
 
     /* Is this block an inlining of another one? */
-    MVMint32 inlined;
+    MVMint16 inlined;
+
+    /* Is this basic block part of a jump list? */
+    MVMint16 jumplist;
 };
 
 /* The SSA phi instruction. */
@@ -241,6 +223,10 @@ struct MVMSpeshAnn {
         MVMint32 frame_handler_index;
         MVMint32 deopt_idx;
         MVMint32 inline_idx;
+        struct {
+            MVMuint32 filename_string_index;
+            MVMuint32 line_number;
+        } lineno;
     } data;
 };
 
@@ -254,8 +240,9 @@ struct MVMSpeshAnn {
 #define MVM_SPESH_ANN_INLINE_END    7
 #define MVM_SPESH_ANN_DEOPT_INLINE  8
 #define MVM_SPESH_ANN_DEOPT_OSR     9
+#define MVM_SPESH_ANN_LINENO        10
 
-/* Functions to create/destory the spesh graph. */
+/* Functions to create/destroy the spesh graph. */
 MVMSpeshGraph * MVM_spesh_graph_create(MVMThreadContext *tc, MVMStaticFrame *sf,
     MVMuint32 cfg_only, MVMuint32 insert_object_nulls);
 MVMSpeshGraph * MVM_spesh_graph_create_from_cand(MVMThreadContext *tc, MVMStaticFrame *sf,
