@@ -3,6 +3,9 @@
 /* This is where the main optimization work on a spesh graph takes place,
  * using facts discovered during analysis. */
 
+/* Writes to stderr about each inline that we perform. */
+#define MVM_LOG_INLINES 0
+
 /* Obtains facts for an operand, just directly accessing them without
  * inferring any kind of usage. */
 static MVMSpeshFacts * get_facts_direct(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshOperand o) {
@@ -171,7 +174,7 @@ static void optimize_is_reprid(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshI
     }
 
     switch (ins->info->opcode) {
-        case MVM_OP_islist: wanted_repr_id = MVM_REPR_ID_MVMArray; break;
+        case MVM_OP_islist: wanted_repr_id = MVM_REPR_ID_VMArray; break;
         case MVM_OP_ishash: wanted_repr_id = MVM_REPR_ID_MVMHash; break;
         case MVM_OP_isint:  wanted_repr_id = MVM_REPR_ID_P6int; break;
         case MVM_OP_isnum:  wanted_repr_id = MVM_REPR_ID_P6num; break;
@@ -823,7 +826,7 @@ static void optimize_smart_coerce(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpe
                 MVM_spesh_manipulate_release_temp_reg(tc, g, temp);
 
                 return;
-            } else if (!is_strify && (REPR(facts->type)->ID == MVM_REPR_ID_MVMArray ||
+            } else if (!is_strify && (REPR(facts->type)->ID == MVM_REPR_ID_VMArray ||
                                      (REPR(facts->type)->ID == MVM_REPR_ID_MVMHash))) {
                 /* A smrt_numify on an array or hash can be replaced by an
                  * elems operation, that can then be optimized by our
@@ -1168,18 +1171,23 @@ static void optimize_call(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb
                 /* Yes. Will we be able to inline? */
                 MVMSpeshGraph *inline_graph = MVM_spesh_inline_try_get_graph(tc, g,
                     target_code, &target_code->body.sf->body.spesh_candidates[spesh_cand]);
-                if (inline_graph) {
-                    /* Yes, have inline graph, so go ahead and do it. */
-                    /*char *c_name_i = MVM_string_utf8_encode_C_string(tc, target_code->body.sf->body.name);
+#if MVM_LOG_INLINES
+                {
+                    char *c_name_i = MVM_string_utf8_encode_C_string(tc, target_code->body.sf->body.name);
                     char *c_cuid_i = MVM_string_utf8_encode_C_string(tc, target_code->body.sf->body.cuuid);
                     char *c_name_t = MVM_string_utf8_encode_C_string(tc, g->sf->body.name);
                     char *c_cuid_t = MVM_string_utf8_encode_C_string(tc, g->sf->body.cuuid);
-                    printf("Can inline %s (%s) into %s (%s)\n",
+                    fprintf(stderr, "%s inline %s (%s) into %s (%s)\n",
+                        (inline_graph ? "Can" : "Can NOT"),
                         c_name_i, c_cuid_i, c_name_t, c_cuid_t);
                     MVM_free(c_name_i);
                     MVM_free(c_cuid_i);
                     MVM_free(c_name_t);
-                    MVM_free(c_cuid_t);*/
+                    MVM_free(c_cuid_t);
+                }
+#endif
+                if (inline_graph) {
+                    /* Yes, have inline graph, so go ahead and do it. */
                     MVM_spesh_inline(tc, g, arg_info, bb, ins, inline_graph, target_code);
                 }
                 else {
@@ -1219,6 +1227,15 @@ static void optimize_call(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb
     }
 }
 
+static void optimize_coverage_log(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb, MVMSpeshIns *ins) {
+    char *cache        = (char *)ins->operands[3].lit_i64;
+    MVMint32 cache_idx = ins->operands[2].lit_i32;
+
+    if (cache[cache_idx] != 0) {
+        MVM_spesh_manipulate_delete_ins(tc, g, bb, ins);
+    }
+}
+
 /* Optimizes an extension op. */
 static void optimize_extop(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb, MVMSpeshIns *ins) {
     MVMExtOpRecord *extops     = g->sf->body.cu->body.extops;
@@ -1358,6 +1375,23 @@ static void optimize_throwcat(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB
     MVM_free(handlers_found);
 }
 
+static void eliminate_phi_dead_reads(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshIns *ins) {
+    MVMuint32 operand = 1;
+    MVMuint32 insert_pos = 1;
+    MVMuint32 num_operands = ins->info->num_operands;
+    while (operand < ins->info->num_operands) {
+        if (get_facts_direct(tc, g, ins->operands[operand])->dead_writer) {
+            num_operands--;
+        }
+        else {
+            ins->operands[insert_pos] = ins->operands[operand];
+            insert_pos++;
+        }
+        operand++;
+    }
+    if (num_operands != ins->info->num_operands)
+        ins->info = get_phi(tc, g, num_operands);
+}
 static void analyze_phi(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshIns *ins) {
     MVMuint32 operand;
     MVMint32 common_flags;
@@ -1366,6 +1400,8 @@ static void analyze_phi(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshIns *ins
     MVMuint32 needs_merged_with_log_guard = 0;
     MVMSpeshFacts *target_facts = get_facts_direct(tc, g, ins->operands[0]);
 
+    eliminate_phi_dead_reads(tc, g, ins);
+
     common_flags       = get_facts_direct(tc, g, ins->operands[1])->flags;
     common_type        = get_facts_direct(tc, g, ins->operands[1])->type;
     common_decont_type = get_facts_direct(tc, g, ins->operands[1])->decont_type;
@@ -1440,7 +1476,6 @@ static void analyze_phi(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshIns *ins
         }
     } else {
         /*fprintf(stderr, "a PHI node of %d operands had no intersecting flags\n", ins->info->num_operands);*/
-        return;
     }
 }
 /* Visits the blocks in dominator tree order, recursively. */
@@ -1658,6 +1693,9 @@ static void optimize_bb(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb)
             /* Profiling entered from spesh should indicate so. */
             ins->info = MVM_op_get_op(MVM_OP_prof_enterspesh);
             break;
+        case MVM_OP_coverage_log:
+            /* A coverage_log op that has already fired can be thrown out. */
+            optimize_coverage_log(tc, g, bb, ins);
         default:
             if (ins->info->opcode == (MVMuint16)-1)
                 optimize_extop(tc, g, bb, ins);
@@ -1810,6 +1848,14 @@ static MVMint64 has_handler_anns(MVMThreadContext *tc, MVMSpeshBB *bb) {
     }
     return 0;
 }
+static void mark_dead_writers(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *dead_bb) {
+    MVMSpeshIns *ins = dead_bb->first_ins;
+    while (ins) {
+        if ((ins->info->operands[0] & MVM_operand_rw_mask) == MVM_operand_write_reg)
+            get_facts_direct(tc, g, ins->operands[0])->dead_writer = 1; 
+        ins = ins->next;
+    }
+}
 static void eliminate_dead_bbs(MVMThreadContext *tc, MVMSpeshGraph *g) {
     /* Iterate to fixed point. */
     MVMint8  *seen     = MVM_malloc(g->num_bbs);
@@ -1837,6 +1883,7 @@ static void eliminate_dead_bbs(MVMThreadContext *tc, MVMSpeshGraph *g) {
             MVMSpeshBB *death_cand = cur_bb->linear_next;
             if (!seen[death_cand->idx]) {
                 if (!death_cand->inlined && !has_handler_anns(tc, death_cand)) {
+                    mark_dead_writers(tc, g, death_cand);
                     cur_bb->linear_next = cur_bb->linear_next->linear_next;
                     g->num_bbs--;
                     death = 1;
@@ -1868,11 +1915,31 @@ static void eliminate_unused_log_guards(MVMThreadContext *tc, MVMSpeshGraph *g)
                 g->log_guards[i].ins);
 }
 
+/* Sometimes - almost always due to other optmimizations having done their
+ * work - we end up with an unconditional goto at the end of a basic block
+ * that points right to the very next basic block. Delete these. */
+static void eliminate_pointless_gotos(MVMThreadContext *tc, MVMSpeshGraph *g) {
+    MVMSpeshBB *cur_bb = g->entry;
+    while (cur_bb) {
+        if (!cur_bb->jumplist) {
+            MVMSpeshIns *last_ins = cur_bb->last_ins;
+            if (last_ins && last_ins->info->opcode == MVM_OP_goto)
+                if (last_ins->operands[0].ins_bb == cur_bb->linear_next) 
+                    MVM_spesh_manipulate_delete_ins(tc, g, cur_bb, last_ins);
+        }
+        cur_bb = cur_bb->linear_next;
+    }
+}
+
 /* Drives the overall optimization work taking place on a spesh graph. */
 void MVM_spesh_optimize(MVMThreadContext *tc, MVMSpeshGraph *g) {
+    /* Before starting, we eliminate dead basic blocks that were tossed by
+     * arg spesh, to simplify the graph. */
+    eliminate_dead_bbs(tc, g);
     optimize_bb(tc, g, g->entry);
     eliminate_dead_ins(tc, g);
     eliminate_dead_bbs(tc, g);
     eliminate_unused_log_guards(tc, g);
+    eliminate_pointless_gotos(tc, g);
     second_pass(tc, g, g->entry);
 }
