@@ -1,8 +1,12 @@
 #include "moar.h"
+#define UNI_CP_MALE_SIGN             0x2642
+#define UNI_CP_FEMALE_SIGN           0x2640
+#define UNI_CP_ZERO_WIDTH_JOINER     0x200D
+#define UNI_CP_ZERO_WIDTH_NON_JOINER 0x200C
 
 /* Maps outside-world normalization form codes to our internal set, validating
  * that we got something valid. */
-MVMNormalization MVN_unicode_normalizer_form(MVMThreadContext *tc, MVMint64 form_in) {
+MVMNormalization MVM_unicode_normalizer_form(MVMThreadContext *tc, MVMint64 form_in) {
     switch (form_in) {
     case 1: return MVM_NORMALIZE_NFC;
     case 2: return MVM_NORMALIZE_NFD;
@@ -15,7 +19,7 @@ MVMNormalization MVN_unicode_normalizer_form(MVMThreadContext *tc, MVMint64 form
 /* Takes two objects, which must be of VMArray representation and holding
  * 32-bit integers. Performs normalization to the specified form. */
 static void assert_codepoint_array(MVMThreadContext *tc, const MVMObject *arr, char *error) {
-    if (IS_CONCRETE(arr) && REPR(arr)->ID == MVM_REPR_ID_MVMArray) {
+    if (IS_CONCRETE(arr) && REPR(arr)->ID == MVM_REPR_ID_VMArray) {
         MVMuint8 slot_type = ((MVMArrayREPRData *)STABLE(arr)->REPR_data)->slot_type;
         if (slot_type == MVM_ARRAY_I32 || slot_type == MVM_ARRAY_U32)
             return;
@@ -77,36 +81,27 @@ void MVM_unicode_normalize_codepoints(MVMThreadContext *tc, const MVMObject *in,
     ((MVMArray *)out)->body.start     = 0;
     ((MVMArray *)out)->body.elems     = result_pos;
 }
-
-/* Takes an object, which must be of VMArray representation and holding
- * 32-bit integers. Treats them as Unicode codepoints, normalizes them at
- * Grapheme level, and returns the resulting NFG string. */
-MVMString * MVM_unicode_codepoints_to_nfg_string(MVMThreadContext *tc, const MVMObject *codes) {
+MVMString * MVM_unicode_codepoints_c_array_to_nfg_string(MVMThreadContext *tc, MVMCodepoint * cp_v, MVMint64 cp_count) {
     MVMNormalizer  norm;
-    MVMCodepoint  *input;
+    MVMint64       input_pos, result_pos, result_alloc;
     MVMGrapheme32 *result;
-    MVMint64       input_pos, input_codes, result_pos, result_alloc;
     MVMint32       ready;
     MVMString     *str;
 
-    /* Get input array; if it's empty, we're done already. */
-    assert_codepoint_array(tc, codes, "Code points to string input must be native array of 32-bit integers");
-    input       = (MVMCodepoint *)((MVMArray *)codes)->body.slots.u32 + ((MVMArray *)codes)->body.start;
-    input_codes = ((MVMArray *)codes)->body.elems;
-    if (input_codes == 0)
+    if (cp_count == 0)
         return tc->instance->str_consts.empty;
 
-    /* Guess output size based on input size. */
-    result_alloc = input_codes;
+    /* Guess output size based on cp_v size. */
+    result_alloc = cp_count;
     result       = MVM_malloc(result_alloc * sizeof(MVMCodepoint));
 
     /* Perform normalization at grapheme level. */
     MVM_unicode_normalizer_init(tc, &norm, MVM_NORMALIZE_NFG);
     input_pos  = 0;
     result_pos = 0;
-    while (input_pos < input_codes) {
+    while (input_pos < cp_count) {
         MVMGrapheme32 g;
-        ready = MVM_unicode_normalizer_process_codepoint_to_grapheme(tc, &norm, input[input_pos], &g);
+        ready = MVM_unicode_normalizer_process_codepoint_to_grapheme(tc, &norm, cp_v[input_pos], &g);
         if (ready) {
             maybe_grow_result(&result, &result_alloc, result_pos + ready);
             result[result_pos++] = g;
@@ -130,6 +125,20 @@ MVMString * MVM_unicode_codepoints_to_nfg_string(MVMThreadContext *tc, const MVM
     return str;
 }
 
+/* Takes an object, which must be of VMArray representation and holding
+ * 32-bit integers. Treats them as Unicode codepoints, normalizes them at
+ * Grapheme level, and returns the resulting NFG string. */
+MVMString * MVM_unicode_codepoints_to_nfg_string(MVMThreadContext *tc, const MVMObject *codes) {
+    MVMCodepoint  *input;
+    MVMint64       input_codes;
+
+    assert_codepoint_array(tc, codes, "Code points to string input must be native array of 32-bit integers");
+
+    input       = (MVMCodepoint *)((MVMArray *)codes)->body.slots.u32 + ((MVMArray *)codes)->body.start;
+    input_codes = ((MVMArray *)codes)->body.elems;
+    return MVM_unicode_codepoints_c_array_to_nfg_string(tc, input, input_codes);
+}
+
 /* Takes an NFG string and populates the array out, which must be a 32-bit
  * integer array, with codepoints normalized according to the specified
  * normalization form. */
@@ -194,6 +203,8 @@ void MVM_unicode_normalizer_init(MVMThreadContext *tc, MVMNormalizer *n, MVMNorm
     n->buffer_end         = 0;
     n->buffer_norm_end    = 0;
     n->translate_newlines = 0;
+    n->prepend_buffer     = 0;
+    n->regional_indicator = 0;
     switch (n->form) {
         case MVM_NORMALIZE_NFD:
             n->first_significant    = MVM_NORMALIZE_FIRST_SIG_NFD;
@@ -281,13 +292,11 @@ static void decomp_codepoint_to_buffer(MVMThreadContext *tc, MVMNormalizer *n, M
     /* See if we actually need to decompose (can skip if the decomposition
      * type is None, or we're only doing Canonical decomposition and it is
      * anything except Canonical). */
-    const char *type = MVM_unicode_codepoint_get_property_cstr(tc, cp, MVM_UNICODE_PROPERTY_DECOMPOSITION_TYPE);
+    MVMint16 cp_DT = MVM_unicode_codepoint_get_property_int(tc, cp, MVM_UNICODE_PROPERTY_DECOMPOSITION_TYPE);
     MVMint64 decompose = 1;
-    if (!type)
+    if (cp_DT == MVM_UNICODE_PVALUE_DT_NONE)
         decompose = 0;
-    else if (strcmp(type, "None") == 0)
-        decompose = 0;
-    else if (!MVM_NORMALIZE_COMPAT_DECOMP(n->form) && strcmp(type, "Canonical") != 0)
+    else if (!MVM_NORMALIZE_COMPAT_DECOMP(n->form) && cp_DT != MVM_UNICODE_PVALUE_DT_CANONICAL )
         decompose = 0;
     if (decompose) {
         /* We need to decompose. Get the decomp spec and go over the things in
@@ -327,7 +336,7 @@ static MVMint64 ccc(MVMThreadContext *tc, MVMCodepoint cp) {
     }
     else {
         const char *ccc_str = MVM_unicode_codepoint_get_property_cstr(tc, cp, MVM_UNICODE_PROPERTY_CANONICAL_COMBINING_CLASS);
-        return !ccc_str || strlen(ccc_str) > 3 ? 0 : atoi(ccc_str);
+        return !ccc_str || strlen(ccc_str) > 3 ? 0 : fast_atoi(ccc_str);
     }
 }
 
@@ -337,25 +346,28 @@ static MVMint64 ccc(MVMThreadContext *tc, MVMCodepoint cp) {
  * a fast path. */
 static MVMint32 is_control_beyond_latin1(MVMThreadContext *tc, MVMCodepoint in) {
     /* U+200C ZERO WIDTH NON-JOINER and U+200D ZERO WIDTH JOINER are excluded. */
-    if (in != 0x200C && in != 0x200D) {
+    if (in != UNI_CP_ZERO_WIDTH_NON_JOINER && in != UNI_CP_ZERO_WIDTH_JOINER) {
         /* Consider general property. */
         const char *genprop = MVM_unicode_codepoint_get_property_cstr(tc, in,
             MVM_UNICODE_PROPERTY_GENERAL_CATEGORY);
-        if (genprop[0] == 'Z') {
-            /* Line_Separator and Paragraph_Separator are controls. */
-            return genprop[1] == 'l' || genprop[1] == 'p';
-        }
-        if (genprop[0] == 'C') {
-            /* Control, Surrogate, and Format are controls. */
-            if (genprop[1] == 'c' || genprop[1] == 's' || genprop[1] == 'f') {
-                return 1;
-            }
-
-            /* Unassigned is, but only for Default_Ignorable_Code_Point. */
-            if (genprop[1] == 'n') {
-                return MVM_unicode_codepoint_get_property_int(tc, in,
-                    MVM_UNICODE_PROPERTY_DEFAULT_IGNORABLE_CODE_POINT) != 0;
-            }
+        switch (genprop[0]) {
+            case 'Z':
+                /* Line_Separator and Paragraph_Separator are controls. */
+                return genprop[1] == 'l' || genprop[1] == 'p';
+            case 'C':
+                /* Control, Surrogate are controls. */
+                if (genprop[1] == 'c' || genprop[1] == 's') {
+                    return 1;
+                }
+                if (genprop[1] == 'f' ) {
+                    /* Format can have special properties (not control) */
+                    return 0;
+                }
+                /* Unassigned is, but only for Default_Ignorable_Code_Point. */
+                if (genprop[1] == 'n') {
+                    return MVM_unicode_codepoint_get_property_int(tc, in,
+                        MVM_UNICODE_PROPERTY_DEFAULT_IGNORABLE_CODE_POINT) != 0;
+                }
         }
     }
     return 0;
@@ -479,34 +491,19 @@ static void canonical_composition(MVMThreadContext *tc, MVMNormalizer *n, MVMint
  * the handling of breaking around controls much earlier, so don't have to
  * consider that case. */
 static MVMint32 maybe_hangul(MVMCodepoint cp) {
-    return cp >= 0x1100 && cp < 0x1200 || cp >= 0xA960 && cp < 0xD7FC;
-}
-static MVMint32 is_regional_indicator(MVMCodepoint cp) {
-    /* U+1F1E6 REGIONAL INDICATOR SYMBOL LETTER A
-     * ..
-     * U+1F1FF REGIONAL INDICATOR SYMBOL LETTER Z */
-    return cp >= 0x1F1E6 && cp <= 0x1F1FF;
+    return (cp >= 0x1100 && cp < 0x1200) || (cp >= 0xA960 && cp < 0xD7FC);
 }
 static MVMint32 is_grapheme_extend(MVMThreadContext *tc, MVMCodepoint cp) {
     return MVM_unicode_codepoint_get_property_int(tc, cp,
         MVM_UNICODE_PROPERTY_GRAPHEME_EXTEND);
 }
-static MVMint32 is_spacing_mark(MVMThreadContext *tc, MVMCodepoint cp) {
-    const char *genprop = MVM_unicode_codepoint_get_property_cstr(tc, cp,
-        MVM_UNICODE_PROPERTY_GENERAL_CATEGORY);
-    if (genprop[0] == 'M' && genprop[1] == 'c') {
-        const char *gcb = MVM_unicode_codepoint_get_property_cstr(tc, cp,
-            MVM_UNICODE_PROPERTY_GRAPHEME_CLUSTER_BREAK);
-        return strcmp(gcb, "Extend") != 0;
-    }
-    else {
-        /* Special cases outside of Mc:
-         * U+0E33 THAI CHARACTER SARA AM
-         * U+0EB3 LAO VOWEL SIGN AM */
-        return cp == 0x0E33 || cp == 0x0EB3;
-    }
+static MVMint32 is_grapheme_prepend(MVMThreadContext *tc, MVMCodepoint cp) {
+    return MVM_unicode_codepoint_get_property_int(tc, cp,
+        MVM_UNICODE_PROPERTY_PREPENDED_CONCATENATION_MARK);
 }
-static MVMint32 should_break(MVMThreadContext *tc, MVMCodepoint a, MVMCodepoint b) {
+static MVMint32 should_break(MVMThreadContext *tc, MVMCodepoint a, MVMCodepoint b, MVMNormalizer *norm) {
+    int GCB_a = MVM_unicode_codepoint_get_property_int(tc, a, MVM_UNICODE_PROPERTY_GRAPHEME_CLUSTER_BREAK);
+    int GCB_b = MVM_unicode_codepoint_get_property_int(tc, b, MVM_UNICODE_PROPERTY_GRAPHEME_CLUSTER_BREAK);
     /* Don't break between \r and \n, but otherwise break around \r. */
     if (a == 0x0D && b == 0x0A)
         return 0;
@@ -528,19 +525,74 @@ static MVMint32 should_break(MVMThreadContext *tc, MVMCodepoint a, MVMCodepoint
             return !(strcmp(hst_b, "T") == 0);
     }
 
-    /* Don't break between regional indicators. */
-    if (is_regional_indicator(a) && is_regional_indicator(b))
-        return 0;
-
-    /* Don't break before extenders or ZERO WIDTH JOINER. */
-    if (b == 0x200D || is_grapheme_extend(tc, b))
-        return 0;
-
-    /* Don't break before spacing marks. (In the Unicode version at the time
-     * of implementing, there were no Prepend characters, so we don't worry
-     * about that rule for now). */
-    if (is_spacing_mark(tc, b))
-        return 0;
+    switch (GCB_a) {
+        case MVM_UNICODE_PVALUE_GCB_REGIONAL_INDICATOR:
+            if (2 <= norm->regional_indicator) {
+                norm->regional_indicator = 0;
+                if (GCB_b == MVM_UNICODE_PVALUE_GCB_REGIONAL_INDICATOR)
+                    return 1;
+            }
+            if (GCB_b == MVM_UNICODE_PVALUE_GCB_REGIONAL_INDICATOR) {
+                if (!norm->regional_indicator)
+                    norm->regional_indicator = 2;
+                else
+                    norm->regional_indicator++;
+                return 0;
+            }
+            break;
+        /* Don't break after Prepend Grapheme_Cluster_Break=Prepend */
+        case MVM_UNICODE_PVALUE_GCB_PREPEND:
+            /* If it's a control character remember to break */
+            if (is_control_beyond_latin1(tc, b )) {
+                return 1;
+            }
+            /* Otherwise don't break */
+            return 0;
+        /* Don't break after ZWJ for E_Base_GAZ or Glue_After_ZWJ */
+        case MVM_UNICODE_PVALUE_GCB_ZWJ:
+            switch (GCB_b) {
+                case MVM_UNICODE_PVALUE_GCB_E_BASE_GAZ:
+                case MVM_UNICODE_PVALUE_GCB_ZWJ:
+                case MVM_UNICODE_PVALUE_GCB_GLUE_AFTER_ZWJ:
+                    return 0;
+            }
+            if ( b == UNI_CP_FEMALE_SIGN || b == UNI_CP_MALE_SIGN )
+                return 0;
+        case MVM_UNICODE_PVALUE_GCB_E_MODIFIER:
+            if (MVM_unicode_codepoint_get_property_int(tc, b, MVM_UNICODE_PROPERTY_EMOJI_MODIFIER_BASE)) {
+                /* Don't break after ZWJ if it's an Emoji Sequence.
+                 * At the moment FEMALE SIGN and MALE SIGN don't have different
+                 * GCB properties, or any special Emoji properties (Unicode 9.0),
+                 * so we explictly check these codepoints here */
+                if ( b == UNI_CP_FEMALE_SIGN || b == UNI_CP_MALE_SIGN )
+                    return 0;
+            }
+            break;
+    }
+    switch (GCB_b) {
+        /* Don't break before extending chars */
+        case MVM_UNICODE_PVALUE_GCB_EXTEND:
+            return 0;
+        /* Don't break before ZWJ */
+        case MVM_UNICODE_PVALUE_GCB_ZWJ:
+            return 0;
+        case MVM_UNICODE_PVALUE_GCB_E_MODIFIER:
+            switch (GCB_a) {
+                case MVM_UNICODE_PVALUE_GCB_E_BASE_GAZ:
+                    return 0;
+                case MVM_UNICODE_PVALUE_GCB_E_BASE:
+                    return 0;
+                /* Don't break
+                 * when in Emoji Sequences
+                 * we don't save state so can't support this now
+                 *case MVM_UNICODE_PVALUE_GCB_EXTEND:
+                 *    return 0; */
+            }
+            break;
+        /* Don't break before spacing marks. */
+        case MVM_UNICODE_PVALUE_GCB_SPACINGMARK:
+            return 0;
+    }
 
     /* Otherwise break. */
     return 1;
@@ -552,7 +604,7 @@ static void grapheme_composition(MVMThreadContext *tc, MVMNormalizer *n, MVMint3
         MVMint32 pos        = from;
         while (pos < to) {
             MVMint32 next_pos = pos + 1;
-            if (next_pos == to || should_break(tc, n->buffer[pos], n->buffer[next_pos])) {
+            if (next_pos == to || should_break(tc, n->buffer[pos], n->buffer[next_pos], n)) {
                 /* Last in buffer or next code point is a non-starter; turn
                  * sequence into a synthetic. */
                 MVMGrapheme32 g = MVM_nfg_codes_to_grapheme(tc, n->buffer + starterish, next_pos - starterish);
@@ -576,21 +628,28 @@ static void grapheme_composition(MVMThreadContext *tc, MVMNormalizer *n, MVMint3
  * significant codepoint identified by a quick check for the target form). We
  * may find the quick check itself is enough; if not, we have to do real work
  * compute the normalization. */
-MVMint32 MVM_unicode_normalizer_process_codepoint_full(MVMThreadContext *tc, MVMNormalizer *n, MVMCodepoint in, MVMCodepoint *out) {
+MVMint32 MVM_unicode_normalizer_process_codepoint_full(MVMThreadContext *tc, MVMNormalizer *norm, MVMCodepoint in, MVMCodepoint *out) {
     MVMint64 qc_in, ccc_in;
+    int is_prepend = is_grapheme_prepend(tc, in);
+
+    if (0 < norm->prepend_buffer)
+        norm->prepend_buffer--;
+    if (is_prepend)
+        norm->prepend_buffer = 2;
 
     /* If it's a control character (outside of the range we checked in the
      * fast path) then it's a normalization terminator. */
-    if (in > 0xFF && is_control_beyond_latin1(tc, in))
-        return MVM_unicode_normalizer_process_codepoint_norm_terminator(tc, n, in, out);
+    if (in > 0xFF && is_control_beyond_latin1(tc, in) && !is_prepend) {
+        return MVM_unicode_normalizer_process_codepoint_norm_terminator(tc, norm, in, out);
+    }
 
     /* Do a quickcheck on the codepoint we got in and get its CCC. */
-    qc_in  = passes_quickcheck(tc, n, in);
+    qc_in  = passes_quickcheck(tc, norm, in);
     ccc_in = ccc(tc, in);
-
-    /* Fast cases when we pass quick check and what we got in has CCC = 0. */
-    if (qc_in && ccc_in == 0) {
-        if (MVM_NORMALIZE_COMPOSE(n->form)) {
+    /* Fast cases when we pass quick check and what we got in has CCC = 0,
+     * and it does not follow a prepend character. */
+    if (qc_in && ccc_in == 0 && norm->prepend_buffer == 0) {
+        if (MVM_NORMALIZE_COMPOSE(norm->form)) {
             /* We're composing. If we have exactly one thing in the buffer and
              * it also passes the quick check, and both it and the thing in the
              * buffer have a CCC of zero, we can hand back the first of the
@@ -598,11 +657,11 @@ MVMint32 MVM_unicode_normalizer_process_codepoint_full(MVMThreadContext *tc, MVM
              * codepoint coming in. Note that the NFG quick-check property
              * factors in grapheme extenders that don't have a CCC of zero,
              * so we're safe. */
-            if (n->buffer_end - n->buffer_start == 1) {
-                MVMCodepoint maybe_result = n->buffer[n->buffer_start];
-                if (passes_quickcheck(tc, n, maybe_result) && ccc(tc, maybe_result) == 0) {
-                    *out = n->buffer[n->buffer_start];
-                    n->buffer[n->buffer_start] = in;
+            if (norm->buffer_end - norm->buffer_start == 1) {
+                MVMCodepoint maybe_result = norm->buffer[norm->buffer_start];
+                if (passes_quickcheck(tc, norm, maybe_result) && ccc(tc, maybe_result) == 0) {
+                    *out = norm->buffer[norm->buffer_start];
+                    norm->buffer[norm->buffer_start] = in;
                     return 1;
                 }
             }
@@ -610,7 +669,7 @@ MVMint32 MVM_unicode_normalizer_process_codepoint_full(MVMThreadContext *tc, MVM
         else {
             /* We're only decomposing. There should probably be nothing in the
              * buffer in this case; if so we can simply return the codepoint. */
-            if (n->buffer_start == n->buffer_end) {
+            if (norm->buffer_start == norm->buffer_end) {
                 *out = in;
                 return 1;
             }
@@ -618,26 +677,26 @@ MVMint32 MVM_unicode_normalizer_process_codepoint_full(MVMThreadContext *tc, MVM
     }
 
     /* If we didn't pass quick check... */
-    if (!qc_in) {
+    if (!qc_in || 0 < norm->prepend_buffer) {
         /* If we're composing, then decompose the last thing placed in the
          * buffer, if any. We need to do this since it may have passed
          * quickcheck, but having seen some character that does pass then we
          * must make sure we decomposed the prior passing one too. */
-        if (MVM_NORMALIZE_COMPOSE(n->form) && n->buffer_end != n->buffer_norm_end) {
-            MVMCodepoint decomp = n->buffer[n->buffer_end - 1];
-            n->buffer_end--;
-            decomp_codepoint_to_buffer(tc, n, decomp);
+        if (MVM_NORMALIZE_COMPOSE(norm->form) && norm->buffer_end != norm->buffer_norm_end && !is_prepend) {
+            MVMCodepoint decomp = norm->buffer[norm->buffer_end - 1];
+            norm->buffer_end--;
+            decomp_codepoint_to_buffer(tc, norm, decomp);
         }
 
         /* Decompose this new character into the buffer. We'll need to see
          * more before we can go any further. */
-        decomp_codepoint_to_buffer(tc, n, in);
+        decomp_codepoint_to_buffer(tc, norm, in);
         return 0;
     }
 
     /* Since anything we have at this point does pass quick check, add it to
      * the buffer directly. */
-    add_codepoint_to_buffer(tc, n, in);
+    add_codepoint_to_buffer(tc, norm, in);
 
     /* If the codepoint has a CCC that is non-zero, it's not a starter so we
      * should see more before normalizing. */
@@ -646,28 +705,28 @@ MVMint32 MVM_unicode_normalizer_process_codepoint_full(MVMThreadContext *tc, MVM
 
     /* If we don't have at least one codepoint in the buffer, it's too early
      * to hand anything back. */
-    if (n->buffer_end - n->buffer_start <= 1)
+    if (norm->buffer_end - norm->buffer_start <= 1)
         return 0;
 
     /* Perform canonical sorting on everything from the start of the not yet
      * normalized things in the buffer, up to but excluding the quick-check
      * passing thing we just added. */
-    canonical_sort(tc, n, n->buffer_norm_end, n->buffer_end - 1);
+    canonical_sort(tc, norm, norm->buffer_norm_end, norm->buffer_end - 1);
 
     /* Perform canonical composition and grapheme composition if needed. */
-    if (MVM_NORMALIZE_COMPOSE(n->form)) {
-        canonical_composition(tc, n, n->buffer_norm_end, n->buffer_end - 1);
-        if (MVM_NORMALIZE_GRAPHEME(n->form))
-            grapheme_composition(tc, n, n->buffer_norm_end, n->buffer_end - 1);
+    if (MVM_NORMALIZE_COMPOSE(norm->form)) {
+        canonical_composition(tc, norm, norm->buffer_norm_end, norm->buffer_end - 1);
+        if (MVM_NORMALIZE_GRAPHEME(norm->form))
+            grapheme_composition(tc, norm, norm->buffer_norm_end, norm->buffer_end - 1);
     }
 
     /* We've now normalized all except the latest, quick-check-passing
      * codepoint. */
-    n->buffer_norm_end = n->buffer_end - 1;
+    norm->buffer_norm_end = norm->buffer_end - 1;
 
     /* Hand back a codepoint, and flag how many more are available. */
-    *out = n->buffer[n->buffer_start];
-    return n->buffer_norm_end - n->buffer_start++;
+    *out = norm->buffer[norm->buffer_start];
+    return norm->buffer_norm_end - norm->buffer_start++;
 }
 
 /* Push a number of codepoints into the "to normalize" buffer. */
@@ -705,7 +764,9 @@ void MVM_unicode_normalizer_eof(MVMThreadContext *tc, MVMNormalizer *n) {
         if (MVM_NORMALIZE_GRAPHEME(n->form))
             grapheme_composition(tc, n, n->buffer_norm_end, n->buffer_end);
     }
-
+    /* Reset these two to ensure their value doesn't stick around */
+    n->prepend_buffer     = 0;
+    n->regional_indicator = 0;
     /* We've now normalized all that remains. */
     n->buffer_norm_end = n->buffer_end;
 }
