@@ -1,5 +1,5 @@
+#include "platform/memmem.h"
 #include "moar.h"
-
 #define MVM_DEBUG_STRANDS 0
 
 #if MVM_DEBUG_STRANDS
@@ -38,6 +38,12 @@ static void copy_strands(MVMThreadContext *tc, const MVMString *from, MVMuint16
         num_strands * sizeof(MVMStringStrand));
 }
 
+MVM_STATIC_INLINE int can_fit_into_8bit (MVMGrapheme32 g) {
+    return -128 <= g && g <= 127;
+}
+MVM_STATIC_INLINE int can_fit_into_ascii (MVMGrapheme32 g) {
+    return 0 <= g && g <= 127;
+}
 /* If a string is currently using 32bit storage, turn it into using
  * 8 bit storage. Doesn't do any checks at all. */
 static void turn_32bit_into_8bit_unchecked(MVMThreadContext *tc, MVMString *str) {
@@ -59,7 +65,6 @@ static MVMString * collapse_strands(MVMThreadContext *tc, MVMString *orig) {
     MVMStringIndex   i;
     MVMuint32        ographs;
     MVMGraphemeIter  gi;
-    MVMint8          can_use_8bit = 1;
 
     MVMROOT(tc, orig, {
         result = (MVMString *)MVM_repr_alloc_init(tc, tc->instance->VMString);
@@ -72,14 +77,18 @@ static MVMString * collapse_strands(MVMThreadContext *tc, MVMString *orig) {
     MVM_string_gi_init(tc, &gi, orig);
     for (i = 0; i < ographs; i++) {
         MVMGrapheme32 g = MVM_string_gi_get_grapheme(tc, &gi);
-        if (g < -127 || g > 127)
-            can_use_8bit = 0;
         result->body.storage.blob_32[i] = g;
+        if (!can_fit_into_8bit(g)) {
+            /* If we know we can't fit into 8 bits, enter a tighter loop for maximum speed */
+            for (i++; i < ographs; i++) {
+                result->body.storage.blob_32[i] = MVM_string_gi_get_grapheme(tc, &gi);
+            }
+            return result;
+        }
     }
-
-    if (can_use_8bit)
-        turn_32bit_into_8bit_unchecked(tc, result);
-
+    /* If we get here, we didn't see any cp's lower than -127 or higher than 127
+     * so turn it into an 8 bit string */
+    turn_32bit_into_8bit_unchecked(tc, result);
     return result;
 }
 
@@ -197,75 +206,118 @@ MVMGrapheme32 MVM_string_get_grapheme_at_nocheck(MVMThreadContext *tc, MVMString
 }
 
 /* Returns the location of one string in another or -1  */
-MVMint64 MVM_string_index(MVMThreadContext *tc, MVMString *haystack, MVMString *needle, MVMint64 start) {
-    MVMint64 result        = -1;
+MVMint64 MVM_string_index(MVMThreadContext *tc, MVMString *Haystack, MVMString *needle, MVMint64 start) {
     size_t index           = (size_t)start;
-    MVMStringIndex hgraphs = MVM_string_graphs(tc, haystack), ngraphs = MVM_string_graphs(tc, needle);
-
-    MVM_string_check_arg(tc, haystack, "index search target");
+    MVMStringIndex H_graphs = MVM_string_graphs(tc, Haystack), n_graphs = MVM_string_graphs(tc, needle);
+    MVM_string_check_arg(tc, Haystack, "index search target");
     MVM_string_check_arg(tc, needle, "index search term");
 
-    if (!ngraphs)
-        return start <= hgraphs ? start : -1; /* the empty string is in any other string */
+    if (!n_graphs)
+        return start <= H_graphs ? start : -1; /* the empty string is in any other string */
 
-    if (!hgraphs)
+    if (!H_graphs)
         return -1;
 
-    if (start < 0 || start >= hgraphs)
+    if (start < 0 || start >= H_graphs)
         return -1;
 
-    if (ngraphs > hgraphs || ngraphs < 1)
+    if (n_graphs > H_graphs || n_graphs < 1)
         return -1;
 
+    /* Fast paths when storage types are identical. Uses memmem function, which
+     * uses Knuth-Morris-Pratt algorithm on Linux and on others
+     * Crochemore+Perrin two-way string matching */
+    switch (Haystack->body.storage_type) {
+        case MVM_STRING_GRAPHEME_32:
+            if (needle->body.storage_type == MVM_STRING_GRAPHEME_32) {
+                void *start_ptr = Haystack->body.storage.blob_32 + start;
+                void *mm_return_32;
+                void *end_ptr = (char*)start_ptr + sizeof(MVMGrapheme32) * (H_graphs - start);
+                do {
+                    /* Keep as void* to not lose precision */
+                    mm_return_32 = MVM_memmem(
+                        start_ptr, /* start position */
+                        (char*)end_ptr - (char*)start_ptr, /* length of Haystack from start position to end */
+                        needle->body.storage.blob_32, /* needle start */
+                        n_graphs * sizeof(MVMGrapheme32) /* needle length */
+                    );
+                    if (mm_return_32 == NULL)
+                        return -1;
+                } /* If we aren't on a 32 bit boundary then continue from where we left off (unlikely but possible) */
+                while ( ( (char*)mm_return_32 - (char*)Haystack->body.storage.blob_32) % sizeof(MVMGrapheme32)
+                    && ( start_ptr = mm_return_32 ) /* Set the new start pointer at where we left off */
+                    && ( end_ptr > start_ptr ) /* Check we aren't past the end of the string just in case */
+                );
+
+                return (MVMGrapheme32*)mm_return_32 - Haystack->body.storage.blob_32;
+            }
+            break;
+        case MVM_STRING_GRAPHEME_8:
+            if (needle->body.storage_type == MVM_STRING_GRAPHEME_8) {
+                void *mm_return_8 = MVM_memmem(
+                    Haystack->body.storage.blob_8 + start, /* start position */
+                    (H_graphs - start) * sizeof(MVMGrapheme8), /* length of Haystack from start position to end */
+                    needle->body.storage.blob_8, /* needle start */
+                    n_graphs * sizeof(MVMGrapheme8) /* needle length */
+                );
+                if (mm_return_8 == NULL)
+                    return -1;
+                else
+                    return (MVMGrapheme8*)mm_return_8 -  Haystack->body.storage.blob_8;
+            }
+            break;
+    }
+
     /* brute force for now. horrible, yes. halp. */
-    while (index <= hgraphs - ngraphs) {
-        if (MVM_string_substrings_equal_nocheck(tc, needle, 0, ngraphs, haystack, index)) {
-            result = (MVMint64)index;
-            break;
+    while (index <= H_graphs - n_graphs) {
+        if (MVM_string_substrings_equal_nocheck(tc, needle, 0, n_graphs, Haystack, index)) {
+            return (MVMint64)index;
         }
         index++;
     }
-    return result;
+    return -1;
 }
 
 /* Returns the location of one string in another or -1  */
-MVMint64 MVM_string_index_from_end(MVMThreadContext *tc, MVMString *haystack, MVMString *needle, MVMint64 start) {
+MVMint64 MVM_string_index_from_end(MVMThreadContext *tc, MVMString *Haystack, MVMString *needle, MVMint64 start) {
     MVMint64 result = -1;
     size_t index;
-    MVMStringIndex hgraphs = MVM_string_graphs(tc, haystack), ngraphs = MVM_string_graphs(tc, needle);
+    MVMStringIndex H_graphs, n_graphs;
 
-    MVM_string_check_arg(tc, haystack, "rindex search target");
+    MVM_string_check_arg(tc, Haystack, "rindex search target");
     MVM_string_check_arg(tc, needle, "rindex search term");
+    H_graphs = MVM_string_graphs_nocheck(tc, Haystack);
+    n_graphs = MVM_string_graphs_nocheck(tc, needle);
 
-    if (!ngraphs) {
+    if (!n_graphs) {
         if (start >= 0)
-            return start <= hgraphs ? start : -1; /* the empty string is in any other string */
+            return start <= H_graphs ? start : -1; /* the empty string is in any other string */
         else
-            return hgraphs; /* no start, so return end */
+            return H_graphs; /* no start, so return end */
     }
 
-    if (!hgraphs)
+    if (!H_graphs)
         return -1;
 
-    if (ngraphs > hgraphs || ngraphs < 1)
+    if (n_graphs > H_graphs || n_graphs < 1)
         return -1;
 
     if (start == -1)
-        start = hgraphs - ngraphs;
+        start = H_graphs - n_graphs;
 
-    if (start < 0 || start >= hgraphs)
+    if (start < 0 || start >= H_graphs)
         /* maybe return -1 instead? */
         MVM_exception_throw_adhoc(tc, "index start offset out of range");
 
     index = start;
 
-    if (index + ngraphs > hgraphs) {
-        index = hgraphs - ngraphs;
+    if (index + n_graphs > H_graphs) {
+        index = H_graphs - n_graphs;
     }
 
     /* brute force for now. horrible, yes. halp. */
     do {
-        if (MVM_string_substrings_equal_nocheck(tc, needle, 0, ngraphs, haystack, index)) {
+        if (MVM_string_substrings_equal_nocheck(tc, needle, 0, n_graphs, Haystack, index)) {
             result = (MVMint64)index;
             break;
         }
@@ -278,10 +330,11 @@ MVMString * MVM_string_substring(MVMThreadContext *tc, MVMString *a, MVMint64 of
     MVMString *result;
     MVMint64   start_pos, end_pos;
 
-    /* convert to signed to avoid implicit arithmetic conversions */
-    MVMint64 agraphs = (MVMint64)MVM_string_graphs(tc, a);
+    MVMint64 agraphs;
 
     MVM_string_check_arg(tc, a, "substring");
+    /* convert to signed to avoid implicit arithmetic conversions */
+    agraphs = (MVMint64)MVM_string_graphs_nocheck(tc, a);
 
     /* -1 signifies go to the end of the string; anything less is a bug */
     if (length < -1)
@@ -350,7 +403,7 @@ MVMString * MVM_string_substring(MVMThreadContext *tc, MVMString *a, MVMint64 of
             MVM_string_gi_move_to(tc, &gi, start_pos);
             for (i = 0; i < result->body.num_graphs; i++) {
                 MVMGrapheme32 g = MVM_string_gi_get_grapheme(tc, &gi);
-                if (g < -128 || g >= 128)
+                if (!can_fit_into_8bit(g))
                     can_fit_into_8 = 0;
                 result->body.storage.blob_32[i] = g;
             }
@@ -397,18 +450,26 @@ static MVMint32 final_strand_matches(MVMThreadContext *tc, MVMString *a, MVMStri
 MVMString * MVM_string_concatenate(MVMThreadContext *tc, MVMString *a, MVMString *b) {
     MVMString *result;
     MVMuint32  agraphs, bgraphs;
+    MVMuint64  total_graphs;
 
     MVM_string_check_arg(tc, a, "concatenate");
     MVM_string_check_arg(tc, b, "concatenate");
 
     /* Simple empty-string cases. */
-    agraphs = MVM_string_graphs(tc, a);
+    agraphs = MVM_string_graphs_nocheck(tc, a);
     if (agraphs == 0)
         return b;
     bgraphs = MVM_string_graphs(tc, b);
     if (bgraphs == 0)
         return a;
 
+    /* Total size of the resulting string can't be bigger than an MVMString is allowed to be. */
+    total_graphs = (MVMuint64)agraphs + (MVMuint64)bgraphs;
+    if (total_graphs > 0xFFFFFFFF)
+        MVM_exception_throw_adhoc(tc,
+            "Can't concatenate strings, required number of graphemes %"PRIu64" > max allowed of %u",
+             total_graphs, 0xFFFFFFFF);
+
     /* Otherwise, we'll assemble a result string. */
     MVMROOT(tc, a, {
     MVMROOT(tc, b, {
@@ -416,7 +477,7 @@ MVMString * MVM_string_concatenate(MVMThreadContext *tc, MVMString *a, MVMString
         result = (MVMString *)MVM_repr_alloc_init(tc, tc->instance->VMString);
 
         /* Total graphemes is trivial; just total up inputs. */
-        result->body.num_graphs = agraphs + bgraphs;
+        result->body.num_graphs = (MVMuint32)total_graphs;
 
         /* Result string will be made of strands. */
         result->body.storage_type = MVM_STRING_STRAND;
@@ -491,6 +552,7 @@ MVMString * MVM_string_concatenate(MVMThreadContext *tc, MVMString *a, MVMString
 MVMString * MVM_string_repeat(MVMThreadContext *tc, MVMString *a, MVMint64 count) {
     MVMString *result;
     MVMuint32  agraphs;
+    MVMuint64  total_graphs;
 
     MVM_string_check_arg(tc, a, "repeat");
 
@@ -505,17 +567,23 @@ MVMString * MVM_string_repeat(MVMThreadContext *tc, MVMString *a, MVMint64 count
         MVM_exception_throw_adhoc(tc, "repeat count > %d arbitrarily unsupported...", (1 << 30));
 
     /* If input string is empty, repeating it is empty. */
-    agraphs = MVM_string_graphs(tc, a);
+    agraphs = MVM_string_graphs_nocheck(tc, a);
     if (agraphs == 0)
         return tc->instance->str_consts.empty;
 
+    /* Total size of the resulting string can't be bigger than an MVMString is allowed to be. */
+    total_graphs = (MVMuint64)agraphs * (MVMuint64)count;
+    if (total_graphs > 0xFFFFFFFF)
+        MVM_exception_throw_adhoc(tc,
+            "Can't repeat string, required number of graphemes %"PRIu64" > max allowed of %u",
+             total_graphs, 0xFFFFFFFF);
+
     /* Now build a result string with the repetition set. */
     MVMROOT(tc, a, {
         result = (MVMString *)MVM_repr_alloc_init(tc, tc->instance->VMString);
         result->body.num_graphs      = agraphs * count;
         result->body.storage_type    = MVM_STRING_STRAND;
         result->body.storage.strands = allocate_strands(tc, 1);
-        result->body.num_strands     = 1;
         if (a->body.storage_type == MVM_STRING_STRAND) {
             if (a->body.num_strands == 1 && a->body.storage.strands[0].repetitions == 0) {
                 copy_strands(tc, a, 0, result, 0, 1);
@@ -535,6 +603,7 @@ MVMString * MVM_string_repeat(MVMThreadContext *tc, MVMString *a, MVMint64 count
             result->body.storage.strands[0].end         = agraphs;
         }
         result->body.storage.strands[0].repetitions = count - 1;
+        result->body.num_strands = 1;
     });
 
     STRAND_CHECK(tc, result);
@@ -543,14 +612,33 @@ MVMString * MVM_string_repeat(MVMThreadContext *tc, MVMString *a, MVMint64 count
 
 void MVM_string_say(MVMThreadContext *tc, MVMString *a) {
     MVM_string_check_arg(tc, a, "say");
-    MVM_io_write_string(tc, tc->instance->stdout_handle, a, 1);
+    MVM_string_print(tc, MVM_string_concatenate(tc, a,
+        tc->instance->str_consts.platform_newline));
 }
 
 void MVM_string_print(MVMThreadContext *tc, MVMString *a) {
+    MVMOSHandle *handle = (MVMOSHandle *)tc->instance->stdout_handle;
+    MVMuint64 encoded_size;
+    char *encoded;
     MVM_string_check_arg(tc, a, "print");
-    MVM_io_write_string(tc, tc->instance->stdout_handle, a, 0);
+    encoded = MVM_string_utf8_encode(tc, a, &encoded_size, MVM_TRANSLATE_NEWLINE_OUTPUT);
+    MVM_io_write_bytes_c(tc, tc->instance->stdout_handle, encoded, encoded_size);
+    MVM_free(encoded);
 }
+/* Meant to be pased in a MVMNormalizer of type MVM_NORMALIZE_NFD */
+static MVMGrapheme32 ord_getbasechar (MVMThreadContext *tc, MVMGrapheme32 g) {
+    MVMGrapheme32 return_g;
+    MVMint32 ready;
+    MVMNormalizer norm;
+    MVM_unicode_normalizer_init(tc, &norm, MVM_NORMALIZE_NFD);
 
+    ready = MVM_unicode_normalizer_process_codepoint_to_grapheme(tc, &norm, g, &return_g);
+    MVM_unicode_normalizer_eof(tc, &norm);
+    if (!ready)
+        return_g = MVM_unicode_normalizer_get_grapheme(tc, &norm);
+    MVM_unicode_normalizer_cleanup(tc, &norm);
+    return return_g;
+}
 /* Tests whether one string a has the other string b as a substring at that index */
 MVMint64 MVM_string_equal_at(MVMThreadContext *tc, MVMString *a, MVMString *b, MVMint64 offset) {
 
@@ -559,8 +647,8 @@ MVMint64 MVM_string_equal_at(MVMThreadContext *tc, MVMString *a, MVMString *b, M
     MVM_string_check_arg(tc, a, "equal_at");
     MVM_string_check_arg(tc, b, "equal_at");
 
-    agraphs = MVM_string_graphs(tc, a);
-    bgraphs = MVM_string_graphs(tc, b);
+    agraphs = MVM_string_graphs_nocheck(tc, a);
+    bgraphs = MVM_string_graphs_nocheck(tc, b);
 
     if (offset < 0) {
         offset += agraphs;
@@ -571,17 +659,150 @@ MVMint64 MVM_string_equal_at(MVMThreadContext *tc, MVMString *a, MVMString *b, M
         return 0;
     return MVM_string_substrings_equal_nocheck(tc, a, offset, bgraphs, b, 0);
 }
+/* Ensure return value can hold numbers at least 3x higher than MVMStringIndex.
+ * Theoretically if the string has all ﬃ ligatures and 1/3 the max size of
+ * MVMStringIndex in length, we could have some weird results. */
 
-MVMint64 MVM_string_equal_at_ignore_case(MVMThreadContext *tc, MVMString *a, MVMString *b, MVMint64 offset) {
-    MVMString *lca;
-    MVMString *lcb;
-    MVMROOT(tc, b, {
-        lca = MVM_string_lc(tc, a);
-        MVMROOT(tc, lca, {
-            lcb = MVM_string_lc(tc, b);
-        });
+/* ignoremark is 0 for normal operation and 1 for ignoring diacritics */
+MVM_STATIC_INLINE MVMint64 string_equal_at_ignore_case_INTERNAL_loop(MVMThreadContext *tc, MVMString *Haystack, MVMString *needle_fc, MVMint64 H_start, MVMint64 H_graphs, MVMint64 n_fc_graphs, int ignoremark) {
+    MVMuint32 H_fc_cps;
+    /* An additional needle offset which is used only when codepoints expand
+     * when casefolded. The offset is the number of additional codepoints that
+     * have been seen so Haystack and needle stay aligned */
+    MVMint64 n_offset = 0;
+    MVMint64 i, j;
+    MVMGrapheme32 H_g, n_g;
+    for (i = 0; i + H_start < H_graphs && i + n_offset < n_fc_graphs; i++) {
+        const MVMCodepoint* H_result_cps;
+        H_g = MVM_string_get_grapheme_at_nocheck(tc, Haystack, H_start + i);
+        if (H_g >= 0 ) {
+            /* For codeponits we can get the case change directly */
+            H_fc_cps = MVM_unicode_get_case_change(tc, H_g, MVM_unicode_case_change_type_fold, &H_result_cps);
+        }
+        else {
+            /* Synthetics must use this function */
+            H_fc_cps = MVM_nfg_get_case_change(tc, H_g, MVM_unicode_case_change_type_fold, (MVMGrapheme32**) &H_result_cps);
+        }
+        /* If we get 0 for the number that means the cp doesn't change when casefolded */
+        if (H_fc_cps == 0) {
+            n_g = MVM_string_get_grapheme_at_nocheck(tc, needle_fc, i + n_offset);
+            if (ignoremark) {
+                H_g = ord_getbasechar(tc, H_g);
+                n_g = ord_getbasechar(tc, n_g);
+            }
+            if (H_g != n_g)
+                return -1;
+        }
+        else if (H_fc_cps >= 1) {
+            for (j = 0; j < H_fc_cps; j++) {
+                n_g = MVM_string_get_grapheme_at_nocheck(tc, needle_fc, i + n_offset);
+                H_g = H_result_cps[j];
+                if (ignoremark) {
+                    H_g = ord_getbasechar(tc, H_g);
+                    n_g = ord_getbasechar(tc, n_g);
+                }
+                if (H_g != n_g)
+                    return -1;
+                n_offset++;
+            }
+            n_offset--;
+        }
+    }
+    return n_offset;
+    /* We return -1 if the strings are not equal and 0 or more if they are equal
+     * The return values from 0, 1 etc designate how many Haystack graphemes
+     * were expanded.
+     * This may seem like an odd arangement, but this extra information is needed
+     * to determine the length of the Haystack which was traversed, as it can
+     * differ from the length of the needle if there are expansions. */
+}
+/* Checks if needle exists at the offset, but ignores case.
+ * Sometimes there is a difference in length of a string before and after foldcase,
+ * because of this we must compare this differently than just foldcasing both
+ * strings to ensure the offset is correct */
+static MVMint64 string_equal_at_ignore_case(MVMThreadContext *tc, MVMString *Haystack, MVMString *needle, MVMint64 H_offset, int ignoremark) {
+    /* Foldcase version of needle */
+    MVMString *needle_fc;
+    MVMStringIndex H_graphs = MVM_string_graphs(tc, Haystack);
+    MVMStringIndex n_graphs = MVM_string_graphs(tc, needle);
+    MVMStringIndex n_fc_graphs;
+    /* H_expansion must be able to hold integers 3x larger than MVMStringIndex */
+    MVMint64 H_expansion;
+
+    if (H_offset < 0) {
+        H_offset += H_graphs;
+        if (H_offset < 0)
+            H_offset = 0; /* XXX I think this is the right behavior here */
+    }
+    /* If the offset is greater or equal to the number of Haystack graphemes
+     * return 0. Since size of graphemes could change under casefolding, we
+     * can't assume too much. If optimizing this be careful */
+    if (H_offset >= H_graphs)
+        return 0;
+
+    MVMROOT(tc, Haystack, {
+        needle_fc = MVM_string_fc(tc, needle);
     });
-    return MVM_string_equal_at(tc, lca, lcb, offset);
+    n_fc_graphs = MVM_string_graphs(tc, needle_fc);
+    H_expansion = string_equal_at_ignore_case_INTERNAL_loop(tc, Haystack, needle_fc, H_offset, H_graphs, n_fc_graphs, ignoremark);
+    if (H_expansion >= 0)
+        return H_graphs + H_expansion - H_offset >= n_fc_graphs  ? 1 : 0;
+    return 0;
+}
+static MVMint64 string_index_ignore_case(MVMThreadContext *tc, MVMString *Haystack, MVMString *needle, MVMint64 start, int ignoremark) {
+    /* Foldcase version of needle */
+    MVMString *needle_fc;
+    MVMStringIndex n_fc_graphs;
+
+    size_t index           = (size_t)start;
+    MVMStringIndex H_graphs, n_graphs;
+    /* H_expansion must be able to hold integers 3x larger than MVMStringIndex */
+    MVMint64 H_expansion;
+    MVMint64 return_val = -1;
+    MVM_string_check_arg(tc, Haystack, ignoremark ? "index ignore case ignore mark search target" : "index ignore case search target");
+    MVM_string_check_arg(tc, needle,   ignoremark ? "index ignore case ignore mark search term"   : "index ignore case search term");
+    H_graphs = MVM_string_graphs(tc, Haystack);
+    n_graphs = MVM_string_graphs(tc, needle);
+    if (!n_graphs)
+        return start <= H_graphs ? start : -1; /* Empty string is in any other string */
+    if (!H_graphs)
+        return -1;
+    if (start < 0 || start >= H_graphs)
+        return -1;
+    /* Codepoints can expand into up to THREE codepoints (as of Unicode 9.0). The next check
+     * checks if it is at all possible for the needle grapheme number to be higher
+     * than the Haystack */
+    if (n_graphs > H_graphs * 3)
+        return -1;
+
+    if (n_graphs < 1)
+        return -1;
+
+    MVMROOT(tc, Haystack, {
+        needle_fc = MVM_string_fc(tc, needle);
+    });
+    n_fc_graphs = MVM_string_graphs(tc, needle_fc);
+    /* brute force for now. horrible, yes. halp. */
+    while (index <= H_graphs) {
+        H_expansion = string_equal_at_ignore_case_INTERNAL_loop(tc, Haystack, needle_fc, index, H_graphs, n_fc_graphs, ignoremark);
+        if (H_expansion >= 0)
+            return H_graphs + H_expansion - index >= n_fc_graphs  ? (MVMint64)index : -1;
+        index++;
+    }
+    return -1;
+}
+
+MVMint64 MVM_string_equal_at_ignore_case(MVMThreadContext *tc, MVMString *Haystack, MVMString *needle, MVMint64 H_offset) {
+    return string_equal_at_ignore_case(tc, Haystack, needle, H_offset, 0);
+}
+MVMint64 MVM_string_equal_at_ignore_case_ignore_mark(MVMThreadContext *tc, MVMString *Haystack, MVMString *needle, MVMint64 H_offset) {
+    return string_equal_at_ignore_case(tc, Haystack, needle, H_offset, 1);
+}
+MVMint64 MVM_string_index_ignore_case(MVMThreadContext *tc, MVMString *Haystack, MVMString *needle, MVMint64 start) {
+    return string_index_ignore_case(tc, Haystack, needle, start, 0);
+}
+MVMint64 MVM_string_index_ignore_case_ignore_mark(MVMThreadContext *tc, MVMString *Haystack, MVMString *needle, MVMint64 start) {
+    return string_index_ignore_case(tc, Haystack, needle, start, 1);
 }
 
 MVMGrapheme32 MVM_string_ord_at(MVMThreadContext *tc, MVMString *s, MVMint64 offset) {
@@ -599,10 +820,10 @@ MVMGrapheme32 MVM_string_ord_at(MVMThreadContext *tc, MVMString *s, MVMint64 off
     return g >= 0 ? g : MVM_nfg_get_synthetic_info(tc, g)->base;
 }
 
+/* Gets the base character at a grapheme position, ignoring things like diacritics */
 MVMGrapheme32 MVM_string_ord_basechar_at(MVMThreadContext *tc, MVMString *s, MVMint64 offset) {
     MVMStringIndex agraphs;
     MVMGrapheme32 g;
-    MVMNormalizer norm;
     MVMint32 ready;
 
     MVM_string_check_arg(tc, s, "ord_basechar_at");
@@ -618,18 +839,13 @@ MVMGrapheme32 MVM_string_ord_basechar_at(MVMThreadContext *tc, MVMString *s, MVM
         g = si->base;
     }
     else {
-        MVM_unicode_normalizer_init(tc, &norm, MVM_NORMALIZE_NFD);
-        ready = MVM_unicode_normalizer_process_codepoint_to_grapheme(tc, &norm, g, &g);
-        MVM_unicode_normalizer_eof(tc, &norm);
-        if (!ready)
-            g = MVM_unicode_normalizer_get_grapheme(tc, &norm);
-
-        MVM_unicode_normalizer_cleanup(tc, &norm);
+        g = ord_getbasechar(tc, g);
     }
 
     return g;
 }
 
+
 /* Compares two strings for equality. */
 MVMint64 MVM_string_equal(MVMThreadContext *tc, MVMString *a, MVMString *b) {
     MVM_string_check_arg(tc, a, "equal");
@@ -890,7 +1106,7 @@ void MVM_string_encode_to_buf(MVMThreadContext *tc, MVMString *s, MVMString *enc
 
     /* Ensure the target is in the correct form. */
     MVM_string_check_arg(tc, s, "encode");
-    if (!IS_CONCRETE(buf) || REPR(buf)->ID != MVM_REPR_ID_MVMArray)
+    if (!IS_CONCRETE(buf) || REPR(buf)->ID != MVM_REPR_ID_VMArray)
         MVM_exception_throw_adhoc(tc, "encode requires a native array to write into");
     buf_rd = (MVMArrayREPRData *)STABLE(buf)->REPR_data;
     if (buf_rd) {
@@ -934,7 +1150,7 @@ MVMString * MVM_string_decode_from_buf(MVMThreadContext *tc, MVMObject *buf, MVM
     MVMuint8 elem_size = 0;
 
     /* Ensure the source is in the correct form. */
-    if (!IS_CONCRETE(buf) || REPR(buf)->ID != MVM_REPR_ID_MVMArray)
+    if (!IS_CONCRETE(buf) || REPR(buf)->ID != MVM_REPR_ID_VMArray)
         MVM_exception_throw_adhoc(tc, "decode requires a native array to read from");
     buf_rd = (MVMArrayREPRData *)STABLE(buf)->REPR_data;
     if (buf_rd) {
@@ -975,8 +1191,8 @@ MVMObject * MVM_string_split(MVMThreadContext *tc, MVMString *separator, MVMStri
         result = MVM_repr_alloc_init(tc, hll->slurpy_array_type);
         MVMROOT(tc, result, {
             start = 0;
-            end = MVM_string_graphs(tc, input);
-            sep_length = MVM_string_graphs(tc, separator);
+            end = MVM_string_graphs_nocheck(tc, input);
+            sep_length = MVM_string_graphs_nocheck(tc, separator);
 
             while (start < end) {
                 MVMString *portion;
@@ -1039,7 +1255,7 @@ MVMString * MVM_string_join(MVMThreadContext *tc, MVMString *separator, MVMObjec
     /* Take a first pass through the string, counting up length and the total
      * number of strands we encounter as well as building a flat array of the
      * strings (to we only have to do the indirect calls once). */
-    sgraphs  = MVM_string_graphs(tc, separator);
+    sgraphs  = MVM_string_graphs_nocheck(tc, separator);
     if (sgraphs)
         sstrands = separator->body.storage_type == MVM_STRING_STRAND
             ? separator->body.num_strands
@@ -1110,9 +1326,17 @@ MVMString * MVM_string_join(MVMThreadContext *tc, MVMString *separator, MVMObjec
 
             /* Add separator if needed. */
             if (i > 0) {
+                /* If there's no separator and one piece is The Empty String we
+                 * have to be extra careful about concat stability */
+                if (sgraphs == 0 && MVM_string_graphs_nocheck(tc, piece) == 0 && concats_stable
+                        && i + 1 < num_pieces
+                        && !MVM_nfg_is_concat_stable(tc, pieces[i - 1], pieces[i + 1])) {
+                    concats_stable = 0;
+                }
+
                 if (sgraphs) {
                     if (!concats_stable)
-                        /* Already stable; no more checks. */;
+                        /* Already unstable; no more checks. */;
                     else if (!MVM_nfg_is_concat_stable(tc, pieces[i - 1], separator))
                         concats_stable = 0;
                     else if (!MVM_nfg_is_concat_stable(tc, separator, piece))
@@ -1182,11 +1406,11 @@ MVMint64 MVM_string_char_at_in_string(MVMThreadContext *tc, MVMString *a, MVMint
     MVM_string_check_arg(tc, b, "char_at_in_string");
 
     /* We return -2 here only to be able to distinguish between "out of bounds" and "not in string". */
-    if (offset < 0 || offset >= MVM_string_graphs(tc, a))
+    if (offset < 0 || offset >= MVM_string_graphs_nocheck(tc, a))
         return -2;
 
     search  = MVM_string_get_grapheme_at_nocheck(tc, a, offset);
-    bgraphs = MVM_string_graphs(tc, b);
+    bgraphs = MVM_string_graphs_nocheck(tc, b);
     switch (b->body.storage_type) {
     case MVM_STRING_GRAPHEME_32: {
         MVMStringIndex i;
@@ -1196,7 +1420,7 @@ MVMint64 MVM_string_char_at_in_string(MVMThreadContext *tc, MVMString *a, MVMint
         break;
     }
     case MVM_STRING_GRAPHEME_ASCII:
-        if (search >= 0 && search <= 127) {
+        if (can_fit_into_ascii(search)) {
             MVMStringIndex i;
             for (i = 0; i < bgraphs; i++)
                 if (b->body.storage.blob_ascii[i] == search)
@@ -1204,7 +1428,7 @@ MVMint64 MVM_string_char_at_in_string(MVMThreadContext *tc, MVMString *a, MVMint
         }
         break;
     case MVM_STRING_GRAPHEME_8:
-        if (search >= -128 && search <= 127) {
+        if (can_fit_into_8bit(search)) {
             MVMStringIndex i;
             for (i = 0; i < bgraphs; i++)
                 if (b->body.storage.blob_8[i] == search)
@@ -1229,7 +1453,7 @@ MVMint64 MVM_string_offset_has_unicode_property_value(MVMThreadContext *tc, MVMS
 
     MVM_string_check_arg(tc, s, "uniprop");
 
-    if (offset < 0 || offset >= MVM_string_graphs(tc, s))
+    if (offset < 0 || offset >= MVM_string_graphs_nocheck(tc, s))
         return 0;
 
     g = MVM_string_get_grapheme_at_nocheck(tc, s, offset);
@@ -1247,27 +1471,27 @@ MVMint64 MVM_string_offset_has_unicode_property_value(MVMThreadContext *tc, MVMS
 MVMString * MVM_string_indexing_optimized(MVMThreadContext *tc, MVMString *s) {
     MVM_string_check_arg(tc, s, "indexingoptimized");
     if (s->body.storage_type == MVM_STRING_STRAND) {
-        MVMGrapheme32   *flat = MVM_malloc(MVM_string_graphs(tc, s) * sizeof(MVMGrapheme32));
-        MVMStringStrand *orig = s->body.storage.strands;
-        MVMuint32        i    = 0;
+        MVMuint32        chars = MVM_string_graphs_nocheck(tc, s);
+        MVMGrapheme32   *flat  = MVM_malloc(chars * sizeof(MVMGrapheme32));
+        MVMuint32        i     = 0;
         MVMString       *res;
-        MVMint8          can_fit_into_8bit = 1;
+        MVMint8          string_can_fit_into_8bit = 1;
 
         MVMGraphemeIter  gi;
         MVM_string_gi_init(tc, &gi, s);
         while (MVM_string_gi_has_more(tc, &gi)) {
             MVMGrapheme32 g = MVM_string_gi_get_grapheme(tc, &gi);
-            if (g < -128 || g >= 128)
-                can_fit_into_8bit = 0;
+            if (! can_fit_into_8bit(g) )
+                string_can_fit_into_8bit = 0;
             flat[i++] = g;
         }
 
         res = (MVMString *)MVM_repr_alloc_init(tc, tc->instance->VMString);
         res->body.storage_type    = MVM_STRING_GRAPHEME_32;
         res->body.storage.blob_32 = flat;
-        res->body.num_graphs      = MVM_string_graphs(tc, s);
+        res->body.num_graphs      = chars;
 
-        if (can_fit_into_8bit)
+        if (string_can_fit_into_8bit)
             turn_32bit_into_8bit_unchecked(tc, res);
 
         return res;
@@ -1286,11 +1510,11 @@ MVMString * MVM_string_escape(MVMThreadContext *tc, MVMString *s) {
     MVMStringIndex  sgraphs, balloc;
     MVMGrapheme32  *buffer;
     MVMGrapheme32   crlf;
-    MVMint8         can_fit_into_8bit = 1;
+    MVMint8         string_can_fit_into_8bit = 1;
 
     MVM_string_check_arg(tc, s, "escape");
 
-    sgraphs = MVM_string_graphs(tc, s);
+    sgraphs = MVM_string_graphs_nocheck(tc, s);
     balloc  = sgraphs;
     buffer  = MVM_malloc(sizeof(MVMGrapheme32) * balloc);
 
@@ -1333,8 +1557,8 @@ MVMString * MVM_string_escape(MVMThreadContext *tc, MVMString *s) {
                 balloc += 32;
                 buffer = MVM_realloc(buffer, sizeof(MVMGrapheme32) * balloc);
             }
-            if (graph < -128 || graph >= 128)
-                can_fit_into_8bit = 0;
+            if (!can_fit_into_8bit(graph))
+                string_can_fit_into_8bit = 0;
             buffer[bpos++] = graph;
         }
     }
@@ -1344,7 +1568,7 @@ MVMString * MVM_string_escape(MVMThreadContext *tc, MVMString *s) {
     res->body.storage.blob_32 = buffer;
     res->body.num_graphs      = bpos;
 
-    if (can_fit_into_8bit)
+    if (string_can_fit_into_8bit)
         turn_32bit_into_8bit_unchecked(tc, res);
 
     STRAND_CHECK(tc, res);
@@ -1359,7 +1583,7 @@ MVMString * MVM_string_flip(MVMThreadContext *tc, MVMString *s) {
     MVMStringIndex  rpos;
 
     MVM_string_check_arg(tc, s, "flip");
-    sgraphs = MVM_string_graphs(tc, s);
+    sgraphs = MVM_string_graphs_nocheck(tc, s);
     rpos    = sgraphs;
 
     if (s->body.storage_type == MVM_STRING_GRAPHEME_8) {
@@ -1403,8 +1627,8 @@ MVMint64 MVM_string_compare(MVMThreadContext *tc, MVMString *a, MVMString *b) {
     MVM_string_check_arg(tc, b, "compare");
 
     /* Simple cases when one or both are zero length. */
-    alen = MVM_string_graphs(tc, a);
-    blen = MVM_string_graphs(tc, b);
+    alen = MVM_string_graphs_nocheck(tc, a);
+    blen = MVM_string_graphs_nocheck(tc, b);
     if (alen == 0)
         return blen == 0 ? 0 : -1;
     if (blen == 0)
@@ -1434,8 +1658,8 @@ MVMString * MVM_string_bitand(MVMThreadContext *tc, MVMString *a, MVMString *b)
     MVM_string_check_arg(tc, a, "bitwise and");
     MVM_string_check_arg(tc, b, "bitwise and");
 
-    alen = MVM_string_graphs(tc, a);
-    blen = MVM_string_graphs(tc, b);
+    alen = MVM_string_graphs_nocheck(tc, a);
+    blen = MVM_string_graphs_nocheck(tc, b);
     sgraphs = alen < blen ? alen : blen;
     buffer = MVM_malloc(sizeof(MVMGrapheme32) * sgraphs);
 
@@ -1462,8 +1686,8 @@ MVMString * MVM_string_bitor(MVMThreadContext *tc, MVMString *a, MVMString *b) {
     MVM_string_check_arg(tc, a, "bitwise or");
     MVM_string_check_arg(tc, b, "bitwise or");
 
-    alen = MVM_string_graphs(tc, a);
-    blen = MVM_string_graphs(tc, b);
+    alen = MVM_string_graphs_nocheck(tc, a);
+    blen = MVM_string_graphs_nocheck(tc, b);
     sgraphs = (alen > blen ? alen : blen);
     buffer = MVM_malloc(sizeof(MVMGrapheme32) * sgraphs);
 
@@ -1499,8 +1723,8 @@ MVMString * MVM_string_bitxor(MVMThreadContext *tc, MVMString *a, MVMString *b)
     MVM_string_check_arg(tc, a, "bitwise xor");
     MVM_string_check_arg(tc, b, "bitwise xor");
 
-    alen = MVM_string_graphs(tc, a);
-    blen = MVM_string_graphs(tc, b);
+    alen = MVM_string_graphs_nocheck(tc, a);
+    blen = MVM_string_graphs_nocheck(tc, b);
     sgraphs = (alen > blen ? alen : blen);
     buffer = MVM_malloc(sizeof(MVMGrapheme32) * sgraphs);
 
@@ -1708,7 +1932,7 @@ static MVMint64 grapheme_is_cclass(MVMThreadContext *tc, MVMint64 cclass, MVMGra
 MVMint64 MVM_string_is_cclass(MVMThreadContext *tc, MVMint64 cclass, MVMString *s, MVMint64 offset) {
     MVMGrapheme32 g;
     MVM_string_check_arg(tc, s, "is_cclass");
-    if (offset < 0 || offset >= MVM_string_graphs(tc, s))
+    if (offset < 0 || offset >= MVM_string_graphs_nocheck(tc, s))
         return 0;
     g = MVM_string_get_grapheme_at_nocheck(tc, s, offset);
     return grapheme_is_cclass(tc, cclass, g);
@@ -1721,7 +1945,7 @@ MVMint64 MVM_string_find_cclass(MVMThreadContext *tc, MVMint64 cclass, MVMString
 
     MVM_string_check_arg(tc, s, "find_cclass");
 
-    length = MVM_string_graphs(tc, s);
+    length = MVM_string_graphs_nocheck(tc, s);
     end    = offset + count;
     end = length < end ? length : end;
     if (offset < 0 || offset >= length)
@@ -1745,7 +1969,7 @@ MVMint64 MVM_string_find_not_cclass(MVMThreadContext *tc, MVMint64 cclass, MVMSt
 
     MVM_string_check_arg(tc, s, "find_not_cclass");
 
-    length = MVM_string_graphs(tc, s);
+    length = MVM_string_graphs_nocheck(tc, s);
     end    = offset + count;
     end = length < end ? length : end;
     if (offset < 0 || offset >= length)
@@ -1815,26 +2039,37 @@ MVMuint8 MVM_string_find_encoding(MVMThreadContext *tc, MVMString *name) {
     }
 }
 
-/* Turns a codepoint into a string. Uses the normalizer to ensure that we get
- * a valid NFG string (NFG is a superset of NFC, and singleton decompositions
- * exist). */
+/* Turns a codepoint into a string. If required uses the normalizer to ensure
+ * that we get a valid NFG string (NFG is a superset of NFC, and singleton
+ * decompositions exist). */
 MVMString * MVM_string_chr(MVMThreadContext *tc, MVMCodepoint cp) {
     MVMString *s;
     MVMGrapheme32 g;
-    MVMNormalizer norm;
 
     if (cp < 0)
         MVM_exception_throw_adhoc(tc, "chr codepoint cannot be negative");
+    /* If the codepoint decomposes we may need to normalize it.
+     * The first cp that decomposes is U+0340, but to be on the safe side
+     * for now we go with the first significant character which at the time
+     * of writing (Unicode 9.0) is COMBINING GRAVE ACCENT U+300 */
+    if (cp >= MVM_NORMALIZE_FIRST_SIG_NFC
+        && MVM_unicode_codepoint_get_property_int(tc, cp, MVM_UNICODE_PROPERTY_DECOMPOSITION_TYPE)
+        != MVM_UNICODE_PVALUE_DT_NONE) {
 
-    MVM_unicode_normalizer_init(tc, &norm, MVM_NORMALIZE_NFG);
-    if (!MVM_unicode_normalizer_process_codepoint_to_grapheme(tc, &norm, cp, &g)) {
-        MVM_unicode_normalizer_eof(tc, &norm);
-        g = MVM_unicode_normalizer_get_grapheme(tc, &norm);
+        MVMNormalizer norm;
+        MVM_unicode_normalizer_init(tc, &norm, MVM_NORMALIZE_NFG);
+        if (!MVM_unicode_normalizer_process_codepoint_to_grapheme(tc, &norm, cp, &g)) {
+            MVM_unicode_normalizer_eof(tc, &norm);
+            g = MVM_unicode_normalizer_get_grapheme(tc, &norm);
+        }
+        MVM_unicode_normalizer_cleanup(tc, &norm);
+    }
+    else {
+        g = (MVMGrapheme32) cp;
     }
-    MVM_unicode_normalizer_cleanup(tc, &norm);
 
     s = (MVMString *)REPR(tc->instance->VMString)->allocate(tc, STABLE(tc->instance->VMString));
-    if (g >= -128 && g < 128) {
+    if (can_fit_into_8bit(g)) {
         s->body.storage_type       = MVM_STRING_GRAPHEME_8;
         s->body.storage.blob_8     = MVM_malloc(sizeof(MVMGrapheme8));
         s->body.storage.blob_8[0]  = g;
@@ -1896,20 +2131,14 @@ void MVM_string_compute_hash_code(MVMThreadContext *tc, MVMString *s) {
     /* Now handle trailing graphemes (must be 2, 1, or 0). */
     switch (graphs_remaining) {
         case 2:
-            hash_block.graphs[0] = MVM_string_gi_get_grapheme(tc, &gi);
             hash_block.graphs[1] = MVM_string_gi_get_grapheme(tc, &gi);
-            break;
-        case 1:
-            hash_block.graphs[0] = MVM_string_gi_get_grapheme(tc, &gi);
-            break;
-    }
-    switch (graphs_remaining * 4) {
-        case 8:
             _hj_j += ( (unsigned)hash_block.bytes[7] << 24 ) +
                      ( (unsigned)hash_block.bytes[6] << 16 ) +
                      ( (unsigned)hash_block.bytes[5] << 8 ) +
                      hash_block.bytes[4];
-        case 4:
+        /* Fallthrough */
+        case 1:
+            hash_block.graphs[0] = MVM_string_gi_get_grapheme(tc, &gi);
             _hj_i += ( (unsigned)hash_block.bytes[3] << 24 ) +
                      ( (unsigned)hash_block.bytes[2] << 16 ) +
                      ( (unsigned)hash_block.bytes[1] << 8 ) +
