@@ -323,12 +323,15 @@ MVMuint32 MVM_string_utf8_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds
     MVMint32 count = 0, total = 0;
     MVMint32 state = 0;
     MVMCodepoint codepoint = 0;
+    MVMCodepoint lag_codepoint = -1;
     MVMint32 bufsize;
     MVMGrapheme32 *buffer;
     MVMDecodeStreamBytes *cur_bytes;
     MVMDecodeStreamBytes *last_accept_bytes = ds->bytes_head;
-    MVMint32 last_accept_pos, ready, at_start;
+    MVMDecodeStreamBytes *lag_last_accept_bytes;
+    MVMint32 last_accept_pos, lag_last_accept_pos, ready, at_start;
     MVMuint32 reached_stopper;
+    MVMuint32 can_fast_path;
 
     /* If there's no buffers, we're done. */
     if (!ds->bytes_head)
@@ -339,8 +342,11 @@ MVMuint32 MVM_string_utf8_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds
     if (stopper_chars && *stopper_chars == 0)
         return 1;
 
-    /* Rough starting-size estimate is number of bytes in the head buffer. */
-    bufsize = ds->bytes_head->length;
+    /* If there's nothing hanging around in the normalization buffer, we can
+     * use the fast path. */
+    can_fast_path = MVM_unicode_normalizer_empty(tc, &(ds->norm));
+
+    bufsize = ds->result_size_guess;
     buffer = MVM_malloc(bufsize * sizeof(MVMGrapheme32));
 
     /* Decode each of the buffers. */
@@ -363,19 +369,52 @@ MVMuint32 MVM_string_utf8_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds
             }
             at_start = 0;
         }
-        while (pos < cur_bytes->length) {
-            switch(decode_utf8_byte(&state, &codepoint, bytes[pos++])) {
-            case UTF8_ACCEPT: {
-                MVMint32 first = 1;
-                MVMGrapheme32 g;
-                last_accept_bytes = cur_bytes;
-                last_accept_pos = pos;
-                ready = MVM_unicode_normalizer_process_codepoint_to_grapheme(tc, &(ds->norm), codepoint, &g);
-                while (ready--) {
-                    if (first)
-                        first = 0;
-                    else
-                        g = MVM_unicode_normalizer_get_grapheme(tc, &(ds->norm));
+
+        /* We have both a fast path and a slow path for UTF-8 decoding. The
+         * fast path covers the common case where we have no chars that are
+         * significant to normalization, and so we can skip the normalizer.
+         * This is true of the ASCII and Latin-1 ranges of UTF-8, with the
+         * exception of \r. Note that since the following codepoint may be
+         * the one that causes us to need to compose, we need a lag of 1
+         * codepoint. */
+        if (can_fast_path) {
+            /* Lift the no lag codepoint case out of the hot loop below,
+             * to save on a couple of branches. */
+            MVMCodepoint first_significant = ds->norm.first_significant;
+            while (lag_codepoint == -1 && pos < cur_bytes->length) {
+                switch(decode_utf8_byte(&state, &codepoint, bytes[pos++])) {
+                case UTF8_ACCEPT: {
+                    if (codepoint == '\r' || codepoint >= first_significant) {
+                        can_fast_path = 0;
+                        goto slow_path;
+                    }
+                    lag_codepoint = codepoint;
+                    lag_last_accept_bytes = cur_bytes;
+                    lag_last_accept_pos = pos;
+                    break;
+                }
+                case UTF8_REJECT:
+                    MVM_exception_throw_adhoc(tc, "Malformed UTF-8");
+                    break;
+                }
+            }
+
+            while (pos < cur_bytes->length) {
+                switch(decode_utf8_byte(&state, &codepoint, bytes[pos++])) {
+                case UTF8_ACCEPT: {
+                    /* If we hit something that needs the normalizer, we put
+                     * any lagging codepoint into its buffer and jump to it. */
+                    if (codepoint == '\r' || codepoint >= first_significant) {
+                        MVM_unicode_normalizer_push_codepoints(tc, &(ds->norm),
+                            &lag_codepoint, 1);
+                        lag_codepoint = -1; /* Invalidate, we used it. */
+                        can_fast_path = 0;
+                        goto slow_path;
+                    }
+
+                    /* As we have a lagging codepoint, and this one does not
+                     * need normalization, then we know we can spit out the
+                     * lagging one. */
                     if (count == bufsize) {
                         /* Valid character, but we filled the buffer. Attach this
                         * one to the buffers linked list, and continue with a new
@@ -384,22 +423,87 @@ MVMuint32 MVM_string_utf8_decodestream(MVMThreadContext *tc, MVMDecodeStream *ds
                         buffer = MVM_malloc(bufsize * sizeof(MVMGrapheme32));
                         count = 0;
                     }
-                    buffer[count++] = g;
+                    buffer[count++] = lag_codepoint;
                     total++;
-                    if (stopper_chars && *stopper_chars == total) {
+                    if (MVM_string_decode_stream_maybe_sep(tc, seps, lag_codepoint)) {
                         reached_stopper = 1;
+                        last_accept_bytes = lag_last_accept_bytes;
+                        last_accept_pos = lag_last_accept_pos;
                         goto done;
                     }
-                    if (MVM_string_decode_stream_maybe_sep(tc, seps, g)) {
+                    else if (stopper_chars && *stopper_chars == total) {
                         reached_stopper = 1;
+                        last_accept_bytes = lag_last_accept_bytes;
+                        last_accept_pos = lag_last_accept_pos;
                         goto done;
                     }
+
+                    /* The current state becomes the lagged state. */
+                    lag_codepoint = codepoint;
+                    lag_last_accept_bytes = cur_bytes;
+                    lag_last_accept_pos = pos;
+                    break;
+                }
+                case UTF8_REJECT:
+                    MVM_exception_throw_adhoc(tc, "Malformed UTF-8");
+                    break;
                 }
-                break;
             }
-            case UTF8_REJECT:
-                MVM_exception_throw_adhoc(tc, "Malformed UTF-8");
-                break;
+            lag_last_accept_bytes = cur_bytes;
+            lag_last_accept_pos = pos;
+
+            /* If we fall out of the loop and have a lagged codepoint, but
+             * no next buffer, then we fall into the slow path to process it
+             * correctly. */
+            if (lag_codepoint != -1 && !cur_bytes->next) {
+                codepoint = lag_codepoint;
+                lag_codepoint = -1;
+                can_fast_path = 0;
+                goto slow_path;
+            }
+        }
+        else {
+            while (pos < cur_bytes->length) {
+                switch(decode_utf8_byte(&state, &codepoint, bytes[pos++])) {
+                case UTF8_ACCEPT: {
+                    MVMGrapheme32 g;
+                    MVMint32 first;
+                  slow_path:
+                    first = 1;
+                    last_accept_bytes = cur_bytes;
+                    last_accept_pos = pos;
+                    ready = MVM_unicode_normalizer_process_codepoint_to_grapheme(tc,
+                        &(ds->norm), codepoint, &g);
+                    while (ready--) {
+                        if (first)
+                            first = 0;
+                        else
+                            g = MVM_unicode_normalizer_get_grapheme(tc, &(ds->norm));
+                        if (count == bufsize) {
+                            /* Valid character, but we filled the buffer. Attach this
+                            * one to the buffers linked list, and continue with a new
+                            * one. */
+                            MVM_string_decodestream_add_chars(tc, ds, buffer, bufsize);
+                            buffer = MVM_malloc(bufsize * sizeof(MVMGrapheme32));
+                            count = 0;
+                        }
+                        buffer[count++] = g;
+                        total++;
+                        if (MVM_string_decode_stream_maybe_sep(tc, seps, g)) {
+                            reached_stopper = 1;
+                            goto done;
+                        }
+                        else if (stopper_chars && *stopper_chars == total) {
+                            reached_stopper = 1;
+                            goto done;
+                        }
+                    }
+                    break;
+                }
+                case UTF8_REJECT:
+                    MVM_exception_throw_adhoc(tc, "Malformed UTF-8");
+                    break;
+                }
             }
         }
         cur_bytes = cur_bytes->next;
@@ -470,9 +574,7 @@ char * MVM_string_utf8_encode_substr(MVMThreadContext *tc,
         else {
             MVM_free(result);
             MVM_free(repl_bytes);
-            MVM_exception_throw_adhoc(tc,
-                "Error encoding UTF-8 string: could not encode codepoint %d",
-                cp);
+            MVM_string_utf8_throw_encoding_exception(tc, cp);
         }
     }
 
@@ -502,3 +604,22 @@ char * MVM_string_utf8_encode_C_string(MVMThreadContext *tc, MVMString *str) {
     result[output_size] = (char)0;
     return result;
 }
+
+void MVM_string_utf8_throw_encoding_exception (MVMThreadContext *tc, MVMCodepoint cp) {
+    const char *gencat = MVM_unicode_codepoint_get_property_cstr(tc, cp, MVM_UNICODE_PROPERTY_GENERAL_CATEGORY);
+    if(cp > 0x10FFFF) {
+        MVM_exception_throw_adhoc(tc,
+            "Error encoding UTF-8 string: could not encode codepoint %d (0x%X), codepoint out of bounds. Cannot encode higher than %d (0x%X)",
+            cp, cp, 0x10FFFF, 0x10FFFF);
+    }
+    else if (strcmp("Cs", gencat) == 0) {
+        MVM_exception_throw_adhoc(tc,
+            "Error encoding UTF-8 string: could not encode Unicode Surrogate codepoint %d (0x%X)",
+            cp, cp);
+    }
+    else {
+        MVM_exception_throw_adhoc(tc,
+            "Error encoding UTF-8 string: could not encode codepoint %d (0x%X)",
+            cp, cp);
+    }
+}
