@@ -42,10 +42,14 @@ typedef struct MVMExceptionBody MVMExceptionBody;
 typedef struct MVMExtOpRecord MVMExtOpRecord;
 typedef struct MVMExtOpRegistry MVMExtOpRegistry;
 typedef struct MVMExtRegistry MVMExtRegistry;
+typedef struct MVMRegionAlloc MVMRegionAlloc;
+typedef struct MVMRegionBlock MVMRegionBlock;
 typedef struct MVMFixedSizeAlloc MVMFixedSizeAlloc;
 typedef struct MVMFixedSizeAllocFreeListEntry MVMFixedSizeAllocFreeListEntry;
 typedef struct MVMFixedSizeAllocSafepointFreeListEntry MVMFixedSizeAllocSafepointFreeListEntry;
 typedef struct MVMFixedSizeAllocSizeClass MVMFixedSizeAllocSizeClass;
+typedef struct MVMFixedSizeAllocThread MVMFixedSizeAllocThread;
+typedef struct MVMFixedSizeAllocThreadSizeClass MVMFixedSizeAllocThreadSizeClass;
 typedef struct MVMFrame MVMFrame;
 typedef struct MVMFrameHandler MVMFrameHandler;
 typedef struct MVMGen2Allocator MVMGen2Allocator;
@@ -77,7 +81,7 @@ typedef struct MVMNFAStateInfo MVMNFAStateInfo;
 typedef struct MVMNFGState MVMNFGState;
 typedef struct MVMNFGSynthetic MVMNFGSynthetic;
 typedef struct MVMNFGTrieNode MVMNFGTrieNode;
-typedef struct MVMNGFTrieNodeEntry MVMNGFTrieNodeEntry;
+typedef struct MVMNFGTrieNodeEntry MVMNFGTrieNodeEntry;
 typedef struct MVMNativeCall MVMNativeCall;
 typedef struct MVMNativeCallBody MVMNativeCallBody;
 typedef struct MVMNativeRef MVMNativeRef;
@@ -187,6 +191,7 @@ typedef struct MVMThreadBody MVMThreadBody;
 typedef struct MVMThreadContext MVMThreadContext;
 typedef struct MVMUnicodeNamedValue MVMUnicodeNamedValue;
 typedef struct MVMUnicodeNameRegistry MVMUnicodeNameRegistry;
+typedef struct MVMUnicodeGraphemeNameRegistry MVMUnicodeGraphemeNameRegistry;
 typedef struct MVMUninstantiable MVMUninstantiable;
 typedef struct MVMWorkThread MVMWorkThread;
 typedef struct MVMIOOps MVMIOOps;
@@ -224,6 +229,7 @@ typedef struct MVMJitGuard MVMJitGuard;
 typedef struct MVMJitInvoke MVMJitInvoke;
 typedef struct MVMJitJumpList MVMJitJumpList;
 typedef struct MVMJitControl MVMJitControl;
+typedef struct MVMJitData MVMJitData;
 typedef struct MVMJitCode MVMJitCode;
 typedef struct MVMProfileThreadData MVMProfileThreadData;
 typedef struct MVMProfileGC MVMProfileGC;
