@@ -1,20 +1,33 @@
+#!/usr/bin/env perl
 use v5.14;
 use warnings; use strict;
+use utf8;
+use feature 'unicode_strings';
 use Data::Dumper;
-use Carp qw(cluck);
+use Carp qw(cluck croak);
 $Data::Dumper::Maxdepth = 1;
 # Make C versions of the Unicode tables.
 
-# Before running, download zip files from http://www.unicode.org/Public/zipped/
-# and extract them in UNIDATA
+# Before running, downloading UNIDATA.zip from http://www.unicode.org/Public/zipped/
+# and extract them into UNIDATA in the root of this repository.
+
+# Download allkeys.txt from http://www.unicode.org/Public/UCA/latest/
+# and place into a folder named UCA under the UNIDATA directory.
 
 my $DEBUG = $ENV{UCD2CDEBUG} // 0;
 
 my @name_lines;
-open(LOG, ">extents") or die "can't create extents: $!" if $DEBUG;
+if $DEBUG {
+    open(LOG, ">extents") or die "can't create extents: $!";
+    binmode LOG, ':encoding(UTF-8)';
+}
+binmode STDOUT, ':encoding(UTF-8)';
+binmode STDERR, ':encoding(UTF-8)';
 my $LOG;
 
 my $db_sections = {};
+my $sequences = {};
+my $hout = "";
 my $h_sections = {};
 my $planes = [];
 my $points_by_hex = {};
@@ -24,6 +37,7 @@ my $binary_properties = {};
 my $first_point = undef;
 my $last_point = undef;
 my $aliases = {};
+my $alias_types = {};
 my $prop_names = {};
 my $named_sequences = {};
 my $bitfield_table = [];
@@ -47,10 +61,17 @@ my %is_subtype = (
 );
 my $gc_alias_checkers = [];
 
+sub trim { my $s = shift; $s =~ s/^\s+|\s+$//g; return $s };
+
 sub progress($);
 sub main {
     $db_sections->{'AAA_header'} = header();
-
+    add_unicode_sequence('emoji/emoji-sequences');
+    add_unicode_sequence('emoji/emoji-zwj-sequences');
+    add_unicode_sequence('NamedSequences');
+    gen_unicode_sequence_keypairs();
+    NameAliases();
+    gen_name_alias_keypairs();
     # Load all the things
     UnicodeData(
         derived_property('BidiClass', 'Bidi_Class', {}, 0),
@@ -58,13 +79,16 @@ sub main {
         derived_property('CombiningClass',
             'Canonical_Combining_Class', { Not_Reordered => 0 }, 1)
     );
+    collation();
+    BidiMirroring();
     goto skip_most if $skip_most_mode;
     binary_props('extracted/DerivedBinaryProperties');
-    enumerated_property('ArabicShaping', 'Joining_Type', {}, 0, 2);
+    binary_props('emoji/emoji-data');
     enumerated_property('ArabicShaping', 'Joining_Group', {}, 0, 3);
-    enumerated_property('BidiMirroring', 'Bidi_Mirroring_Glyph', { '' => 0 }, 1, 1);
     enumerated_property('Blocks', 'Block', { No_Block => 0 }, 1, 1);
     enumerated_property('extracted/DerivedDecompositionType', 'Decomposition_Type', { None => 0 }, 1, 1);
+    enumerated_property('extracted/DerivedEastAsianWidth', 'East_Asian_Width', {}, 0, 1);
+    enumerated_property('ArabicShaping', 'Joining_Type', {}, 0, 2);
     CaseFolding();
     SpecialCasing();
     enumerated_property('DerivedAge',
@@ -89,12 +113,11 @@ sub main {
         'Hangul_Syllable_Type', { Not_Applicable => 0 }, 1, 1);
     Jamo();
     LineBreak();
-    NameAliases();
     NamedSequences();
     binary_props('PropList');
     enumerated_property('Scripts', 'Script', { Unknown => 0 }, 1, 1);
     # XXX StandardizedVariants.txt # no clue what this is
-    break_property('Grapheme', 'Grapheme_Cluster_Break');
+    grapheme_cluster_break('Grapheme', 'Grapheme_Cluster_Break');
     break_property('Sentence', 'Sentence_Break');
   skip_most:
     break_property('Word', 'Word_Break');
@@ -186,6 +209,9 @@ sub apply_to_range {
     # be applied to all/any in between.
     my $range = shift;
     chomp($range);
+    if ( !defined $range ) {
+        cluck "Did not get any range in apply_to_range";
+    }
     my $fn = shift;
     my ($first, $last) = split '\\.\\.', $range;
     $first ||= $range;
@@ -233,6 +259,15 @@ sub break_property {
     enumerated_property("auxiliary/${fname}BreakProperty",
         $pname, { Other => 0 }, 1, 1);
 }
+sub grapheme_cluster_break {
+    my ($fname, $pname) = @_;
+    enumerated_property("auxiliary/${fname}BreakProperty",
+        $pname, {
+            # Should not be set to Other for this one ?
+            Other => 0,
+
+        }, 1, 1);
+}
 
 sub derived_property {
     # filename, property name, property object, starting counter
@@ -456,7 +491,7 @@ sub emit_extent_fate {
 sub add_extent($$) {
     my ($extents, $extent) = @_;
     if ($DEBUG) {
-        $LOG .= "\n" . join '', 
+        $LOG .= "\n" . join '',
             grep /code|fate|name|bitfield/,
             sort split /^/m, "EXTENT " . Dumper($extent);
     }
@@ -601,7 +636,7 @@ sub emit_codepoint_row_lookup {
     MVMint32 plane = codepoint >> 16;
 
     if (codepoint < 0) {
-        MVM_exception_throw_adhoc(tc, \"should eventually be unreachable\");
+        MVM_exception_throw_adhoc(tc, \"Error, MoarVM cannot get Unicode codepoint property for synthetic codepoint %i", codepoint);
     }
 
     if (plane == 0) {"
@@ -687,7 +722,7 @@ sub emit_bitfield {
 sub emit_property_value_lookup {
     my $allocated = shift;
     my $enumtables = "\n\n";
-    my $hout = "typedef enum {\n";
+    our $hout = "typedef enum {\n";
     my $out = "
 static MVMint32 MVM_unicode_get_property_int(MVMThreadContext *tc, MVMint32 codepoint, MVMint64 property_code) {
     MVMuint32 switch_val = (MVMuint32)property_code;
@@ -799,7 +834,27 @@ static const char* MVM_unicode_get_property_str(MVMThreadContext *tc, MVMint32 c
 ";  # or should we try to stringify numeric value?
 
     $hout .= "} MVM_unicode_property_codes;";
- 
+
+    sub gen_pvalue_defines {
+        my ( $property_name_mvm, $property_name, $short_pval_name ) = @_;
+        my $GCB_h;
+        $GCB_h .= "\n\n/* $property_name_mvm */\n";
+        my %seen;
+        foreach my $key (sort keys % {$enumerated_properties->{$property_name}->{'enum'} }  ) {
+            next if $seen{$key};
+            my $value = $enumerated_properties->{$property_name}->{'enum'}->{$key};
+            $key = 'MVM_UNICODE_PVALUE_' . $short_pval_name . '_' . uc $key;
+            $key =~ tr/\./_/;
+            $GCB_h .= "#define $key $value\n";
+            $seen{$key} = 1;
+        }
+        $hout .= $GCB_h;
+    }
+    gen_pvalue_defines('MVM_UNICODE_PROPERTY_GRAPHEME_CLUSTER_BREAK', 'Grapheme_Cluster_Break', 'GCB');
+    gen_pvalue_defines('MVM_UNICODE_PROPERTY_DECOMPOSITION_TYPE', 'Decomposition_Type', 'DT');
+    gen_pvalue_defines('MVM_UNICODE_PROPERTY_CANONICAL_COMBINING_CLASS', 'Canonical_Combining_Class', 'CCC');
+    gen_pvalue_defines('MVM_UNICODE_PROPERTY_NUMERIC_TYPE', 'Numeric_Type', 'Numeric_Type');
+
     $db_sections->{MVM_unicode_get_property_int} = $enumtables . $eout . $out;
     $h_sections->{property_code_definitions} = $hout;
 }
@@ -869,7 +924,6 @@ MVMint32 MVM_unicode_is_in_block(MVMThreadContext *tc, MVMString *str, MVMint64
 
     return in_block;
 }";
-
     $db_sections->{block_lookup} = $out;
     $h_sections->{block_lookup} = $hout;
 }
@@ -899,6 +953,8 @@ static void generate_codepoints_by_name(MVMThreadContext *tc) {
     MVMint32 extent_index = 0;
     MVMint32 codepoint = 0;
     MVMint32 codepoint_table_index = 0;
+    MVMint16 i = num_unicode_namealias_keypairs - 1;
+
     MVMUnicodeNameRegistry *entry;
     for (; extent_index < MVM_NUM_UNICODE_EXTENTS; extent_index++) {
         MVMint32 length;
@@ -941,25 +997,14 @@ static void generate_codepoints_by_name(MVMThreadContext *tc) {
             }
         }
     }
-    entry = MVM_malloc(sizeof(MVMUnicodeNameRegistry));
-    entry->name = "LF";
-    entry->codepoint = 10;
-    HASH_ADD_KEYPTR(hash_handle, codepoints_by_name, "LF", 2, entry);
+    for (; i >= 0; i--) {
+        entry = MVM_malloc(sizeof(MVMUnicodeNameRegistry));
+        entry->name = uni_namealias_pairs[i].name;
+        entry->codepoint =  uni_namealias_pairs[i].codepoint;
+        HASH_ADD_KEYPTR(hash_handle, codepoints_by_name,  uni_namealias_pairs[i].name,  uni_namealias_pairs[i].strlen, entry);
 
-    entry = MVM_malloc(sizeof(MVMUnicodeNameRegistry));
-    entry->name = "FF";
-    entry->codepoint = 12;
-    HASH_ADD_KEYPTR(hash_handle, codepoints_by_name, "FF", 2, entry);
+    }
 
-    entry = MVM_malloc(sizeof(MVMUnicodeNameRegistry));
-    entry->name = "CR";
-    entry->codepoint = 13;
-    HASH_ADD_KEYPTR(hash_handle, codepoints_by_name, "CR", 2, entry);
-
-    entry = MVM_malloc(sizeof(MVMUnicodeNameRegistry));
-    entry->name = "NEL";
-    entry->codepoint = 133;
-    HASH_ADD_KEYPTR(hash_handle, codepoints_by_name, "NEL", 3, entry);
 }
 END
     $db_sections->{names_hash_builder} = $out;
@@ -1050,9 +1095,111 @@ static const MVMUnicodeNamedValue unicode_property_keypairs[".scalar(@lines)."]
     $db_sections->{BBB_unicode_property_keypairs} = $out;
     $h_sections->{MVMUnicodeNamedValue} = $hout;
 }
+sub add_unicode_sequence {
+    my ($filename) = @_;
+    each_line($filename, sub { my $line = shift;
+        if ( $line =~ /^#/ or $line =~ /^\s*$/) {
+            return;
+        }
+        my (@list, $hex_ords, $type, $name);
+        @list = split /;|   \#/, $line;
+        if ($filename =~ /emoji/) {
+            $hex_ords = trim shift @list;
+            $type = trim shift @list;
+            $name = trim shift @list;
+        }
+        else {
+            $name = trim shift @list;
+            $hex_ords = trim shift @list;
+            $type = 'NamedSequences';
+        }
+        # Make sure it's uppercase since the Emoji sequences are not all in
+        # uppercase.
+        $name = uc $name;
+        # Emoji sequences have commas in some and these cannot be included
+        # since they seperate seperate named items in ISO notation that P6 uses
+        $name =~ s/,//g;
+        $sequences->{$name}->{'type'} = $type;
 
+        for my $hex (split ' ', $hex_ords) {
+            push @{$sequences->{$name}->{'ords'}}, hex $hex;
+        }
+    } );
+}
+sub gen_unicode_sequence_keypairs {
+    my $count = 0;
+    my $string_seq;
+    my $seq_c_hash_str;
+    my @seq_c_hash_array;
+    my $enum_table;
+    $string_seq .= "/* Unicode sequences such as Emoji sequences */\n";
+    for my $thing ( sort keys %$sequences ) {
+        my $seq_name = "uni_seq_$count";
+        $string_seq .=  "static const MVMint32 $seq_name" . "[] = {";
+        $seq_c_hash_str .= '{"' . $thing . '",' . $count . '},';
+        my $ord_data;
+        for my $ord ( @{$sequences->{$thing}->{'ords'}} ) {
+            $ord_data .= '0x' . uc sprintf("%x", $ord) . ',';
+        }
+        $ord_data = scalar @{$sequences->{$thing}->{'ords'}} . ',' . $ord_data;
+        $string_seq .= $ord_data;
+        $ord_data =~ s/,$//;
+        $string_seq =~ s/,$//;
+        $string_seq = $string_seq . "}; " . "/* $thing */ /*" . $sequences->{$thing}->{'type'} . " */\n";
+        $enum_table = $enum_table . "$seq_name,\n";
+        $count++;
+        if ( length $seq_c_hash_str > 80 ) {
+            push @seq_c_hash_array, $seq_c_hash_str . "\n";
+            $seq_c_hash_str = '';
+        }
+    }
+    push @seq_c_hash_array, $seq_c_hash_str . "\n";
+    $seq_c_hash_str = join '    ', @seq_c_hash_array;
+    $seq_c_hash_str =~ s/\s*,\s*$//;
+    $seq_c_hash_str .= "\n};";
+    $seq_c_hash_str = "static const MVMUnicodeNamedValue uni_seq_pairs[$count] = {\n    " . $seq_c_hash_str;
+
+    $enum_table =~ s/\s*,\s*$/};/;
+    $enum_table = "static const MVMint32 * uni_seq_enum[$count] = {\n" . $enum_table;
+    $db_sections->{uni_seq} = $seq_c_hash_str . $string_seq . $enum_table;
+    $hout .= "#define num_unicode_seq_keypairs " . $count ."\n";
+}
+sub gen_name_alias_keypairs {
+    my $count = 0;
+    my $seq_c_hash_str;
+    my @seq_c_hash_array;
+    for my $thing ( sort keys %$alias_types ) {
+        my $ord_data;
+        my $ord = $alias_types->{$thing}->{'code'};
+        $ord_data .= '0x' . uc sprintf("%x", $ord) . ',';
+        $seq_c_hash_str .= '{"' . $thing . '",' . $ord_data . (length $thing) . '},';
+        $ord_data =~ s/,$//;
+        my $type = $alias_types->{$thing}->{'type'};
+        $count++;
+        if ( length $seq_c_hash_str > 80 ) {
+            push @seq_c_hash_array, $seq_c_hash_str . "\n";
+            $seq_c_hash_str = '';
+        }
+    }
+    push @seq_c_hash_array, $seq_c_hash_str . "\n";
+    $seq_c_hash_str = join '    ', @seq_c_hash_array;
+    $seq_c_hash_str =~ s/\s*,\s*$//;
+    $seq_c_hash_str .= "\n};";
+    $seq_c_hash_str = "static const MVMUnicodeNamedAlias uni_namealias_pairs[$count] = {\n    " . $seq_c_hash_str;
+
+    $seq_c_hash_str = "/* Unicode Name Aliases */\n" . $seq_c_hash_str;
+    $db_sections->{Auni_namealias} = $seq_c_hash_str;
+    $hout .= "#define num_unicode_namealias_keypairs " . $count ."\n";
+    $hout .= <<'END'
+struct MVMUnicodeNamedAlias {
+    char *name;
+    MVMGrapheme32 codepoint;
+    MVMint16 strlen;
+};
+typedef struct MVMUnicodeNamedAlias MVMUnicodeNamedAlias;
+END
+}
 sub emit_unicode_property_value_keypairs {
-    my $hout = "";
     my @lines = ();
     my $property;
     for (keys %$enumerated_properties) {
@@ -1310,6 +1457,7 @@ authorization of the copyright holder. */
 sub read_file {
     my $fname = shift;
     open FILE, $fname or die "Couldn't open file '$fname': $!";
+    binmode FILE, ':encoding(UTF-8)';
     my @lines = ();
     while( <FILE> ) {
         push @lines, $_;
@@ -1321,6 +1469,7 @@ sub read_file {
 sub write_file {
     my ($fname, $contents) = @_;
     open FILE, ">$fname" or die "Couldn't open file '$fname': $!";
+    binmode FILE, ':encoding(UTF-8)';
     print FILE $contents;
     close FILE;
 }
@@ -1367,10 +1516,13 @@ sub UnicodeData {
 
         my $code = hex $code_str;
         my $plane_num = $code >> 16;
-        if ($name eq '<control>' || $name eq '') {
-            $name = $u1name;
+        if ($name eq '<control>' ) {
+            $name = sprintf '<control-%.4X>', $code;
         }
         my $point = {
+            # Unicode_1_Name is not used yet. We should make sure it ends up
+            # in some data structure
+            Unicode_1_Name => $u1name,
             code_str => $code_str,
             name => $name,
             gencat_name => $gencat,
@@ -1385,6 +1537,7 @@ sub UnicodeData {
             NFKD_QC => 1,
             NFKC_QC => 1,
             NFG_QC => 1,
+            MVM_COLLATION_QC => 1,
             code => $code,
             Any => 1
         };
@@ -1572,16 +1725,131 @@ sub Jamo {
         $points_by_hex->{$code_str}->{Jamo_Short_Name} = $name;
     });
 }
+sub BidiMirroring {
+    my $file = 'BidiMirroring';
+    my $propname = 'Bidi_Mirroring_Glyph';
+    my $max_size = 0;
+    each_line('BidiMirroring', sub { $_ = shift;
+        my $line = $_;
+        my ($range, $int) = split /\s*[;#]\s*/, $line;
+        $int = hex $int or die;
+        $max_size = $int if $max_size < $int;
+        apply_to_range($range, sub {
+            my $point = shift;
+            $point->{$propname} = $int;
+        });
+    });
 
+    register_int_property($propname, $max_size);
+}
+sub collation {
+    my $enum = {};
+    my $base = { enum => $enum };
+    my $j = 0;
+    my $name_primary = 'MVM_COLLATION_PRIMARY';
+    my $name_secondary = 'MVM_COLLATION_SECONDARY';
+    my $name_tertiary = 'MVM_COLLATION_TERTIARY';
+    my $implicit = 0;
+    # Record the highest value we see so we can save some bits in the bitfield
+    my $primary_max = 0;
+    my $secondary_max = 0;
+    my $tertiary_max = 0;
+    ## Sample line from allkeys.txt
+    #1D4FF ; [.1EE3.0020.0005] # MATHEMATICAL BOLD SCRIPT SMALL V
+    my $line_no = 0;
+    each_line('UCA/allkeys', sub { $_ = shift;
+        my $line = $_;
+        $line_no++;
+        my ($code, $weight1, $weight2, $weight3, $temp);
+        if ($line =~ s/ ^ \@implicitweights \s+ //xms ) {
+            ($code, $weight1) = split / (?: [;\[\]]|\s )+ /xms, $line;
+            $weight1 = hex $weight1 or croak;
+            $code or croak;
+            $implicit = 1;
+        }
+        elsif ( $line =~ /^\s*[#@]/ or $line =~ /^\s*$/ ) {
+            $line = '';
+            return;
+        }
+        else {
+            ($code, $temp) = split /[;#]+/, $line;
+            $code = trim $code;
+            my @codes = split / /, $code;
+            # we don't yet support collation for multiple codepoints
+            if ( scalar @codes > 1 ) {
+                # For now set MVM_COLLATION_QC = 0 for these cp
+                apply_to_range($codes[0], sub {
+                    my $point = shift;
+                    $point->{'MVM_COLLATION_QC'} = 0;
+                });
+                return;
+            }
+            # We capture the `.` or `*` before each weight. Currently we do
+            # not use this information, but it may be of use later (we currently
+            # don't put their values into the data structure.
+
+            # When multiple tables are specified for a character, it is because those
+            # are the composite values for the decomposed character. Since we compare
+            # in NFC form not NFD, let's add these together.
+
+            while ( $temp =~ / (:? \[ ([.*]) (\p{AHex}+) ([.*]) (\p{AHex}+) ([.*]) (\p{AHex}+) \] ) /xmsg ) {
+                $weight1 += hex $3;
+                $weight2 += hex $5;
+                $weight3 += hex $7;
+            }
+
+        }
+        if ( !defined $code or !defined $weight1 or !defined $weight2 or !defined $weight3 ) {
+            unless ( $implicit and defined $weight1 ) {
+                die "Line no $line_no: line:[$line] weight1:[$weight1] weight2:[$weight2] weight3:[$weight3]";
+            }
+        }
+        apply_to_range($code, sub {
+            my $point = shift;
+            # Add one to the value so we can distinguish between specified values
+            # of zero for collation weight and null values.
+            $point->{$name_primary} = 1;
+            if ($weight1) {
+                $point->{$name_primary} += $weight1 if $weight1;
+                $primary_max = $weight1 if $weight1 > $primary_max;
+            }
+            $point->{$name_secondary} = 1;
+            if ($weight2) {
+                $point->{$name_secondary} += $weight2;
+                $secondary_max = $weight2 if $weight2 > $secondary_max;
+            }
+            $point->{$name_tertiary}  = 1;
+            if ($weight3) {
+                $point->{$name_tertiary} += $weight3  if $weight3 ;
+                $tertiary_max = $weight3 if $weight3 > $tertiary_max;
+            }
+        });
+    });
+    for ( $primary_max, $secondary_max, $tertiary_max ) {
+        if ( $_ < 1 ) {
+            die "Oh no! One of the highest collation numbers I saw is less than 0. Something is wrong" .
+              "Primary max: $primary_max secondary max: $secondary_max tertiary_max: $tertiary_max";
+        }
+    }
+
+    register_int_property($name_primary, $primary_max);
+    register_int_property($name_secondary, $secondary_max);
+    register_int_property($name_tertiary, $tertiary_max);
+    register_binary_property('MVM_COLLATION_QC');
+}
 sub LineBreak {
     my $enum = {};
     my $base = { enum => $enum };
     my $j = 0;
-    $enum->{$_} = $j++ for ("BK", "CR", "LF", "CM", "SG", "GL",
-        "CB", "SP", "ZW", "NL", "WJ", "JL", "JV", "JT", "H2", "H3");
+    $enum->{$_} = $j++ for ("BK", "CM", "CR", "GL", "LF", "NL", "SP",
+        "WJ", "ZW", "ZWJ", "AI", "AL", "B2", "BA", "BB", "CB", "CJ", "CL", "CP", "EB",
+        "EM", "EX", "H2", "H3", "HL", "HY", "ID", "IN", "IS", "JL",
+        "JT", "JV", "NS", "NU", "OP", "PO", "PR", "QU", "RI", "SA",
+        "SG", "SY", "XX"
+        );
     each_line('LineBreak', sub { $_ = shift;
         my ($range, $name) = split /\s*[;#]\s*/;
-        return unless exists $enum->{$name}; # only normative
+        die "Can't find Line_Break property $name in the enum" unless exists $enum->{$name}; # only normative
         apply_to_range($range, sub {
             my $point = shift;
             $point->{Line_Break} = $enum->{$name};
@@ -1592,14 +1860,16 @@ sub LineBreak {
         $keys[$base->{enum}->{$key}] = $key;
     }
     $base->{keys} = \@keys;
-    $base->{bit_width} = int(log($j)/log(2) - 0.00001) + 1;
+    $base->{bit_width} = least_int_ge_lg2($j);
     register_enumerated_property('Line_Break', $base);
 }
 
 sub NameAliases {
     each_line('NameAliases', sub { $_ = shift;
-        my ($code_str, $name) = split /\s*[;#]\s*/;
+        my ($code_str, $name, $type) = split /\s*[;#]\s*/;
         $aliases->{$name} = hex $code_str;
+        $alias_types->{$name}->{'code'} = hex $code_str;
+        $alias_types->{$name}->{'type'} = $type;
     });
 }
 
@@ -1645,6 +1915,18 @@ sub tweak_nfg_qc {
         elsif ($point->{'gencat_name'} eq 'Mc' || $code == 0x0E33 || $code == 0x0EB3) {
             $point->{'NFG_QC'} = 0;
         }
+        # For now set all Emoji to NFG_QC 0
+        # Eventually we will only want to set the ones that are NOT specified
+        # as ZWJ sequences
+        elsif ($point->{'Emoji'} ) {
+            $point->{'NFG_QC'} = 0;
+        }
+        elsif ($point->{'Grapheme_Cluster_Break'}) {
+            $point->{'NFG_QC'} = 0;
+        }
+        elsif ($point->{'Prepended_Concatenation_Mark'}) {
+            $point->{'NFG_QC'} = 0;
+        }
     }
 }
 
@@ -1657,6 +1939,16 @@ sub register_binary_property {
     } unless exists $binary_properties->{$name};
 }
 
+sub register_int_property {
+    my ( $name, $elems ) = @_;
+    # add to binary_properties for now
+    $all_properties->{$name} = $binary_properties->{$name} = {
+        property_index => $property_index++,
+        name => $name,
+        bit_width => least_int_ge_lg2($elems)
+    } unless exists $binary_properties->{$name};
+}
+
 sub register_enumerated_property {
     my ($pname, $obj) = @_;
     die if exists $enumerated_properties->{$pname};
